"""Module contains a list sdkloop errors. One exception class - one error kind"""

from typing import Optional

from lunavl.sdk.errors.errors import ErrorInfo


class LoopError(Exception):
    """
    Base class for sdkloop exceptions

    Attributes:
        error (ErrorInfo): error info
    """

    # default error, must be determinate in 'final' class
    error: ErrorInfo

    def __init__(self, error: Optional[ErrorInfo] = None, message=""):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
        if error is not None:
            self.error = error
        else:
            self.error = self.error

    def asDict(self) -> dict:
        """
        Convert exception to dict
        """
        return {"error_code": self.errorCode, "desc": self.description, "detail": self.detail, "link": self.link}

    @classmethod
    def format(cls, *formatArgs) -> "LoopError":
        """
        Format error detail
        Args:
            *formatArgs: format args

        Returns:
            new error with detail formatting
        """
        newException = cls(cls.error.format(*formatArgs))
        return newException

    @property
    def detail(self) -> str:
        """Get error detail, mimicry to ErrorInfo class"""
        return self.error.detail

    @property
    def description(self) -> str:
        """Get error description, mimicry to ErrorInfo class"""
        return self.error.description

    @property
    def errorCode(self) -> int:
        """Get error code, mimicry to ErrorInfo class"""
        return self.error.errorCode

    @property
    def link(self) -> str:
        """Get link with error info"""
        return f"https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-{self.errorCode}"

    def __eq__(self, other):
        """Check equality, mimicry to ErrorInfo class"""
        return isinstance(other, LoopError) and self.errorCode == other.errorCode


class PlatformError(LoopError):
    """Platform specific errors"""


class ReadImageError(PlatformError):
    """Load image error"""

    error = ErrorInfo(18003, "Bad/incomplete input data", "Failed read bytes as image")


class BadWarp(PlatformError):
    """Image is not VisionLabs error"""

    error = ErrorInfo(11043, "Bad/incomplete input data", "{}")


class MultipleFaces(PlatformError):
    """Processing several face detection error"""

    error = ErrorInfo(11038, "Multiple faces found", "Multiple faces found in image {} , detect {}")


class MultipleBodies(PlatformError):
    """Processing several body detection error"""

    error = ErrorInfo(34008, "Multiple bodies found", "Multiple bodies found in image {} , detect {}")


class InvalidType(PlatformError):
    """Unsupported image format (including broken image)"""

    error = ErrorInfo(100033, "Unsupported image type", "{}")


class LoopRuntimeError(LoopError):
    """Loop runtime errors"""


class UnknownDescriptor(LoopRuntimeError):
    """Not supported descriptor version (including not valid version)"""

    error = ErrorInfo(200003, "Bad/incomplete input data", "Not supported descriptor version {}")


class EstimatorInitializationError(LoopRuntimeError):
    """An estimator initialization failed"""

    error = ErrorInfo(200004, "Internal error", "Estimator '{}' initialization fail")


class EstimatorWasNotInitializedError(LoopRuntimeError):
    """An estimator was not initialized"""

    error = ErrorInfo(200005, "Internal error", "'{}' estimator was not initialized")


class SDKError(LoopError):
    """SDL exception wrap"""


def _check() -> None:
    """
    Check error codes' uniqueness.
    """
    registeredErrorCodes = set()

    def rCheck(cls):
        if error := getattr(cls, "error", None):
            if not isinstance(error, ErrorInfo):
                raise RuntimeError("Does not valid error type")
            if error.errorCode in registeredErrorCodes:
                raise RuntimeError(f"Duplicated {error.errorCode} for error '{cls.__name__}'")
            registeredErrorCodes.add(error.errorCode)
        for subclasses in cls.__subclasses__():
            rCheck(subclasses)

    for subclasses in LoopError.__subclasses__():
        rCheck(subclasses)


_check()
del _check
