"""Module contains body descriptor estimator realization"""

from dataclasses import dataclass
from typing import Tuple, Union

from lunavl.sdk.descriptors.descriptors import BodyDescriptor, BodyDescriptorBatch
from lunavl.sdk.estimators.body_estimators.body_descriptor import BodyDescriptorEstimator
from lunavl.sdk.estimators.body_estimators.bodywarper import BodyWarp, BodyWarpedImage
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM
from ...utils.async_tools import gather


class BDMonitoringPointBase(EstimationPoint):
    """Body descriptor monitoring point base class.

    Attributes:
        descriptorVersion (int): version of descriptors which was estimated
    """

    __slots__ = ("descriptorVersion",)

    def __init__(self, descriptorVersion: int, executionTime: float = 0, batchSize: int = 0):
        super().__init__(batchSize=batchSize, executionTime=executionTime)
        self.descriptorVersion = descriptorVersion

    def getTags(self) -> dict:
        """Get monitoring tags. Add additional tag `descriptor_version`"""
        tags = super().getTags()
        tags["descriptor_version"] = self.descriptorVersion
        return tags


class BDMonitoringPoint(BDMonitoringPointBase):
    """Monitoring point for body descriptor estimation"""

    name = "body_descriptor"


class ABDMonitoringPoint(BDMonitoringPointBase):
    """Monitoring point for aggregated body descriptor estimation"""

    name = "aggregated_body_descriptor"


@dataclass
class BodyDescriptorSettings(EstimatorSettingsBase):
    """Body descriptor estimation settings"""

    # descriptor version
    descriptorVersion: int = 116


class BodyDescriptorQueueEstimator(
    QueueEstimatorBase[BodyWarp, BodyDescriptor, BodyDescriptorEstimator, BodyDescriptorSettings]
):
    """Body descriptor queue estimator realization"""

    async def _estimate(self, dataBatch: list[BodyWarp]) -> BodyDescriptorBatch:
        """
        Batch estimation body descriptor
        Args:
            dataBatch: body warp batch
        Returns:
            descriptors batch in the corresponding dataBatch order
        """
        with EM.monitoring(
            BDMonitoringPoint(batchSize=len(dataBatch), descriptorVersion=self.settings.descriptorVersion)
        ):
            res, _ = await self._estimator.estimateDescriptorsBatch(dataBatch, asyncEstimate=True)
            return res


class BodyDescriptorEstimationBroker(
    EstimationBrokerBase[BodyWarp, BodyDescriptor, BodyDescriptorEstimator, BodyDescriptorSettings]
):
    """
    Body descriptor estimation broker. Each broker corresponds to only one descriptor version.
    """

    @property
    def name(self) -> str:
        """Broker name"""
        return f"body_descriptor_{self._estimatorSettings.descriptorVersion}"

    @classmethod
    def createEstimator(cls, settings: BodyDescriptorSettings) -> tuple[BodyDescriptorEstimator, VLFaceEngine]:
        """Create body descriptor estimator"""
        return createEstimator(
            settings, lambda fe, lo: fe.createBodyDescriptorEstimator(settings.descriptorVersion, launchOptions=lo)
        )


class AggregatedBodyDescriptorQueueEstimator(
    QueueEstimatorBase[
        Union[BodyWarp, BodyWarpedImage],
        Tuple[BodyDescriptorBatch, BodyDescriptor],
        BodyDescriptorEstimator,
        BodyDescriptorSettings,
    ]
):
    """Aggregated body descriptor queue estimator realization"""

    async def _estimate(
        self, dataBatch: list[list[Union[BodyWarp, BodyWarpedImage]]]
    ) -> list[Tuple[BodyDescriptorBatch, BodyDescriptor]]:
        """
        Batch estimation aggregate body descriptor
        Args:
            dataBatch: batch of body warp batch
        Returns:
            list. Item of list is tuple <descriptors batch, aggregate descriptor> in the corresponding dataBatch order
        """
        with EM.monitoring(
            ABDMonitoringPoint(batchSize=len(dataBatch), descriptorVersion=self.settings.descriptorVersion)
        ):
            futures = (
                self._estimator.estimateDescriptorsBatch(warps=batch, asyncEstimate=True, aggregate=True)
                for batch in dataBatch
            )
            return await gather(futures)


class AggregatedBodyDescriptorEstimationBroker(
    EstimationBrokerBase[
        Union[BodyWarp, BodyWarpedImage],
        Tuple[BodyDescriptorBatch, BodyDescriptor],
        BodyDescriptorEstimator,
        BodyDescriptorSettings,
    ]
):
    """
    Aggregated body descriptor estimation broker. Each broker corresponds to only one descriptor version.
    """

    @property
    def name(self) -> str:
        """Broker name"""
        return f"aggregated_body_descriptor_{self._estimatorSettings.descriptorVersion}"

    def createEstimator(self, settings: BodyDescriptorSettings) -> tuple[BodyDescriptorEstimator, VLFaceEngine]:
        """Create aggregated body descriptor estimator"""
        return createEstimator(
            settings, lambda fe, lo: fe.createBodyDescriptorEstimator(settings.descriptorVersion, launchOptions=lo)
        )
