"""
Module contains body detector estimator realization.
"""

from dataclasses import dataclass

from lunavl.sdk.detectors.base import ImageForDetection, ImageForRedetection
from lunavl.sdk.detectors.bodydetector import BodyDetection, BodyDetector
from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.image_utils.image import VLImage

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM
from ...utils.validators import assertImageForReDetection


class BodyDetMonitoringPoint(EstimationPoint):
    """
    Body detector monitoring point. We get first (random) image for monitoring an image size and a body detection size.

    Attributes:
        detectionCount: detection count on all images from batch
        imageWidth: first image width from batch
        imageHeight: first image height from batch
        detectionWidth: first body detection width on first image from batch
        detectionHeight: first body detection height on first image from batch
    """

    __slots__ = (
        "detectionCount",
        "imageWidth",
        "imageHeight",
        "detectionWidth",
        "detectionHeight",
    )
    name = "body_detection"

    def __init__(self, imageWidth: int, imageHeight: int, executionTime: float = 0, batchSize: int = 0):
        super().__init__(batchSize=batchSize, executionTime=executionTime)
        self.detectionCount = 0
        self.imageWidth = imageWidth
        self.imageHeight = imageHeight
        self.detectionWidth = None
        self.detectionHeight = None

    def updateFromFirstImageResult(self, detectionResult: list[list[BodyDetection]]):
        """
        Update point of detection results
        Args:
            detectionResult: detection result
        """
        # firstImageDetections[0] may be None for re-detect
        if (firstImageDetections := detectionResult[0]) and firstImageDetections[0]:
            self.detectionWidth = firstImageDetections[0].boundingBox.rect.width
            self.detectionHeight = firstImageDetections[0].boundingBox.rect.height
        self.detectionCount = sum((len(detections) for detections in detectionResult))

    def getFields(self) -> dict:
        """Get fields for monitoring"""
        fields = super().getFields()
        fields["image_width"] = self.imageWidth
        fields["image_height"] = self.imageHeight
        fields["detection_count"] = self.detectionCount
        fields["body_detection_width"] = self.detectionWidth
        fields["body_detection_height"] = self.detectionHeight

        return fields


class BodyReDetMonitoringPointBase(BodyDetMonitoringPoint):
    """
    Body detector monitoring point. We get first (random) image for monitoring an image size, bbox
    and a body detection size.

    Attributes:
        bboxWidth: first bbox width on first image from batch
        bboxHeight: first bbox height on first image from batch
    """

    __slots__ = ("bboxWidth", "bboxHeight")

    name = "body_redetection"

    def __init__(
        self,
        bboxWidth: int,
        bboxHeight: int,
        imageWidth: int,
        imageHeight: int,
        executionTime: float = 0,
        batchSize: int = 0,
    ):
        super().__init__(
            batchSize=batchSize,
            executionTime=executionTime,
            imageWidth=imageWidth,
            imageHeight=imageHeight,
        )

        self.bboxWidth = bboxWidth
        self.bboxHeight = bboxHeight

    def getFields(self) -> dict:
        """Get fields for monitoring"""
        fields = super().getFields()
        fields["bbox_width"] = self.bboxWidth
        fields["bbox_height"] = self.bboxHeight
        return fields


@dataclass
class BodyDetectorSettings(EstimatorSettingsBase):
    """Body detector settings"""

    # max bodies count per image
    maxBodiesCount: int = 128


class BodyBrokerDetectorMixin:
    """Body detection base broker"""

    def createEstimator(self, settings: BodyDetectorSettings) -> tuple[BodyDetector, VLFaceEngine]:
        """Create body detector"""
        return createEstimator(settings, lambda fe, lo: fe.createBodyDetector(launchOptions=lo))


class BodyDetectionQueueEstimator(
    QueueEstimatorBase[VLImage | ImageForDetection, list[BodyDetection], BodyDetector, BodyDetectorSettings]
):
    """Body detection estimator estimator"""

    async def _estimate(self, dataBatch: list[VLImage | ImageForDetection]) -> list[list[BodyDetection]]:
        """Detect bodies on images"""
        firstImage = dataBatch[0]
        if isinstance(firstImage, VLImage):
            rect = firstImage.rect
        else:
            rect = firstImage.detectArea
        point = BodyDetMonitoringPoint(
            batchSize=len(dataBatch),
            imageWidth=rect.width,
            imageHeight=rect.height,
        )
        with EM.monitoring(point):
            res = await self._estimator.detect(dataBatch, asyncEstimate=True, limit=self.settings.maxBodiesCount)
            point.updateFromFirstImageResult(res)
            return res


DETECTOR_GENERIC = EstimationBrokerBase[
    VLImage | ImageForDetection, list[BodyDetection], BodyDetector, BodyDetectorSettings
]
REDETECTOR_GENERIC = EstimationBrokerBase[ImageForRedetection, list[BodyDetection], BodyDetector, BodyDetectorSettings]


class BodyDetectionEstimationBroker(BodyBrokerDetectorMixin, DETECTOR_GENERIC):
    """Body detection broker"""

    # broker name
    name = "body_detection"


class BodyReDetectionQueueEstimator(
    QueueEstimatorBase[ImageForRedetection, list[BodyDetection], BodyDetector, BodyDetectorSettings]
):
    """Body ReDetection estimator"""

    def assertData(self, image: ImageForRedetection):
        """assert image for redetection"""
        assertImageForReDetection(self._estimator._detector, image)

    async def _estimate(self, dataBatch: list[ImageForRedetection]) -> list[list[BodyDetection]]:
        """ReDetect bodies on images"""
        point = BodyReDetMonitoringPointBase(
            batchSize=len(dataBatch),
            imageWidth=dataBatch[0].image.rect.width,
            imageHeight=dataBatch[0].image.rect.height,
            bboxWidth=dataBatch[0].bBoxes[0].width,
            bboxHeight=dataBatch[0].bBoxes[0].height,
        )
        with EM.monitoring(point):
            res = await self._estimator.redetect(dataBatch, asyncEstimate=True)
            point.updateFromFirstImageResult(res)
            return res


class BodyReDetectionEstimationBroker(BodyBrokerDetectorMixin, REDETECTOR_GENERIC):
    """Body ReDetection broker"""

    # broker name
    name = "body_redetection"
