"""Module contains engine for estimation batching"""

import asyncio
import concurrent.futures
from datetime import datetime
from math import sqrt
from typing import List, Optional, Union

import FaceEngine as fe
from lunavl.sdk.descriptors.descriptors import BodyDescriptor, FaceDescriptor, FaceDescriptorBatch
from lunavl.sdk.detectors.base import ImageForRedetection
from lunavl.sdk.detectors.bodydetector import BodyDetection
from lunavl.sdk.detectors.facedetector import FaceDetection, Landmarks5, Landmarks68
from lunavl.sdk.detectors.humandetector import HumanDetection
from lunavl.sdk.errors.exceptions import LunaSDKException
from lunavl.sdk.estimators.body_estimators.body_attributes import BodyAttributes
from lunavl.sdk.estimators.body_estimators.bodywarper import BodyWarp, BodyWarpedImage
from lunavl.sdk.estimators.face_estimators.background import FaceDetectionBackground
from lunavl.sdk.estimators.face_estimators.basic_attributes import BasicAttributes
from lunavl.sdk.estimators.face_estimators.deepfake import Deepfake
from lunavl.sdk.estimators.face_estimators.emotions import Emotions
from lunavl.sdk.estimators.face_estimators.eyebrow_expressions import EyebrowExpression
from lunavl.sdk.estimators.face_estimators.eyes import (
    EyesEstimation,
    GazeDirection,
    WarpWithLandmarks,
    WarpWithLandmarks5,
)
from lunavl.sdk.estimators.face_estimators.face_occlusion import FaceOcclusion
from lunavl.sdk.estimators.face_estimators.facewarper import FaceWarpedImage
from lunavl.sdk.estimators.face_estimators.facial_hair import FacialHair
from lunavl.sdk.estimators.face_estimators.fisheye import Fisheye
from lunavl.sdk.estimators.face_estimators.glasses import Glasses
from lunavl.sdk.estimators.face_estimators.head_pose import HeadPose
from lunavl.sdk.estimators.face_estimators.headwear import Headwear
from lunavl.sdk.estimators.face_estimators.image_type import ImageColorType
from lunavl.sdk.estimators.face_estimators.livenessv1 import LivenessV1
from lunavl.sdk.estimators.face_estimators.mask import Mask
from lunavl.sdk.estimators.face_estimators.mouth_state import MouthStates
from lunavl.sdk.estimators.face_estimators.natural_light import FaceNaturalLight
from lunavl.sdk.estimators.face_estimators.portrait_style import PortraitStyle
from lunavl.sdk.estimators.face_estimators.red_eye import RedEyes
from lunavl.sdk.estimators.face_estimators.warp_quality import Quality
from lunavl.sdk.estimators.image_estimators.image_modification import ImageModification
from lunavl.sdk.estimators.image_estimators.orientation_mode import OrientationType
from lunavl.sdk.image_utils.image import VLImage

from .constants import FSDK_THREAD_POOL_SIZE, IMAGE_DECODER_POOL_SIZE, SYNC_IMAGE_DECODER_THRESHOLD
from .estimators_collection import EstimatorsCollection
from .face_estimators.deepfake_estimator import DeepfakeInputData
from .face_estimators.face_occlusion_estimator import FaceOcclusionInputData
from .face_estimators.liveness1_estimator import LivenessInputData
from ..errors import errors
from ..models.body_sample import AggregatedBodySample, BodySample
from ..models.face_sample import AggregatedFaceSample, FaceSample
from ..models.image import Image, ImageType, InputImage
from ..models.sample import Sample
from ..monitoring_utils.storages import IMAGE_LOAD_MONITORING as ILM
from ..utils.async_tools import gather
from ..utils.rotation import getImageRotationAngle, rotateBBoxes, rotateImage

# image decoder thread pool, 4 - imperial constant
DECODER_POOL = concurrent.futures.ThreadPoolExecutor(max_workers=IMAGE_DECODER_POOL_SIZE)

from FaceEngine import AsyncThreadPool

AsyncThreadPool.initialize(FSDK_THREAD_POOL_SIZE)


class ImageLoadPoint:
    """
    Image loading monitoring point

    Attributes:
        executionTime (float): loading time
        eventTime (float): event time timestamp
        format (str): image format
        width (int): image width
        size (int): image size (bytes)
        height (int): image height
        exif (bool): load exif or not
        autorotated (bool): use auto rotate or not
    """

    __slots__ = ("executionTime", "eventTime", "format", "width", "height", "size", "exif", "autorotated")
    name: str = "image_load"

    def __init__(self, eventTime: Optional[float] = None, executionTime: float = 0):
        self.executionTime = executionTime
        self.eventTime = eventTime if eventTime is not None else datetime.now().timestamp()
        self.format = None
        self.width = None
        self.height = None
        self.size = None
        self.exif = False
        self.autorotated = False

    def getFields(self) -> dict:
        """Get monitoring fields"""
        return {"execution_time": self.executionTime, "size": self.size, "height": self.height, "width": self.width}

    def getTags(self) -> dict:
        """Get monitoring tags"""
        return {"exif": self.exif, "autorotated": self.autorotated, "format": self.format}

    def update(self, img: Image):
        """Update  monitoring data from loaded image"""
        if img.sdkImage:
            self.width = img.sdkImage.rect.width
            self.height = img.sdkImage.rect.height
        if pilImage := img._pillowImage:
            self.format = pilImage.format
        if isinstance(img.origin.body, (bytes, bytearray)):
            self.size = len(img.origin.body)
        if img.orientation != OrientationType.NORMAL:
            self.autorotated = True
        if img.exif is not None:
            self.exif = True


def checkDetectionsRelated(bodyDetection: BodyDetection, faceDetection: FaceDetection) -> bool:
    """
    Check the body detection and face detection are related to the same person.

    Heuristic algorithm. If a сenter of the top side of the face bbox is near a center of the top side of the body bbox
    then we assume that it same person.

    Args:
        bodyDetection: body_estimators detection
        faceDetection: face detection

    Returns:

        True if body detection and face detection are related to the same person, otherwise False

    """
    xBodyCenter = bodyDetection.boundingBox.rect.x + bodyDetection.boundingBox.rect.width / 2
    yBodyCenter = bodyDetection.boundingBox.rect.y
    xFaceCenter = faceDetection.boundingBox.rect.x + faceDetection.boundingBox.rect.width / 2
    yFaceCenter = faceDetection.boundingBox.rect.y
    distance = sqrt((xBodyCenter - xFaceCenter) ** 2 + (yBodyCenter - yFaceCenter) ** 2)
    # 1.5 - heuristic
    if distance < faceDetection.boundingBox.rect.height / 1.5:
        return True
    return False


def loadSdkImage(image: Image, pilImage, filename: str = ""):
    """
    Load sdk image from pillow.
    Args:
        image: destination image
        pilImage: pillow image
        filename: optional filename
    """
    image.sdkImage = VLImage(pilImage, filename=filename)


def loadFaceWarp(image: Image, pilImage, filename: str = ""):
    """
    Load sdk face warp from pillow.
    Args:
        image: destination image
        pilImage: pillow image
        filename: optional filename
    """
    try:
        image.sdkImage = FaceWarpedImage(pilImage, filename=filename)
    except ValueError as e:
        image.error = errors.BadWarp.format(str(e))


def loadBodyWarp(image: Image, pilImage, filename: str = ""):
    """
    Load sdk face warp from pillow.
    Args:
        image: destination image
        pilImage: pillow image
        filename: optional filename"""
    try:
        image.sdkImage = BodyWarpedImage(pilImage, filename=filename)
    except ValueError as e:
        image.error = errors.BadWarp.format(str(e))


def prepareImage(inputImage: InputImage, useExifInfo: bool = False, estimateExif: bool = False) -> Image:
    """
    Prepare image. Load image to sdk structure.
    Args:
        inputImage: input image
        useExifInfo: use or not exif info for loading image (orientation tag)
        estimateExif: estimate exif or not

    Returns:
        loaded image, if error is occurred in loading process see image.error
    """
    image = Image(
        inputImage,
        faceBoxes=inputImage.faceBoxes,
        bodyBoxes=inputImage.bodyBoxes,
        trustedDetections=inputImage.trustedDetections,
    )
    if useExifInfo and inputImage.imageType == ImageType.IMAGE:
        image = Image.exifTranspose(image)
    elif estimateExif:
        image.extractExif()

    if image.sdkImage is not None:
        return image

    pilImage = image.getAsPillow()
    if not pilImage:
        return image
    try:
        if image.origin.imageType == ImageType.FACE_WARP:
            loadFaceWarp(image, pilImage, filename=inputImage.filename)
        elif image.origin.imageType == ImageType.BODY_WARP:
            loadBodyWarp(image, pilImage, filename=inputImage.filename)
        else:
            loadSdkImage(image, pilImage, filename=inputImage.filename)
    except LunaSDKException as e:
        image.error = e.error
    except Exception as exc:
        image.error = errors.ReadImageError(message=str(exc))
    return image


class SDKEngine:
    """
    Engine. Class realizes a batching of input data and following estimation.
    Attributes:
        estimators(EstimatorsCollection): all allowed estimators
        _loop (BaseLoop): current asyncio loop
    """

    def __init__(self, estimatorCollection: EstimatorsCollection, **kwargs):
        self._loop = asyncio.get_running_loop()
        self.estimators = estimatorCollection

    async def close(self, withEstimators: bool = True):
        """
        Close all engine
        Args:
            withEstimators: close or not estimators.
        """
        if withEstimators:
            await self.estimators.close()

    async def prepareImage(
        self, inputImage: InputImage, useExifInfo: bool, autoRotation: bool, estimateExif: bool
    ) -> Image:
        """
        Prepare image prepare for the next estimation.

        Prepare features:
            - load image to SDK format
            - rotate image base on exif if needed
            - rotate image based on orientation image estimation
        Args:
            inputImage: input image
            useExifInfo: rotate or not image based on a `orientation` exif tag
            autoRotation: detect image orientation and rotate or not
            estimateExif: estimate exif or not

        Returns:
            loaded image. If load is failed, output image will be contained an error
        """
        with ILM.monitoring(ImageLoadPoint()) as point:
            originImage = inputImage.body
            if isinstance(originImage, (bytes, bytearray)) and len(originImage) > SYNC_IMAGE_DECODER_THRESHOLD:
                # for big image decoding to a thread
                image = await self._loop.run_in_executor(
                    DECODER_POOL, prepareImage, inputImage, useExifInfo, estimateExif
                )
            else:
                image = prepareImage(inputImage, useExifInfo, estimateExif)
            point.update(image)
        # loaded to sdk image
        if image.error:
            return image
        if autoRotation and inputImage.imageType == ImageType.IMAGE:
            # check orientation
            orientationEstimator = self.estimators.getImageOrientationsEstimator()
            orientation = await orientationEstimator.estimate(image.sdkImage)
            image.orientation = orientation
            if orientation != OrientationType.NORMAL:
                rotationAngle = getImageRotationAngle(orientation)
                imageRect = image.sdkImage.rect
                if image.faceBoxes:
                    image.faceBoxes = rotateBBoxes(
                        image.faceBoxes, imageWidth=imageRect.width, imageHeight=imageRect.height, angle=rotationAngle
                    )
                if image.bodyBoxes:
                    image.bodyBoxes = rotateBBoxes(
                        image.bodyBoxes, imageWidth=imageRect.width, imageHeight=imageRect.height, angle=rotationAngle
                    )
                image.sdkImage = rotateImage(image.sdkImage, rotationAngle)
        return image

    async def extractExif(self, image: Image) -> dict:
        """
        Extract exif tags from image. The extraction is proceeded in separate thread.

        Args:
            image: image

        Returns:
            exif dict <tag: value>
        """
        exif = await self._loop.run_in_executor(DECODER_POOL, image.extractExif)
        return exif

    async def extractExifB(self, images: list[Image]) -> list[dict]:
        """
        Extract exif tags from batch images.
        Args:
            images: image

        Returns:
            exif dict <tag: value>
        """
        futures = (self.extractExif(image) for image in images)
        return await gather(futures)

    async def prepareImages(
        self, images: list[InputImage], useExifInfo: bool, autoRotation: bool, estimateExif: bool
    ) -> List[Image]:
        """
        Prepare batch images
        Args:
            images: images
            useExifInfo: use or no exif info for load image
            autoRotation: auto detect image orientation or not
            estimateExif: estimate exif or not

        Returns:
            list of prepared images
        """
        futures = (self.prepareImage(image, useExifInfo, autoRotation, estimateExif) for image in images)
        return await gather(futures)

    async def detectFaces(
        self,
        image: Image,
    ) -> Union[list[FaceDetection], list[Optional[FaceDetection]], None]:
        """
        Detect faces on image.

        Args:
            image: image

        Returns:
            None: if estimation is failed (see error in `image.error`)
            list[FaceDetection]: if detect faces
            list[Optional[FaceDetection]]: if redetect faces
        """
        try:
            if image.faceBoxes:
                if image.trustedDetections:
                    detections = []
                    sdkImage = image.sdkImage
                    for det in image.faceBoxes:

                        coreDet = fe.Detection(det.coreRectF, 1)
                        coreFace = fe.Face(sdkImage.coreImage, coreDet)
                        detection = FaceDetection(coreFace, sdkImage)
                        detections.append(detection)
                else:
                    img = ImageForRedetection(image.sdkImage, image.faceBoxes)
                    detections = await self.estimators.getFaceReDetector().estimate(img)
            else:
                detections = await self.estimators.getFaceDetector().estimate(image.sdkImage)
        except LunaSDKException as e:
            # expected redetect errors with incorrect bboxes
            image.error = errors.SDKError(e.error)
            return None
        detections = list(filter(None, detections))
        return detections

    async def detectHumans(
        self,
        image: Image,
    ) -> list[HumanDetection]:
        """
        Detect faces on image.

        Args:
            image: image

        Returns:
            None: if estimation is failed (see error in `image.error`)
            list[FaceDetection]: if detect faces
            list[Optional[FaceDetection]]: if redetect faces
        """
        detections = await self.estimators.getHumanDetector().estimate(image.sdkImage)
        return detections

    async def detectFacesB(self, images) -> list[Union[list[FaceDetection], list[Optional[FaceDetection]], None]]:
        """Batch detect. Return result by every image in corresponding order"""

        futures = (self.detectFaces(image) for image in images)
        return await gather(futures)

    async def warpFaceSample(self, faceSample: FaceSample) -> tuple[FaceWarpedImage, Landmarks5]:
        """
        Estimate face warp.

        Args:
            faceSample: face sample

        Returns:
            <face warp, transformed Landmarks5>
        """
        faceWarp, transformedLandmarks5 = await self.estimators.getFaceWarpEstimator().estimate(faceSample.detection)
        faceSample.warp, faceSample.transformedLandmarks5 = faceWarp.warpedImage, transformedLandmarks5
        return faceWarp, transformedLandmarks5

    async def warpBodySample(self, bodySample: BodySample) -> BodyWarp:
        """
        Estimate body warp.

        Args:
            bodySample: body sample

        Returns:
            body warp
        """
        warp = await self.estimators.getBodyWarpEstimator().estimate(bodySample.detection)
        bodySample.warp = warp.warpedImage
        return warp

    async def warpFaceSampleB(self, faceSamples: list[FaceSample]) -> list[FaceWarpedImage]:
        """Estimate face warps (batch)"""
        return await gather((self.warpFaceSample(sample) for sample in faceSamples))

    async def warpBodySampleB(self, bodySamples: list[BodySample]) -> list[BodyWarpedImage]:
        """Estimate body warps (batch)"""
        return await gather((self.warpBodySample(sample) for sample in bodySamples))

    async def detectBodies(self, image: Image) -> Union[list[BodyDetection], list[Optional[BodyDetection]], None]:
        """
        Detect bodies on image.

        Args:
            image: image

        Returns:
            None: if estimation is failed (see error in `image.error`)
            list[BodyDetection]: if detect bodies
            list[Optional[BodyDetection]]: if redetect bodies
        """
        try:
            if image.bodyBoxes:
                if image.trustedDetections:
                    detections = []
                    sdkImage = image.sdkImage
                    for det in image.bodyBoxes:

                        coreDet = fe.Detection(det.coreRectF, 1)
                        coreBody = fe.Human()
                        coreBody.detection = coreDet
                        coreBody.img = sdkImage.coreImage
                        detection = BodyDetection(coreBody, sdkImage)
                        detections.append(detection)
                else:
                    img = ImageForRedetection(image.sdkImage, image.bodyBoxes)
                    detections = await self.estimators.getBodyReDetector().estimate(img)
            else:
                detections = await self.estimators.getBodyDetector().estimate(image.sdkImage)
        except LunaSDKException as e:
            # expected redetect errors with incorrect bboxes
            image.error = errors.SDKError(e.error)
            return None
        detections = list(filter(None, detections))
        return detections

    async def detectBodiesB(self, images: list[Image]) -> list[list[BodyDetection]]:
        """Batch bodies detect. Return result by every image in corresponding order"""
        futures = (self.detectBodies(image) for image in images)
        return await gather(futures)

    def _merge(self, faceDetections: List[FaceDetection], bodyDetections: List[BodyDetection]) -> list[Sample]:
        """
        Merge bodies and faces detections one same image. If face and body detection
        belongs to one human function will aggregate them to one sample.

        Args:
            faceDetections: list of face detection
            bodyDetections:  list of bodies detection

        Returns:
            list of common samples. Sample contains face and body detection if these detection belongs to one human
        """

        if not faceDetections and not bodyDetections:
            return []
        elif faceDetections and not bodyDetections:
            samples = [Sample(face=FaceSample(detection=faceDetection)) for faceDetection in faceDetections]
            return samples
        elif bodyDetections and not faceDetections:
            samples = [Sample(body=BodySample(detection=bodyDetection)) for bodyDetection in bodyDetections]
            return samples

        samples = []
        for faceDetection in faceDetections:
            faceSample = FaceSample(faceDetection)
            sample = Sample(face=faceSample)
            for idx, bodyDetection in enumerate(bodyDetections):
                if checkDetectionsRelated(bodyDetection, faceDetection):
                    sample.body = BodySample(detection=bodyDetection)
                    bodyDetections.pop(idx)
                    break
            samples.append(sample)
        for bodyDetection in bodyDetections:
            sample = Sample()
            sample.body = BodySample(detection=bodyDetection)
            samples.append(sample)
        return samples

    async def estimateImageOrientation(self, image: Image) -> OrientationType:
        """
        Estimate image orientation
        Args:
            image: image

        Returns:
            image orientation
        """
        image.orientation = await self.estimators.getImageOrientationsEstimator().estimate(image)
        return image.orientation

    async def estimateImageOrientationB(self, images: list[Image]) -> list[OrientationType]:
        """Batch image orientation estimation."""
        return await gather(self.estimateImageOrientation(image) for image in images)

    async def estimatePeopleCount(self, image: Image):
        """
        Estimate people count
        Args:
            image: image
        Retruns:
            people count
        """
        image.peopleCount = await self.estimators.getPeopleCountEstimator().estimate(image.sdkImage)
        return image.peopleCount

    async def estimatePeopleCountB(self, images: List[Image]):
        """Batch people count estimation"""
        return await gather(self.estimatePeopleCount(image) for image in images)

    async def estimateImageModification(self, image: Image) -> ImageModification:
        """
        Estimate image modification
        Args:
            image: image
        Retruns:
            image modification
        """
        image.imageModification = await self.estimators.getImageModificationEstimator().estimate(image.sdkImage)
        return image.imageModification

    async def estimateImageModificationB(self, images: List[Image]) -> list[ImageModification]:
        """Batch  image modification estimation"""
        return await gather(self.estimateImageModification(image) for image in images)

    async def estimateDynamicRange(self, sample: FaceSample) -> float:
        """
        Estimate dynamic range
        Args:
            sample: face sample

        Returns:
            estimated dynamic range
        """
        dynamicRange = await self.estimators.getDynamicRangeEstimator().estimate(sample.detection)
        sample.dynamicRange = dynamicRange
        return dynamicRange

    async def estimateDynamicRangeB(self, samples: list[FaceSample]) -> list[float]:
        """Batch dynamic range estimation."""
        return await gather(self.estimateDynamicRange(sample) for sample in samples)

    async def estimateAGSRange(self, sample: FaceSample) -> float:
        """
        Estimate ags
        Args:
            sample: face sample

        Returns:
            estimated ags
        """
        ags = await self.estimators.getAGSEstimator().estimate(sample.detection)
        sample.ags = ags
        return ags

    async def estimateAGSRangeB(self, samples: list[FaceSample]) -> list[float]:
        """Batch ags estimation."""
        return await gather(self.estimateAGSRange(sample) for sample in samples)

    async def estimatePortraitStyle(self, sample: FaceSample) -> PortraitStyle:
        """Perform estimation, assign result to the samples appropriate field."""
        portraitStyle = await self.estimators.getPortraitStyleEstimator().estimate(sample.detection)
        sample.portraitStyle = portraitStyle
        return portraitStyle

    async def estimatePortraitStyleB(self, samples: list[FaceSample]) -> list[PortraitStyle]:
        """Batch estimation."""
        return await gather(self.estimatePortraitStyle(sample) for sample in samples)

    async def detectSamples(self, image: Image, detectFaces: bool = True, detectBodies: bool = False) -> list[Sample]:
        """
        Detect samples on image

        Args:
            image: image
            detectFaces: detect or not faces
            detectBodies: detect or not bodies

        Returns:
            List of Common samples (body + face)
        """

        faceDetections = []
        bodyDetections = []
        if detectFaces and detectBodies:
            if not self.estimators.humanDetector or image.faceBoxes or image.bodyBoxes:
                faceDetections, bodyDetections = await asyncio.gather(
                    self.detectFaces(image), self.detectBodies(image)
                )  # get landmarks for merging
            else:
                humanDetections = await self.detectHumans(image)
                samples = []
                for humanDetection in humanDetections:
                    if face := humanDetection.face:
                        faceSample = FaceSample(face)
                    else:
                        faceSample = None
                    if body := humanDetection.body:
                        bodySample = BodySample(body)
                    else:
                        bodySample = None
                    sample = Sample(face=faceSample, body=bodySample)
                    samples.append(sample)
                image.samples = samples
                return samples

        elif detectBodies:
            bodyDetections = await self.detectBodies(image)
        elif detectFaces:
            faceDetections = await self.detectFaces(image)

        faceDetections = faceDetections if faceDetections is not None else []
        bodyDetections = bodyDetections if bodyDetections is not None else []
        samples = self._merge(faceDetections, bodyDetections)
        image.samples = samples
        return samples

    async def detectSamplesB(
        self,
        images: List[Image],
        detectFaces: bool = True,
        detectBodies: bool = False,
    ) -> list[list[Sample]]:
        """Batch samples detect."""
        if not images:
            return []
        return await gather(
            self.detectSamples(image, detectFaces=detectFaces, detectBodies=detectBodies) for image in images
        )

    async def estimateLivenessV1(
        self,
        sample: FaceSample,
        scoreThreshold: Optional[float] = None,
    ) -> LivenessV1:
        """
        Estimate livenessV1
        Args:
            sample: face sample
            scoreThreshold: score threshold for prediction

        Returns:
            estimated liveness
        """
        data = LivenessInputData(sample.detection, scoreThreshold=scoreThreshold)
        livenessV1 = await self.estimators.getLivenessv1Estimator().estimate(data)
        sample.livenessV1 = livenessV1
        return livenessV1

    async def estimateLivenessV1B(
        self,
        samples: list[FaceSample],
        scoreThreshold: Optional[float] = None,
    ) -> list[LivenessV1]:
        """Batch livenessV1 estimation."""
        return await gather((self.estimateLivenessV1(sample, scoreThreshold=scoreThreshold) for sample in samples))

    async def estimateDeepfakeV2(
        self,
        sample: FaceSample,
        scoreThreshold: Optional[float] = None,
    ) -> Deepfake:
        """
        Estimate deepfakeV2
        Args:
            sample: face sample
            scoreThreshold: score threshold for prediction

        Returns:
            estimated deepfake
        """
        data = DeepfakeInputData(sample.detection, scoreThreshold=scoreThreshold)
        deepfake = await self.estimators.getDeepfakeV2Estimator().estimate(data)
        sample.deepfake = deepfake
        return deepfake

    async def estimateDeepfakeV2B(
        self,
        samples: list[FaceSample],
        scoreThreshold: Optional[float] = None,
    ) -> list[Deepfake]:
        """Batch deepfakeV2 estimation."""
        return await gather((self.estimateDeepfakeV2(sample, scoreThreshold=scoreThreshold) for sample in samples))

    async def estimateDeepfakeV1(
        self,
        sample: FaceSample,
        scoreThreshold: Optional[float] = None,
    ) -> Deepfake:
        """
        Estimate deepfakeV1
        Args:
            sample: face sample
            scoreThreshold: score threshold for prediction

        Returns:
            estimated deepfake
        """
        data = DeepfakeInputData(sample.detection, scoreThreshold=scoreThreshold)
        deepfake = await self.estimators.getDeepfakeV1Estimator().estimate(data)
        sample.deepfake = deepfake
        return deepfake

    async def estimateDeepfakeV1B(
        self,
        samples: list[FaceSample],
        scoreThreshold: Optional[float] = None,
    ) -> list[Deepfake]:
        """Batch deepfakeV1 estimation."""
        return await gather((self.estimateDeepfakeV1(sample, scoreThreshold=scoreThreshold) for sample in samples))

    async def estimateHeadPose(self, sample: FaceSample) -> HeadPose:
        """
        Estimate head pose
        Args:
            sample: face sample

        Returns:
            estimated head pose
        """
        headPose = await self.estimators.getHeadPoseEstimator().estimate(sample.detection)
        sample.headPose = headPose
        return headPose

    async def estimateHeadPoseB(self, samples: list[FaceSample]) -> list[HeadPose]:
        """Batch livenessV1 estimation."""
        return await gather((self.estimateHeadPose(sample) for sample in samples))

    async def estimateMask(self, sample: FaceSample) -> Mask:
        """
        Estimate mask
        Args:
            sample: face sample

        Returns:
            estimated head pose
        """
        mask = await self.estimators.getMaskEstimator().estimate(sample.warp)
        sample.mask = mask
        return mask

    async def estimateMaskB(self, samples: list[FaceSample]) -> list[Mask]:
        """Batch mask estimation."""
        return await gather((self.estimateMask(sample) for sample in samples))

    async def estimateGaze(self, sample: FaceSample) -> GazeDirection:
        """
        Estimate gaze direction
        Args:
            sample: face sample

        Returns:
            estimated gaze direction
        """
        gaze = await self.estimators.getGazeDirectionEstimator().estimate(
            WarpWithLandmarks5(sample.warp, sample.transformedLandmarks5)
        )
        sample.gaze = gaze
        return gaze

    async def estimateLandmarks5(self, sample: FaceSample) -> Landmarks5:
        """
        Estimate landmarks5
        Args:
            sample: face sample

        Returns:
            estimated landmarks5
        """
        if sample.detection.landmarks5:
            # redetect datected landmarks 5
            return sample.detection.landmarks5
        landmarks5 = await self.estimators.getLandmarks5Estimator().estimate(sample.detection)
        sample.detection.landmarks5 = landmarks5
        return landmarks5

    async def estimateLandmarks5B(self, samples: list[FaceSample]) -> list[Landmarks5]:
        """Batch landmarks5 estimation."""
        return await gather(self.estimateLandmarks5(sample) for sample in samples)

    async def estimateLandmarks68(self, sample: FaceSample) -> Landmarks68:
        """
        Estimate landmarks68
        Args:
            sample: face sample

        Returns:
            estimated landmarks68
        """
        landmarks68 = await self.estimators.getLandmarks68Estimator().estimate(sample.detection)
        sample.detection.landmarks68 = landmarks68
        return landmarks68

    async def estimateLandmarks68B(self, samples: list[FaceSample]) -> list[Landmarks68]:
        """Batch landmarks68 estimation."""
        return await gather(self.estimateLandmarks68(sample) for sample in samples)

    async def estimateGazeB(self, samples: list[FaceSample]) -> list[GazeDirection]:
        """Batch gaze direction estimation."""
        return await gather(self.estimateGaze(sample) for sample in samples)

    async def estimateFaceDescriptor(self, sample: FaceSample, version: int) -> FaceDescriptor:
        """
        Estimate face descriptor

        Args:
            sample: face sample
            version: descriptor version

        Returns:
            estimated descriptor
        Raises:
            errors.UnknownDescriptor: if version  is not valid or unsupported
        """
        estimator = self.estimators.getFaceDescriptorEstimator(version)
        descriptor = await estimator.estimate(sample.warp)
        sample.descriptor = descriptor
        return descriptor

    async def estimateFaceDescriptorB(self, samples: list[FaceSample], version: int) -> list[FaceDescriptor]:
        """Batch face descriptor estimation."""
        if not samples:
            return []
        return await gather(self.estimateFaceDescriptor(sample, version) for sample in samples)

    async def estimateAggregatedFaceDescriptor(
        self, sample: AggregatedFaceSample, version: int
    ) -> tuple[FaceDescriptorBatch, FaceDescriptor]:
        """
        Estimated aggregated face descriptor.

        Args:
            sample: aggregated face sample with non empty samples
            version: descriptor version

        Returns:
            tuple <descriptor for every sample, aggregated descriptor>
        Raises:
            errors.UnknownDescriptor: if version  is not valid or unsupported
        """
        estimator = self.estimators.getAggreagtedFaceDescriptorEstimator(version)
        descriptors, aggregatedDescriptor = await estimator.estimate([sample.warp for sample in sample.samples])
        sample.descriptor = aggregatedDescriptor
        for idx, sample in enumerate(sample.samples):
            sample.descriptor = descriptors[idx]
        return descriptors, aggregatedDescriptor

    async def estimateBasicAttributes(self, sample: FaceSample) -> BasicAttributes:
        """
        Estimate basic attributes
        Args:
            sample: face sample

        Returns:
            estimated basic attributes
        """
        basicAttributes = await self.estimators.getBasicAttributesEstimator().estimate(sample.warp)
        sample.basicAttributes = basicAttributes
        return basicAttributes

    async def estimateBasicAttributesB(self, samples: list[FaceSample]) -> list[BasicAttributes]:
        """Batch basic attributes estimation."""
        return await gather(self.estimateBasicAttributes(sample) for sample in samples)

    async def estimateAggregatedBasicAttributes(
        self,
        sample: AggregatedFaceSample,
    ) -> tuple[list[BasicAttributes], BasicAttributes]:
        """
        Estimate aggregated basic attributes
        Args:
            sample: aggregated face sample with non empty samples

        Returns:
            tuple <basic attributes for every sample, aggregated basic attributes>
        """
        attributesBatch, aggregatedAttributes = await self.estimators.getAggreagtedBasicAttributesEstimator().estimate(
            [sample.warp for sample in sample.samples]
        )
        sample.basicAttributes = aggregatedAttributes
        for idx, sample in enumerate(sample.samples):
            sample.basicAttributes = attributesBatch[idx]
        return attributesBatch, aggregatedAttributes

    async def estimateEmotions(self, sample: FaceSample) -> Emotions:
        """
        Estimate emotions
        Args:
            sample: face sample

        Returns:
            estimated emotions
        """
        emotions = await self.estimators.getEmotionsEstimator().estimate(sample.warp)
        sample.emotions = emotions
        return emotions

    async def estimateEmotionsB(self, samples: list[FaceSample]) -> list[Emotions]:
        """Batch emotions estimation."""
        return await gather(self.estimateEmotions(sample) for sample in samples)

    async def estimateFacialHair(self, sample: FaceSample) -> FacialHair:
        """
        Estimate facial hair
        Args:
            sample: face sample

        Returns:
            estimated facial hair
        """
        facialHair = await self.estimators.getFacialHairEstimator().estimate(sample.warp)
        sample.facialHair = facialHair
        return facialHair

    async def estimateFacialHairB(self, samples: list[FaceSample]) -> list[FacialHair]:
        """Batch facial hair estimation."""
        return await gather(self.estimateFacialHair(sample) for sample in samples)

    async def estimateGlasses(self, sample: FaceSample) -> Glasses:
        """
        Estimate glasses
        Args:
            sample: face sample

        Returns:
            estimated glasses
        """
        glasses = await self.estimators.getGlassesEstimator().estimate(sample.warp)
        sample.glasses = glasses
        return glasses

    async def estimateGlassesB(self, samples: list[FaceSample]) -> list[Glasses]:
        """Batch glasses estimation."""
        return await gather(self.estimateGlasses(sample) for sample in samples)

    async def estimateFaceOcclusion(
        self,
        sample: FaceSample,
        occlusionThreshold: float | None = None,
        hairThreshold: float | None = None,
        foreheadThreshold: float | None = None,
        eyeThreshold: float | None = None,
        noseThreshold: float | None = None,
        mouthThreshold: float | None = None,
        lowerFaceThreshold: float | None = None,
    ) -> FaceOcclusion:
        """
        Estimate face cocclusion
        Args:
            sample: face sample
            occlusionThreshold: overall occlusion threshold
            hairThreshold: hair occlusion threshold
            foreheadThreshold: forehead occlusion threshold
            eyeThreshold: eye occlusion threshold
            noseThreshold: nose occlusion threshold
            mouthThreshold: mouse occlusion threshold
            lowerFaceThreshold: lower face occlusion threshold

        Returns:
            estimated face occlusion
        """
        data = FaceOcclusionInputData(
            WarpWithLandmarks(sample.warp, sample.transformedLandmarks5),
            overallThreshold=occlusionThreshold,
            hairThreshold=hairThreshold,
            foreheadThreshold=foreheadThreshold,
            leftEyeThreshold=eyeThreshold,
            rightEyeThreshold=eyeThreshold,
            noseThreshold=noseThreshold,
            mouthThreshold=mouthThreshold,
            lowerFaceThreshold=lowerFaceThreshold,
        )
        faceOcclusion = await self.estimators.getFaceOcclusionEstimator().estimate(data)
        sample.faceOcclusion = faceOcclusion
        return faceOcclusion

    async def estimateFaceOcclusionB(self, samples: list[FaceSample], **thresholds) -> list[FaceOcclusion]:
        """Batch face occlusion estimation."""
        return await gather(self.estimateFaceOcclusion(sample, **thresholds) for sample in samples)

    async def estimateEyes(self, sample: FaceSample) -> EyesEstimation:
        """
        Estimate eyes
        Args:
            sample: face sample

        Returns:
            estimated eyes
        """
        eyes = await self.estimators.getEyesEstimator().estimate(
            WarpWithLandmarks(sample.warp, sample.transformedLandmarks5)
        )
        sample.eyes = eyes
        return eyes

    async def estimateEyesB(self, samples: list[FaceSample]) -> list[EyesEstimation]:
        """Batch eyes estimation."""
        return await gather(self.estimateEyes(sample) for sample in samples)

    async def estimateBodyDescriptor(self, sample: BodySample, version: int) -> BodyDescriptor:
        """
        Estimate body descriptor

        Args:
            sample: body sample
            version: descriptor version

        Returns:
            estimated descriptor
        Raises:
            errors.UnknownDescriptor: if version  is not valid or unsupported
        """
        estimator = self.estimators.getBodyDescriptorEstimator(version)
        descriptor = await estimator.estimate(sample.warp)
        sample.descriptor = descriptor
        return descriptor

    async def estimateAggregatedBodyDescriptor(
        self, sample: AggregatedBodySample, version
    ) -> tuple[list[BodyDescriptor], BodyDescriptor]:
        """
        Estimated aggregated body descriptor.

        Args:
            sample: aggregated body sample with non empty samples
            version: descriptor version

        Returns:
            tuple <descriptor for every sample, aggregated descriptor>
        Raises:
            errors.UnknownDescriptor: if version  is not valid or unsupported
        """
        estimator = self.estimators.getAggreagtedBodyDescriptorEstimator(version)
        descriptors, aggregatedDescriptor = await estimator.estimate([sample.warp for sample in sample.samples])
        sample.descriptor = aggregatedDescriptor
        for idx, sample in enumerate(sample.samples):
            sample.descriptor = descriptors[idx]
        return descriptors, aggregatedDescriptor

    async def estimateBodyDescriptorB(self, samples: list[BodySample], version: int):
        """Batch body descriptors estimation."""
        return await gather(self.estimateBodyDescriptor(sample, version) for sample in samples)

    async def estimateMouth(self, sample: FaceSample) -> MouthStates:
        """
        Estimate mouth states
        Args:
            sample: face sample

        Returns:
            estimated mouth states
        """
        mouthState = await self.estimators.getMouthStateEstimator().estimate(sample.warp)
        sample.mouthState = mouthState
        return mouthState

    async def estimateMouthB(self, samples: list[FaceSample]) -> list[MouthStates]:
        """Batch mouth states estimation."""
        return await gather(self.estimateMouth(sample) for sample in samples)

    async def estimateFaceWarpQuality(self, sample: FaceSample) -> Quality:
        """
        Estimate face warp quality
        Args:
            sample: face sample

        Returns:
            estimated face warp quality
        """
        warpQuality = await self.estimators.getFaceWarpQualityEstimator().estimate(sample.warp)
        sample.warpQuality = warpQuality
        return warpQuality

    async def estimateFaceWarpQualityB(self, samples: list[FaceSample]) -> list[Quality]:
        """Batch face warp quality estimation."""
        return await gather(self.estimateFaceWarpQuality(sample) for sample in samples)

    async def estimateHeadwear(self, sample: FaceSample) -> Headwear:
        """
        Estimate headwear
        Args:
            sample: face sample

        Returns:
            estimated headwear
        """
        headwear = await self.estimators.getHeadwearEstimator().estimate(sample.warp)
        sample.headwear = headwear
        return headwear

    async def estimateHeadwearB(self, samples: list[FaceSample]) -> list[Headwear]:
        """Batch headwear estimation."""
        return await gather(self.estimateHeadwear(sample) for sample in samples)

    async def estimateFaceNaturalLight(self, sample: FaceSample) -> FaceNaturalLight:
        """
        Estimate face natural light
        Args:
            sample: face sample

        Returns:
            estimated face natural light
        """
        naturalLight = await self.estimators.getNaturalLightEstimator().estimate(sample.warp)
        sample.naturalLight = naturalLight
        return naturalLight

    async def estimateFaceNaturalLightB(self, samples: list[FaceSample]) -> list[FaceNaturalLight]:
        """Batch face natural light estimation."""
        return await gather(self.estimateFaceNaturalLight(sample) for sample in samples)

    async def estimateImageColorType(self, sample: FaceSample) -> ImageColorType:
        """
        Estimate image color type
        Args:
            sample: face sample

        Returns:
            estimated  image color type
        """
        imageColorType = await self.estimators.getImageColorTypeEstimator().estimate(sample.warp)
        sample.imageColorType = imageColorType
        return imageColorType

    async def estimateImageColorTypeB(self, samples: list[FaceSample]) -> list[ImageColorType]:
        """Batch image color tType estimation."""
        return await gather(self.estimateImageColorType(sample) for sample in samples)

    async def estimateEyebrowExpression(self, sample: FaceSample) -> EyebrowExpression:
        """
        Estimate eyebrow expression
        Args:
            sample: face sample

        Returns:
            estimated eyebrow expression
        """
        eyebrowExpression = await self.estimators.getEyebrowExpressionEstimator().estimate(sample.warp)
        sample.eyebrowExpression = eyebrowExpression
        return eyebrowExpression

    async def estimateEyebrowExpressionB(self, samples: list[FaceSample]) -> list[EyebrowExpression]:
        """Batch eyebrow expression estimation."""
        return await gather(self.estimateEyebrowExpression(sample) for sample in samples)

    async def estimateRedEyes(self, sample: FaceSample) -> RedEyes:
        """
        Estimate red-eyes
        Args:
            sample: face sample

        Returns:
            estimated red-eyes
        """
        redEyes = await self.estimators.getRedEyesEstimator().estimate(
            WarpWithLandmarks5(sample.warp, sample.transformedLandmarks5)
        )
        sample.redEyes = redEyes
        return redEyes

    async def estimateRedEyesB(self, samples: list[FaceSample]) -> list[RedEyes]:
        """Batch red-eyes estimation."""
        return await gather(self.estimateRedEyes(sample) for sample in samples)

    async def estimateFishEye(self, sample: FaceSample) -> Fisheye:
        """
        Estimate fisheye
        Args:
            sample: face sample

        Returns:
            estimated fisheye
        """
        fisheye = await self.estimators.getFisheyeEstimator().estimate(sample.warp)
        sample.fisheye = fisheye
        return fisheye

    async def estimateFishEyeB(self, samples: list[FaceSample]) -> list[Fisheye]:
        """Batch Fisheye estimation."""
        return await gather((self.estimateFishEye(sample) for sample in samples))

    async def estimateFaceDetectionBackground(self, sample: FaceSample) -> FaceDetectionBackground:
        """
        Estimate face detection background
        Args:
            sample: face sample

        Returns:
            estimated face detection background
        """
        faceDetectionBackground = await self.estimators.getFaceDetectionBackgroundEstimator().estimate(sample.detection)
        sample.detectionBackground = faceDetectionBackground
        return faceDetectionBackground

    async def estimateFaceDetectionBackgroundB(self, samples: list[FaceSample]) -> list[FaceDetectionBackground]:
        """Batch face detection background estimation."""
        return await gather((self.estimateFaceDetectionBackground(sample) for sample in samples))

    async def estimateBodyAttributes(self, sample: BodySample) -> BodyAttributes:
        """
        Estimate body attributes
        Args:
            sample: body sample

        Returns:
            estimated body attributes
        """
        bodyAttributes = await self.estimators.getBodyAttributesEstimator().estimate(sample.warp)
        sample.attributes = bodyAttributes
        return bodyAttributes

    async def estimateBodyAttributesB(self, samples: list[BodySample]) -> list[BodyAttributes]:
        """Batch body attributes estimation."""
        return await gather((self.estimateBodyAttributes(sample) for sample in samples))
