from contextlib import contextmanager
from copy import deepcopy
from typing import ContextManager, Literal, Optional, Type, TypeVar

from attr import define, field
from vlutils.descriptors.data import DescriptorsEnum

from .body_estimators.body_attributes_estimator import BodyAttributesEstimationBroker, BodyAttributesQueueEstimator
from .body_estimators.body_descriptor_estimator import (
    AggregatedBodyDescriptorEstimationBroker,
    AggregatedBodyDescriptorQueueEstimator,
    BodyDescriptorEstimationBroker,
    BodyDescriptorQueueEstimator,
    BodyDescriptorSettings,
)
from .body_estimators.body_detector import (
    BodyDetectionEstimationBroker,
    BodyDetectionQueueEstimator,
    BodyReDetectionEstimationBroker,
    BodyReDetectionQueueEstimator,
)
from .body_estimators.body_warp_estimator import BodyWarpEstimationBroker, BodyWarpQueueEstimator
from .face_estimators.ags_estimator import AGSEstimationBroker, AGSQueueEstimator
from .face_estimators.basic_attributes_estimator import (
    AggregatedBasicAttributesEstimationBroker,
    AggregatedBasicAttributesQueueEstimator,
    BasicAttributesEstimationBroker,
    BasicAttributesQueueEstimator,
)
from .face_estimators.deepfake_estimator import (
    DeepfakeV1QueueEstimationBroker,
    DeepfakeV1QueueEstimator,
    DeepfakeV2QueueEstimationBroker,
    DeepfakeV2QueueEstimator,
)
from .face_estimators.dynamic_range_estimator import DynamicRangeEstimationBroker, DynamicRangeQueueEstimator
from .face_estimators.emotions_estimator import EmotionsEstimationBroker, EmotionsQueueEstimator
from .face_estimators.eyebrow_expression_estimator import (
    EyebrowExpressionEstimationBroker,
    EyebrowExpressionQueueEstimator,
)
from .face_estimators.eyes_estimator import EyesEstimationBroker, EyesQueueEstimator
from .face_estimators.face_descriptor_estimator import (
    AggregatedFaceDescriptorEstimationBroker,
    AggregatedFaceDescriptorQueueEstimator,
    FaceDescriptorEstimationBroker,
    FaceDescriptorQueueEstimator,
    FaceDescriptorSettings,
)
from .face_estimators.face_detection_background_estimator import (
    FaceDetectionBackgroundEstimationBroker,
    FaceDetectionBackgroundQueueEstimator,
)
from .face_estimators.face_detector import (
    FaceDetectionEstimationBroker,
    FaceDetectionQueueEstimator,
    FaceReDetectionEstimationBroker,
    FaceReDetectionQueueEstimator,
)
from .face_estimators.face_natural_light_estimator import (
    FaceNaturalLightEstimationBroker,
    FaceNaturalLightQueueEstimator,
)
from .face_estimators.face_occlusion_estimator import FaceOcclusionEstimationBroker, FaceOcclusionQueueEstimator
from .face_estimators.face_warp_estimator import FaceWarpEstimationBroker, FaceWarpFaceWarpQueueEstimator
from .face_estimators.face_warp_quality_estimator import FaceWarpQualityEstimationBroker, FaceWarpQualityQueueEstimator
from .face_estimators.facial_hair_estimator import FacialHairEstimationBroker, FacialHairQueueEstimator
from .face_estimators.fisheye_estimator import FisheyeEstimationBroker, FisheyeQueueEstimator
from .face_estimators.gaze_direction_estimator import GazeDirectionQueueEstimator, GazeEstimationBroker
from .face_estimators.glasses_estimator import GlassesEstimationBroker, GlassesQueueEstimator
from .face_estimators.head_pose_estimator import HeadPoseEstimationBroker, HeadPoseQueueEstimator
from .face_estimators.headwear import HeadwearEstimationBroker, HeadwearQueueEstimator
from .face_estimators.image_color_type import ImageColorTypeEstimationBroker, ImageColorTypeQueueEstimator
from .face_estimators.landmarks_estimator import (
    Landmarks5EstimationBroker,
    Landmarks5QueueEstimator,
    Landmarks68EstimationBroker,
    Landmarks68QueueEstimator,
)
from .face_estimators.liveness1_estimator import Livenessv1QueueEstimationBroker, LivenessV1QueueEstimator
from .face_estimators.mask_estimator import MaskEstimationBroker, MaskQueueEstimator
from .face_estimators.mouth_estimator import MouthStateEstimationBroker, MouthStatesQueueEstimator
from .face_estimators.portrait_style_estimator import PortraitStyleEstimationBroker, PortraitStyleQueueEstimator
from .face_estimators.red_eyes_estimator import RedEyesEstimationBroker, RedEyesQueueEstimator
from .image_estimators.image_modification import ImageModificationEstimationBroker, ImageModificationQueueEstimator
from .image_estimators.orientation_mode_estimator import ImageOrientatioEstimationBroker, ImageOrientationQueueEstimator
from .image_estimators.people_count_estimator import PeopleCountEstimationBroker, PeopleCountQueueEstimator
from .others_estimators.human_detector import HumanDetectionEstimationBroker, HumanDetectionQueueEstimator
from .settings import (
    BodyDescriptorBrokerSettings,
    BrokerSettingsProtocol,
    FaceDescriptorBrokerSettings,
    SDKLoopSettings,
)
from .te.human_tracker import TEBroker, TEWorker
from .worker import EstimationBrokerBase, QueueEstimatorBase
from ..enums import Estimators
from ..errors import errors
from ..errors.errors import EstimatorInitializationError, EstimatorWasNotInitializedError
from ..utils.logger import logger

# broker generic type
BROKER_GENERIC = TypeVar("BROKER_GENERIC", bound=EstimationBrokerBase)
# estimator generic type
ESTIMATOR_GENERIC = TypeVar("ESTIMATOR_GENERIC", bound=QueueEstimatorBase)


@contextmanager
def initializationContext(
    brokerClass: Type[BROKER_GENERIC], settings: BrokerSettingsProtocol, name: str | None = None
) -> ContextManager:
    """
    Estimator initialization context manager. A centralization logging and error catching.

    Args:
        brokerClass: broker class
        settings: broker settings
        name: estimation name
    Raises:
        EstimatorInitializationError: if  initialization is failed
    """
    try:
        yield
    except Exception as e:
        if name is None:
            name = brokerClass.name
        logger.exception(f"failed to init estimator '{brokerClass.__name__}', settings: {settings}")
        raise EstimatorInitializationError.format(name) from e
    else:
        logger.info(f"broker {brokerClass.__name__} has been initialized with {settings=}")


def _initBroker(
    brokerClass: Type[BROKER_GENERIC],
    estimatorClass: Type[ESTIMATOR_GENERIC],
    settings: BrokerSettingsProtocol,
    name: str | None = None,
) -> BROKER_GENERIC:
    """
    Init broker helper.

    Args:
        brokerClass: broker class
        estimatorClass: broker estimator class
        settings: broker settings
        name: estimation name

    Returns:
        broker instance
    Raises:
        EstimatorInitializationError: if initialization failed
    """
    with initializationContext(brokerClass, settings, name=name):
        return brokerClass(
            workerClass=estimatorClass,
            optimalBatchSize=settings.optimalBatchSize,
            estimatorSettings=settings.estimatorSettings,
            workerCount=settings.workerCount,
        )


def assertDescriptorVersion(version: int):
    """
    Assert that descriptor version is supported
    Args:
        version: descriptor version

    Raises:
        errors.UnknownDescriptor: if descriptor version is unknown
    """
    for descriptor in DescriptorsEnum:
        if descriptor.value.version == version:
            return
    raise errors.UnknownDescriptor.format(version)


def assertBroker(broker: Optional[BROKER_GENERIC], name: str) -> BROKER_GENERIC:
    """
    Assert that broker is not None
    Args:
        broker: broker for check
        name: broker name
    Returns:
        broker
    Raises:
        EstimatorWasNotInitializedError: if broker is None
    """
    if broker is None:
        raise EstimatorWasNotInitializedError.format(name)
    return broker


@define
class EstimatorsCollection:
    """
    Estimator collection
    """

    # list of estimators for an initialization and a subsequent work
    targets: list[Estimators]
    # settings
    settings: SDKLoopSettings

    """estimators on image """

    # image orientation estimator
    imageOrientationEstimator: Optional[ImageOrientatioEstimationBroker] = field(init=False)
    # image modification estimator
    imageModificationEstimator: Optional[ImageModificationEstimationBroker] = field(init=False)
    # people count estimator
    peopleCountEstimator: Optional[PeopleCountEstimationBroker] = field(init=False)
    # dynamic range estimator
    dynamicRangeEstimator: Optional[DynamicRangeEstimationBroker] = field(init=False)
    # human detection estimator
    humanDetector: Optional[HumanDetectionEstimationBroker] = field(init=False)

    """estimators on face detection """

    # face detection estimator (without landmarks68 and landmarks5)
    faceDetector: Optional[FaceDetectionEstimationBroker] = field(init=False)
    # face ReDetection estimator (without landmarks68 and landmarks5)
    faceReDetector: Optional[FaceReDetectionEstimationBroker] = field(init=False)
    # gaze direction estimator
    gazeDirectionEstimator: Optional[GazeEstimationBroker] = field(init=False)
    # head pose estimator
    headPoseEstimator: Optional[HeadPoseEstimationBroker] = field(init=False)
    # face warp estimator
    faceWarpEstimator: Optional[FaceWarpEstimationBroker] = field(init=False)
    # face occlusion estimator
    faceOcclusionEstimator: Optional[FaceOcclusionEstimationBroker] = field(init=False)
    # face occlusion estimator
    facialHairEstimator: Optional[FacialHairEstimationBroker] = field(init=False)
    # eyes estimator
    eyesEstimator: Optional[EyesEstimationBroker] = field(init=False)
    # livenessV1 estimator
    livenessV1Estimator: Optional[Livenessv1QueueEstimationBroker] = field(init=False)
    # deepfake estimator, estimation mode 1
    deepfakeV1Estimator: Optional[DeepfakeV1QueueEstimationBroker] = field(init=False)
    # deepfake estimator, estimation mode 2
    deepfakeV2Estimator: Optional[DeepfakeV2QueueEstimationBroker] = field(init=False)
    # red-eyes estimator
    redEyesEstimator: Optional[RedEyesEstimationBroker] = field(init=False)
    # fisheye estimator
    fisheyeEstimator: Optional[FisheyeEstimationBroker] = field(init=False)
    portraitStyleEstimator: Optional[PortraitStyleEstimationBroker] = field(init=False)
    # face detection background estimator
    faceDetectionBackgroundEstimator: Optional[FaceDetectionBackgroundEstimationBroker] = field(init=False)
    # face detection ags estimator
    agsEstimator: Optional[AGSEstimationBroker] = field(init=False)

    """estimators on face warp"""

    # face descriptor estimators, map descriptor version to a corresponding broker
    faceDescriptorEstimatorMap: Optional[dict[int, FaceDescriptorEstimationBroker]] = field(init=False)
    # aggregated face descriptor estimators, map descriptor version to a corresponding broker
    aggregatedFaceDescriptorEstimatorMap: Optional[dict[int, AggregatedFaceDescriptorEstimationBroker]] = field(
        init=False
    )
    # emotions estimator
    emotionsEstimator: Optional[EmotionsEstimationBroker] = field(init=False)
    # face warp quality estimator
    faceWarpQualityEstimator: Optional[FaceWarpQualityEstimationBroker] = field(init=False)
    # mouth state estimator
    mouthStateEstimator: Optional[MouthStateEstimationBroker] = field(init=False)
    # mask estimator
    maskEstimator: Optional[MaskEstimationBroker] = field(init=False)
    # glasses estimator
    glassesEstimator: Optional[GlassesEstimationBroker] = field(init=False)
    # face natural light estimator
    faceNaturalLightEstimator: Optional[FaceNaturalLightEstimationBroker] = field(init=False)
    # headwear estimator
    headwearEstimator: Optional[HeadwearEstimationBroker] = field(init=False)
    # eyebrow expression estimator
    eyebrowExpressionEstimator: Optional[EyebrowExpressionEstimationBroker] = field(init=False)
    # image color type estimator
    imageColorTypeEstimator: Optional[ImageColorTypeEstimationBroker] = field(init=False)
    # basic attributes estimator
    basicAttributesEstimator: Optional[BasicAttributesEstimationBroker] = field(init=False)
    # aggregated basic attributes estimator
    aggregatedBasicAttributesEstimator: Optional[AggregatedBasicAttributesEstimationBroker] = field(init=False)
    # face landmarks5 estimator
    landmarks5Estimator: Optional[Landmarks5EstimationBroker] = field(init=False)
    # face landmarks68 estimator
    landmarks68Estimator: Optional[Landmarks68EstimationBroker] = field(init=False)

    """estimators on body detection """

    # body detection estimator
    bodyDetector: Optional[BodyDetectionEstimationBroker] = field(init=False)
    # body redetection estimator
    bodyReDetector: Optional[BodyReDetectionEstimationBroker] = field(init=False)
    # body estimator
    bodyWarpEstimator: Optional[BodyWarpEstimationBroker] = field(init=False)
    # body attributes estimator
    bodyAttributesEstimator: Optional[BodyAttributesEstimationBroker] = field(init=False)

    """estimators on body warp"""

    # body descriptor estimators, map descriptor version to a corresponding broker
    bodyDescriptorEstimatorMap: Optional[dict[int, BodyDescriptorEstimationBroker]] = field(init=False)
    # aggregated body descriptor estimators, map descriptor version to a corresponding broker
    aggregatedBodyDescriptorEstimatorMap: Optional[dict[int, AggregatedBodyDescriptorEstimationBroker]] = field(
        init=False
    )

    #  Map tracker name (name determines with which settings tracker was created) to tracker instance
    humanTrackers: Optional[dict[str, TEBroker]] = field(init=False)

    def __attrs_post_init__(self):
        """Post initialization estimators based on targets and settings"""

        self.initImageOrientationEstimator()
        self.initImageModificationEstimator()
        self.initPeopleCountEstimator()
        self.initDynamicRangeEstimator()
        self.initHumanDetectionEstimator()

        self.initFaceDetectionEstimator()
        self.initFaceDescriptorEstimator()
        self.initHeadPoseEstimator()
        self.initFaceWarpEstimator()
        self.initGazeDirectionEstimator()
        self.initFaceOcclusionEstimator()
        self.initFacialHairEstimator()
        self.initEyesAttributesEstimator()
        self.initEmotionsEstimator()
        self.initMouthStateEstimator()
        self.initFaceWarpQualityEstimator()
        self.initBasicAttributesEstimator()
        self.initMaskEstimator()
        self.initGlassesEstimator()
        self.initLivenessV1Estimator()
        self.initDeepfakeEstimator()
        self.initEyebrowExpressionEstimator()
        self.initFisheyeEstimator()
        self.initPortraitStyleEstimator()
        self.initRedEyesEstimator()
        self.initFaceDetectionBackgroundEstimator()
        self.initImageColorTypeEstimator()
        self.initHeadwearExpressionEstimator()
        self.initFaceNaturalLightExpressionEstimator()
        self.initLandmarks5Estimator()
        self.initLandmarks68Estimator()
        self.initAGSEstimator()

        self.initBodyDetectionEstimator()
        self.initBodyWarpEstimator()
        self.initBodyDescriptorEstimator()
        self.initBodyAttributesEstimator()

        self.initHumanTracksEstimator()

    def initFaceDetectionEstimator(self):
        """Init a face detector brokers if needed"""
        if Estimators.faceDetection in self.targets:
            settings = self.settings.faceDetectionBrokerSettings
            self.faceDetector = _initBroker(FaceDetectionEstimationBroker, FaceDetectionQueueEstimator, settings)
            self.faceReDetector = _initBroker(FaceReDetectionEstimationBroker, FaceReDetectionQueueEstimator, settings)
        else:
            self.faceDetector = None
            self.faceReDetector = None

    def initGazeDirectionEstimator(self):
        """Init a gaze direction broker if needed"""
        if Estimators.gazeDirection in self.targets:
            self.gazeDirectionEstimator = _initBroker(
                GazeEstimationBroker, GazeDirectionQueueEstimator, self.settings.gazeDirectionBrokerSettings
            )
        else:
            self.gazeDirectionEstimator = None

    def initHeadPoseEstimator(self):
        """Init a head pose broker if needed"""
        if Estimators.headPose in self.targets:
            self.headPoseEstimator = _initBroker(
                HeadPoseEstimationBroker, HeadPoseQueueEstimator, self.settings.faceDetectionBrokerSettings
            )
        else:
            self.headPoseEstimator = None

    def initImageOrientationEstimator(self):
        """Init an image orientation broker if needed"""
        if Estimators.imageOrientation in self.targets:
            self.imageOrientationEstimator = _initBroker(
                ImageOrientatioEstimationBroker,
                ImageOrientationQueueEstimator,
                self.settings.imageOrientationBrokerSettings,
            )
        else:
            self.imageOrientationEstimator = None

    def initImageModificationEstimator(self):
        """Init an image modification broker if needed"""
        if Estimators.imageModification in self.targets:
            self.imageModificationEstimator = _initBroker(
                ImageModificationEstimationBroker,
                ImageModificationQueueEstimator,
                self.settings.imageModificationBrokerSettings,
            )
        else:
            self.imageModificationEstimator = None

    def initPeopleCountEstimator(self):
        """Init a people count estimator"""
        if Estimators.peopleCount in self.targets:
            self.peopleCountEstimator = _initBroker(
                PeopleCountEstimationBroker, PeopleCountQueueEstimator, self.settings.peopleCountBrokerSettings
            )
        else:
            self.peopleCountEstimator = None

    def initDynamicRangeEstimator(self):
        """Init dynamic range estimator"""
        if Estimators.dynamicRange in self.targets:
            self.dynamicRangeEstimator = _initBroker(
                DynamicRangeEstimationBroker, DynamicRangeQueueEstimator, self.settings.dynamicRangeBrokerSettings
            )
        else:
            self.dynamicRangeEstimator = None

    def initAGSEstimator(self):
        """Init ags estimator"""
        if Estimators.ags in self.targets:
            self.agsEstimator = _initBroker(AGSEstimationBroker, AGSQueueEstimator, self.settings.agsBrokerSettings)
        else:
            self.agsEstimator = None

    def initFaceWarpEstimator(self):
        """Init a face warp broker if needed"""
        if Estimators.faceWarp in self.targets:
            self.faceWarpEstimator = _initBroker(
                FaceWarpEstimationBroker, FaceWarpFaceWarpQueueEstimator, self.settings.faceWarpBrokerSettings
            )
        else:
            self.faceWarpEstimator = None

    def initFaceOcclusionEstimator(self):
        """Init a face occlusion broker if needed"""
        if Estimators.faceOcclusion in self.targets:
            self.faceOcclusionEstimator = _initBroker(
                FaceOcclusionEstimationBroker, FaceOcclusionQueueEstimator, self.settings.faceOcclusionBrokerSettings
            )
        else:
            self.faceOcclusionEstimator = None

    def initFacialHairEstimator(self):
        """Init facial hair broker if needed"""
        if Estimators.facialHair in self.targets:
            self.facialHairEstimator = _initBroker(
                FacialHairEstimationBroker, FacialHairQueueEstimator, self.settings.facialHairBrokerSettings
            )
        else:
            self.facialHairEstimator = None

    def initEyesAttributesEstimator(self):
        """Init an eyes attributes broker if needed"""
        if Estimators.eyes in self.targets:
            self.eyesEstimator = _initBroker(EyesEstimationBroker, EyesQueueEstimator, self.settings.eyesBrokerSettings)
        else:
            self.eyesEstimator = None

    def initEmotionsEstimator(self):
        """Init an emotions broker if needed"""
        if Estimators.emotions in self.targets:
            self.emotionsEstimator = _initBroker(
                EmotionsEstimationBroker, EmotionsQueueEstimator, self.settings.emotionsBrokerSettings
            )
        else:
            self.emotionsEstimator = None

    def initMouthStateEstimator(self):
        """Init a mouth state broker if needed"""
        if Estimators.mouthState in self.targets:
            self.mouthStateEstimator = _initBroker(
                MouthStateEstimationBroker, MouthStatesQueueEstimator, self.settings.mouthStateBrokerSettings
            )
        else:
            self.mouthStateEstimator = None

    def initFaceWarpQualityEstimator(self):
        """Init a face warp quality broker if needed"""
        if Estimators.quality in self.targets:
            self.faceWarpQualityEstimator = _initBroker(
                FaceWarpQualityEstimationBroker,
                FaceWarpQualityQueueEstimator,
                self.settings.faceWarpQualityBrokerSettings,
            )
        else:
            self.faceWarpQualityEstimator = None

    def initBasicAttributesEstimator(self):
        """Init a basic attributes broker if needed"""
        if Estimators.basicAttributes in self.targets:
            self.basicAttributesEstimator = _initBroker(
                BasicAttributesEstimationBroker,
                BasicAttributesQueueEstimator,
                self.settings.basicAttributesBrokerSettings,
            )
            self.aggregatedBasicAttributesEstimator = _initBroker(
                AggregatedBasicAttributesEstimationBroker,
                AggregatedBasicAttributesQueueEstimator,
                self.settings.basicAttributesBrokerSettings,
            )
        else:
            self.basicAttributesEstimator = None
            self.aggregatedBasicAttributesEstimator = None

    def initMaskEstimator(self):
        """Init a mask broker if needed"""
        if Estimators.mask in self.targets:
            self.maskEstimator = _initBroker(MaskEstimationBroker, MaskQueueEstimator, self.settings.maskBrokerSettings)
        else:
            self.maskEstimator = None

    def initGlassesEstimator(self):
        """Init a glasses broker if needed"""
        if Estimators.glasses in self.targets:
            self.glassesEstimator = _initBroker(
                GlassesEstimationBroker, GlassesQueueEstimator, self.settings.glassesBrokerSettings
            )
        else:
            self.glassesEstimator = None

    def initLivenessV1Estimator(self):
        """Init a livenessV1 broker if needed"""
        if Estimators.livenessV1 in self.targets:
            self.livenessV1Estimator = _initBroker(
                Livenessv1QueueEstimationBroker, LivenessV1QueueEstimator, self.settings.livenessV1BrokerSettings
            )
        else:
            self.livenessV1Estimator = None

    def initDeepfakeEstimator(self):
        """Init a deepfake broker if needed"""

        if Estimators.deepfake in self.targets:
            self.deepfakeV1Estimator = _initBroker(
                DeepfakeV1QueueEstimationBroker, DeepfakeV1QueueEstimator, self.settings.deepfakeBrokerSettings
            )
            self.deepfakeV2Estimator = _initBroker(
                DeepfakeV2QueueEstimationBroker, DeepfakeV2QueueEstimator, self.settings.deepfakeBrokerSettings
            )
        else:
            self.deepfakeV1Estimator = None
            self.deepfakeV2Estimator = None

    def initImageColorTypeEstimator(self):
        """Init an image color broker if needed"""
        if Estimators.imageColorType in self.targets:
            self.imageColorTypeEstimator = _initBroker(
                ImageColorTypeEstimationBroker, ImageColorTypeQueueEstimator, self.settings.imageColorTypeBrokerSettings
            )
        else:
            self.imageColorTypeEstimator = None

    def initFisheyeEstimator(self):
        """Init a fisheye broker if needed"""
        if Estimators.fisheye in self.targets:
            self.fisheyeEstimator = _initBroker(
                FisheyeEstimationBroker, FisheyeQueueEstimator, self.settings.fisheyeBrokerSettings
            )
        else:
            self.fisheyeEstimator = None

    def initPortraitStyleEstimator(self):
        """Init a portrait style broker if needed"""
        if Estimators.portraitStyle in self.targets:
            self.portraitStyleEstimator = _initBroker(
                PortraitStyleEstimationBroker, PortraitStyleQueueEstimator, self.settings.portraitStyleBrokerSettings
            )
        else:
            self.portraitStyleEstimator = None

    def initRedEyesEstimator(self):
        """Init a red-eyes broker if needed"""
        if Estimators.redEyes in self.targets:
            self.redEyesEstimator = _initBroker(
                RedEyesEstimationBroker, RedEyesQueueEstimator, self.settings.redEyesBrokerSettings
            )
        else:
            self.redEyesEstimator = None

    def initEyebrowExpressionEstimator(self):
        """Init an eyebrow expression broker if needed"""
        if Estimators.eyebrowExpression in self.targets:
            self.eyebrowExpressionEstimator = _initBroker(
                EyebrowExpressionEstimationBroker,
                EyebrowExpressionQueueEstimator,
                self.settings.eyebrowExpressionBrokerSettings,
            )
        else:
            self.eyebrowExpressionEstimator = None

    def initHeadwearExpressionEstimator(self):
        """Init a headwear broker if needed"""
        if Estimators.headwear in self.targets:
            self.headwearEstimator = _initBroker(
                HeadwearEstimationBroker, HeadwearQueueEstimator, self.settings.headwearBrokerSettings
            )
        else:
            self.headwearEstimator = None

    def initFaceDetectionBackgroundEstimator(self):
        """Init a face detection background broker if needed"""
        if Estimators.faceDetectionBackground in self.targets:
            self.faceDetectionBackgroundEstimator = _initBroker(
                FaceDetectionBackgroundEstimationBroker,
                FaceDetectionBackgroundQueueEstimator,
                self.settings.faceDetectionBackgroundBrokerSettings,
            )
        else:
            self.faceDetectionBackgroundEstimator = None

    def initFaceNaturalLightExpressionEstimator(self):
        """Init a face natural light broker if needed"""
        if Estimators.faceNaturalLight in self.targets:
            self.faceNaturalLightEstimator = _initBroker(
                FaceNaturalLightEstimationBroker,
                FaceNaturalLightQueueEstimator,
                self.settings.faceNaturalLightBrokerSettings,
            )
        else:
            self.faceNaturalLightEstimator = None

    def initLandmarks5Estimator(self):
        """Init a landmarks5 broker if needed"""
        if Estimators.faceLandmarks5 in self.targets:
            self.landmarks5Estimator = _initBroker(
                Landmarks5EstimationBroker,
                Landmarks5QueueEstimator,
                self.settings.landmarks5BrokerSettings,
            )
        else:
            self.landmarks5Estimator = None

    def initLandmarks68Estimator(self):
        """Init a landmarks68 broker if needed"""
        if Estimators.faceLandmarks68 in self.targets:
            self.landmarks68Estimator = _initBroker(
                Landmarks68EstimationBroker,
                Landmarks68QueueEstimator,
                self.settings.landmarks68BrokerSettings,
            )
        else:
            self.landmarks68Estimator = None

    def createFaceDescriptorEstimators(self, version: int):
        """Create face descriptor broker by descriptor version if needed"""
        lo = self.settings.faceDescriptorBrokerSettings.estimatorSettings.launchOptions
        extractorSettings = FaceDescriptorSettings(descriptorVersion=version, launchOptions=lo)
        settings = FaceDescriptorBrokerSettings(
            optimalBatchSize=self.settings.faceDescriptorBrokerSettings.optimalBatchSize,
            workerCount=self.settings.faceDescriptorBrokerSettings.workerCount,
            estimatorSettings=extractorSettings,
        )
        name = f"face_descriptor_{version}"
        estimator = _initBroker(FaceDescriptorEstimationBroker, FaceDescriptorQueueEstimator, settings, name=name)
        name = f"aggregated_face_descriptor_{version}"
        with initializationContext(AggregatedFaceDescriptorEstimationBroker, settings, name=name):
            aggregatedEstimator = AggregatedFaceDescriptorEstimationBroker(
                estimatorSettings=settings.estimatorSettings,
                workerCount=self.settings.faceDescriptorBrokerSettings.workerCount,
                optimalBatchSize=1,
                workerClass=AggregatedFaceDescriptorQueueEstimator,
            )
        return estimator, aggregatedEstimator

    def initFaceDescriptorEstimator(self, descriptorVersion: Optional[int] = None):
        """
        Init a face descriptor broker if needed

        Args:
            descriptorVersion: descriptor version, if None will be used default from settings
        """
        if Estimators.faceDescriptor in self.targets:
            if descriptorVersion is None:
                descriptorVersion = self.settings.faceDescriptorBrokerSettings.estimatorSettings.descriptorVersion
            estimator, aggregatedEstimator = self.createFaceDescriptorEstimators(descriptorVersion)
            if not hasattr(self, "faceDescriptorEstimatorMap") or self.faceDescriptorEstimatorMap is None:
                self.faceDescriptorEstimatorMap = {}
                self.aggregatedFaceDescriptorEstimatorMap = {}
            self.faceDescriptorEstimatorMap[descriptorVersion] = estimator
            self.aggregatedFaceDescriptorEstimatorMap[descriptorVersion] = aggregatedEstimator
        else:
            self.aggregatedFaceDescriptorEstimatorMap = None
            self.faceDescriptorEstimatorMap = None

    def initBodyDetectionEstimator(self):
        """Init a body detector brokers if needed"""
        if Estimators.bodyDetection in self.targets:
            self.bodyDetector = _initBroker(
                BodyDetectionEstimationBroker,
                BodyDetectionQueueEstimator,
                self.settings.bodyDetectionBrokerSettings,
            )
            self.bodyReDetector = _initBroker(
                BodyReDetectionEstimationBroker,
                BodyReDetectionQueueEstimator,
                self.settings.bodyDetectionBrokerSettings,
            )
        else:
            self.bodyDetector = None
            self.bodyReDetector = None

    def initBodyWarpEstimator(self):
        """Init a body warp broker if needed"""
        if Estimators.bodyWarp in self.targets:
            self.bodyWarpEstimator = _initBroker(
                BodyWarpEstimationBroker, BodyWarpQueueEstimator, self.settings.bodyWarpBrokerSettings
            )
        else:
            self.bodyWarpEstimator = None

    def initBodyAttributesEstimator(self):
        """Init a body warp broker if needed"""
        if Estimators.bodyAttributes in self.targets:
            self.bodyAttributesEstimator = _initBroker(
                BodyAttributesEstimationBroker, BodyAttributesQueueEstimator, self.settings.bodyAttributesBrokerSettings
            )
        else:
            self.bodyAttributesEstimator = None

    def createBodyDescriptorEstimators(self, version: int):
        """Create body descriptor broker by descriptor version if needed"""
        lo = self.settings.bodyDescriptorBrokerSettings.estimatorSettings.launchOptions
        extractorSettings = BodyDescriptorSettings(descriptorVersion=version, launchOptions=lo)
        settings = BodyDescriptorBrokerSettings(
            optimalBatchSize=self.settings.bodyDescriptorBrokerSettings.optimalBatchSize,
            workerCount=self.settings.bodyDescriptorBrokerSettings.workerCount,
            estimatorSettings=extractorSettings,
        )
        name = f"body_descriptor_{version}"
        estimator = _initBroker(BodyDescriptorEstimationBroker, BodyDescriptorQueueEstimator, settings, name=name)
        name = f"aggregated_body_descriptor_{version}"
        with initializationContext(AggregatedFaceDescriptorEstimationBroker, settings, name=name):
            aggregatedEstimator = AggregatedBodyDescriptorEstimationBroker(
                estimatorSettings=settings.estimatorSettings,
                workerCount=self.settings.bodyDescriptorBrokerSettings.workerCount,
                optimalBatchSize=1,
                workerClass=AggregatedBodyDescriptorQueueEstimator,
            )
        return estimator, aggregatedEstimator

    def initBodyDescriptorEstimator(self, descriptorVersion: Optional[int] = None):
        """
        Init a body descriptor broker if needed

        Args:
            descriptorVersion: descriptor version, if None will be used default from settings
        """
        if Estimators.bodyDescriptor in self.targets:
            if descriptorVersion is None:
                descriptorVersion = self.settings.bodyDescriptorBrokerSettings.estimatorSettings.descriptorVersion
            estimator, aggregatedEstimator = self.createBodyDescriptorEstimators(descriptorVersion)
            if not hasattr(self, "bodyDescriptorEstimatorMap") or self.bodyDescriptorEstimatorMap is None:
                self.bodyDescriptorEstimatorMap = {}
                self.aggregatedBodyDescriptorEstimatorMap = {}
            self.bodyDescriptorEstimatorMap[descriptorVersion] = estimator
            self.aggregatedBodyDescriptorEstimatorMap[descriptorVersion] = aggregatedEstimator
        else:
            self.bodyDescriptorEstimatorMap = None
            self.aggregatedBodyDescriptorEstimatorMap = None

    def initHumanDetectionEstimator(self):
        """Init a human detector brokers if needed"""
        if Estimators.humanDetection in self.targets:
            self.humanDetector = _initBroker(
                HumanDetectionEstimationBroker,
                HumanDetectionQueueEstimator,
                self.settings.humanDetectionBrokerSettings,
            )
        else:
            self.humanDetector = None

    def createHumanTracksBroker(self, name, detectFace: int, detectBody: int, useBodyReId: int) -> TEBroker:
        """
        Create Human Tracker with corresponding settings.

        Args:
            name: human tracker alias name (for logging)
            detectFace: detect or not face
            detectBody: detect or not body
            useBodyReId: use body reid for body tracking

        Returns:
            tracker
        """
        with initializationContext(TEBroker, self.settings.humanTracksBrokerSettings, name=name):
            estimatorSettings = deepcopy(self.settings.humanTracksBrokerSettings.estimatorSettings)
            estimatorSettings.te.detectFace = detectFace
            estimatorSettings.te.detectBody = detectBody
            estimatorSettings.te.useBodyReId = useBodyReId
            humanTracker = TEBroker(
                estimatorSettings=estimatorSettings,
                optimalBatchSize=self.settings.humanTracksBrokerSettings.optimalBatchSize,
                workerCount=1,
                workerClass=TEWorker,
            )
            return humanTracker

    def initHumanTracksEstimator(self):
        """
        Init a default human tracker if needed.

        Load default tracker settings from trackengine.conf, create HumanTrack, put
        to tracker container by key with default settings.
        """
        detectFace = self.settings.humanTracksBrokerSettings.estimatorSettings.te.detectFace

        detectBody = self.settings.humanTracksBrokerSettings.estimatorSettings.te.detectBody
        useBodyReId = self.settings.humanTracksBrokerSettings.estimatorSettings.te.useBodyReId
        if Estimators.humanTracker in self.targets:
            name = f"human_tracker_face_{detectFace}_body_{detectBody}_reid_{useBodyReId}"
            self.humanTrackers = {
                name: self.createHumanTracksBroker(
                    name, detectFace=detectFace, detectBody=detectBody, useBodyReId=useBodyReId
                )
            }
        else:
            self.humanTrackers = None

    async def close(self):
        """
        Stop all estimation and delete all brokers/estimators.
        """

        async def closeEstimator(estimatorName: str):
            estimator: EstimationBrokerBase = getattr(self, estimatorName)
            if estimator:
                await estimator.stop()
            setattr(self, estimatorName, None)

        estimators = [
            "faceDetector",
            "faceReDetector",
            "faceWarpEstimator",
            "gazeDirectionEstimator",
            "eyesEstimator",
            "faceOcclusionEstimator",
            # "facialHairEstimator",  # LUNA-7305
            "headPoseEstimator",
            "mouthStateEstimator",
            "faceWarpQualityEstimator",
            "emotionsEstimator",
            "maskEstimator",
            "glassesEstimator",
            "livenessV1Estimator",
            "deepfakeV1Estimator",
            "deepfakeV2Estimator",
            "basicAttributesEstimator",
            "aggregatedBasicAttributesEstimator",
            "fisheyeEstimator",
            "portraitStyleEstimator",
            "redEyesEstimator",
            "eyebrowExpressionEstimator",
            "imageColorTypeEstimator",
            "headwearEstimator",
            "faceNaturalLightEstimator",
            "faceDetectionBackgroundEstimator",
            "dynamicRangeEstimator",
            "landmarks5Estimator",
            "landmarks68Estimator",
            "bodyDetector",
            "bodyReDetector",
            "bodyWarpEstimator",
            "bodyAttributesEstimator",
            "imageOrientationEstimator",
            "imageModificationEstimator",
            "humanDetector",
            "peopleCountEstimator",
            "agsEstimator",
        ]

        for name in estimators:
            await closeEstimator(name)

        async def stopDescriptorsBrokers(mapVersionToBroker):
            for version, descriptorEstimator in mapVersionToBroker.items():
                await descriptorEstimator.stop()

        if self.faceDescriptorEstimatorMap:
            await stopDescriptorsBrokers(self.faceDescriptorEstimatorMap)
            self.faceDescriptorEstimatorMap = None

        if self.aggregatedFaceDescriptorEstimatorMap:
            await stopDescriptorsBrokers(self.aggregatedFaceDescriptorEstimatorMap)
            self.aggregatedFaceDescriptorEstimatorMap = None

        if self.bodyDescriptorEstimatorMap:
            await stopDescriptorsBrokers(self.bodyDescriptorEstimatorMap)
            self.bodyDescriptorEstimatorMap = None

        if self.aggregatedBodyDescriptorEstimatorMap:
            await stopDescriptorsBrokers(self.aggregatedBodyDescriptorEstimatorMap)
            self.bodyDescriptorEstimatorMap = None
        if self.humanTrackers:
            for tracker in self.humanTrackers.values():
                await tracker.stop()

    def getImageOrientationsEstimator(self) -> ImageOrientatioEstimationBroker:
        """Get image image orientation estimator"""
        return assertBroker(self.imageOrientationEstimator, "image_orientation")

    def getImageModificationEstimator(self) -> ImageModificationEstimationBroker:
        """Get image image modification estimator"""
        return assertBroker(self.imageModificationEstimator, "image_modification")

    def getPeopleCountEstimator(self) -> PeopleCountEstimationBroker:
        """Get people count estimator"""
        return assertBroker(self.peopleCountEstimator, "people_count")

    def getDynamicRangeEstimator(self) -> DynamicRangeEstimationBroker:
        """Get dynamic range estimator"""
        return assertBroker(self.dynamicRangeEstimator, "dynamic_range")

    def getAGSEstimator(self) -> AGSEstimationBroker:
        """Get ags range estimator"""
        return assertBroker(self.agsEstimator, "ags")

    def getFaceDetector(self) -> FaceDetectionEstimationBroker:
        """Get face detector (without 68 landmarks)"""
        return assertBroker(self.faceDetector, "face_detection")

    def getFaceReDetector(self) -> FaceReDetectionEstimationBroker:
        """Get face redetector (without 68 landmarks)"""
        return assertBroker(self.faceReDetector, "face_detection")

    def getFaceWarpEstimator(self) -> FaceWarpEstimationBroker:
        """Get face warp estimator"""
        return assertBroker(self.faceWarpEstimator, "face_warp")

    def getBasicAttributesEstimator(self) -> BasicAttributesEstimationBroker:
        """Get basic attributes estimator"""
        return assertBroker(self.basicAttributesEstimator, "basic_attributes")

    def getAggreagtedBasicAttributesEstimator(self) -> AggregatedBasicAttributesEstimationBroker:
        """Get aggregated basic attributes estimator"""
        return assertBroker(self.aggregatedBasicAttributesEstimator, "aggregated_basic_attributes")

    def getEmotionsEstimator(self) -> EmotionsEstimationBroker:
        """Get emotions estimator"""
        return assertBroker(self.emotionsEstimator, "emotions")

    def getFaceWarpQualityEstimator(self) -> FaceWarpQualityEstimationBroker:
        """Get face warp quality estimator"""
        return assertBroker(self.faceWarpQualityEstimator, "face_warp_quality")

    def getMouthStateEstimator(self) -> MouthStateEstimationBroker:
        """Get mouth state estimator"""
        return assertBroker(self.mouthStateEstimator, "mouth_state")

    def getMaskEstimator(self) -> MaskEstimationBroker:
        """Get mask estimator"""
        return assertBroker(self.maskEstimator, "mask")

    def getLivenessv1Estimator(self) -> Livenessv1QueueEstimationBroker:
        """Get livenessV1 estimator"""
        return assertBroker(self.livenessV1Estimator, "livenessV1")

    def getDeepfakeV1Estimator(self) -> DeepfakeV1QueueEstimationBroker:
        """Get deepfake estimator"""
        return assertBroker(self.deepfakeV1Estimator, "deepfakeV1")

    def getDeepfakeV2Estimator(self) -> DeepfakeV2QueueEstimationBroker:
        """Get deepfake estimator"""
        return assertBroker(self.deepfakeV2Estimator, "deepfakeV2")

    def getHeadPoseEstimator(self) -> HeadPoseEstimationBroker:
        """Get head pose estimator"""
        return assertBroker(self.headPoseEstimator, "head_pose")

    def getGazeDirectionEstimator(self) -> GazeEstimationBroker:
        """Get gaze direction estimator"""
        return assertBroker(self.gazeDirectionEstimator, "gaze_direction")

    def getEyesEstimator(self) -> EyesEstimationBroker:
        """Get eyes estimator"""
        return assertBroker(self.eyesEstimator, "eyes")

    def getFaceOcclusionEstimator(self) -> FaceOcclusionEstimationBroker:
        """Get face occlusion estimator"""
        return assertBroker(self.faceOcclusionEstimator, "face_occlusion")

    def getFacialHairEstimator(self) -> FacialHairEstimationBroker:
        """Get facial hair estimator"""
        return assertBroker(self.facialHairEstimator, "facial_hair")

    def getGlassesEstimator(self) -> GlassesEstimationBroker:
        """Get glasses estimator"""
        return assertBroker(self.glassesEstimator, "glasses")

    def getFaceDescriptorEstimator(self, version: Optional[int] = None) -> FaceDescriptorEstimationBroker:
        """Get face descriptor estimator"""
        if self.faceDescriptorEstimatorMap is None:
            raise EstimatorWasNotInitializedError.format("face_descriptor")
        if version is None:
            version = self.settings.faceDescriptorBrokerSettings.estimatorSettings.descriptorVersion
        estimator = self.faceDescriptorEstimatorMap.get(version)
        if not estimator:
            assertDescriptorVersion(version)
            self.initFaceDescriptorEstimator(descriptorVersion=version)
            estimator = self.faceDescriptorEstimatorMap[version]
        return estimator

    def getAggreagtedFaceDescriptorEstimator(
        self, version: Optional[int] = None
    ) -> AggregatedFaceDescriptorEstimationBroker:
        """Get aggregated face descriptor estimator"""
        if self.aggregatedFaceDescriptorEstimatorMap is None:
            raise EstimatorWasNotInitializedError.format("aggregated_face_descriptor")
        if version is None:
            version = self.settings.faceDescriptorBrokerSettings.estimatorSettings.descriptorVersion
        estimator = self.aggregatedFaceDescriptorEstimatorMap.get(version)
        if not estimator:
            assertDescriptorVersion(version)
            self.initFaceDescriptorEstimator(descriptorVersion=version)
            estimator = self.aggregatedFaceDescriptorEstimatorMap[version]
        return estimator

    def getBodyDetector(self) -> BodyDetectionEstimationBroker:
        """Get body redetector"""
        return assertBroker(self.bodyDetector, "body_detection")

    def getBodyReDetector(self) -> BodyReDetectionEstimationBroker:
        """Get body redetector"""
        return assertBroker(self.bodyReDetector, "body_redetection")

    def getBodyWarpEstimator(self) -> BodyWarpEstimationBroker:
        """Get body warp estimator"""
        return assertBroker(self.bodyWarpEstimator, "body_warp")

    def getBodyDescriptorEstimator(self, version: Optional[int] = None) -> BodyDescriptorEstimationBroker:
        """Get body descriptor estimator"""
        if self.bodyDescriptorEstimatorMap is None:
            raise EstimatorWasNotInitializedError.format("body_descriptor")
        if version is None:
            version = self.settings.bodyDescriptorBrokerSettings.estimatorSettings.descriptorVersion
        estimator = self.bodyDescriptorEstimatorMap.get(version)
        if not estimator:
            assertDescriptorVersion(version)
            self.initBodyDescriptorEstimator(descriptorVersion=version)
            estimator = self.bodyDescriptorEstimatorMap[version]
        return estimator

    def getAggreagtedBodyDescriptorEstimator(
        self, version: Optional[int] = None
    ) -> AggregatedBodyDescriptorEstimationBroker:
        """Get aggregated body descriptor estimator"""
        if self.aggregatedBodyDescriptorEstimatorMap is None:
            raise EstimatorWasNotInitializedError.format("aggregated_body_descriptor")
        if version is None:
            version = self.settings.bodyDescriptorBrokerSettings.estimatorSettings.descriptorVersion
        estimator = self.aggregatedBodyDescriptorEstimatorMap.get(version)
        if not estimator:
            assertDescriptorVersion(version)
            self.initBodyDescriptorEstimator(descriptorVersion=version)
            estimator = self.aggregatedBodyDescriptorEstimatorMap[version]
        return estimator

    def getRedEyesEstimator(self) -> RedEyesEstimationBroker:
        """Get red-eyes estimator"""
        return assertBroker(self.redEyesEstimator, "red-eyes")

    def getFisheyeEstimator(self) -> FisheyeEstimationBroker:
        """Get fisheye estimator"""
        return assertBroker(self.fisheyeEstimator, "fisheye")

    def getPortraitStyleEstimator(self) -> PortraitStyleEstimationBroker:
        return assertBroker(self.portraitStyleEstimator, "portrait_style")

    def getHeadwearEstimator(self) -> HeadwearEstimationBroker:
        """Get headwear estimator"""
        return assertBroker(self.headwearEstimator, "headwear")

    def getNaturalLightEstimator(self) -> FaceNaturalLightEstimationBroker:
        """Get face natural light estimator"""
        return assertBroker(self.faceNaturalLightEstimator, "face_natural_light")

    def getFaceDetectionBackgroundEstimator(self) -> FaceDetectionBackgroundEstimationBroker:
        """Get face detection background estimator"""
        return assertBroker(self.faceDetectionBackgroundEstimator, "face_detection_background")

    def getImageColorTypeEstimator(self) -> ImageColorTypeEstimationBroker:
        """Get image color type estimator"""
        return assertBroker(self.imageColorTypeEstimator, "image_color_type")

    def getEyebrowExpressionEstimator(self) -> EyebrowExpressionEstimationBroker:
        """Get eyebrow expression estimator"""
        return assertBroker(self.eyebrowExpressionEstimator, "eyebrow_expression")

    def getBodyAttributesEstimator(self) -> BodyAttributesEstimationBroker:
        """Get body attributes estimator"""
        return assertBroker(self.bodyAttributesEstimator, "body_attributes")

    def getLandmarks5Estimator(self) -> Landmarks5EstimationBroker:
        """Get landmarks5 estimator"""
        return assertBroker(self.landmarks5Estimator, "landmarks5")

    def getLandmarks68Estimator(self) -> Landmarks68EstimationBroker:
        """Get landmarks68 estimator"""
        return assertBroker(self.landmarks68Estimator, "landmarks68")

    def getHumanDetector(self) -> HumanDetectionEstimationBroker:
        """Get human detection estimator"""
        return assertBroker(self.humanDetector, "human_detection")

    def getHumanTracker(
        self, detectFace: Literal[0, 1], detectBody: Literal[0, 1], useBodyReId: Literal[0, 1]
    ) -> TEBroker:
        """
        Get human tracker estimator with corresponding tracker settings
        Args:
            detectFace: tracker must or not detect faces
            detectBody: tracker must or not detect bodies
            useBodyReId: tracker must or not use body reId for tracking

        Returns:
            Tracker instance with corresponding tracker settings
        Raises:
            EstimatorWasNotInitializedError: if trackers was not initialized
        """
        if self.humanTrackers is None:
            raise EstimatorWasNotInitializedError.format("human_tracker")
        name = f"human_tracker_face_{detectFace}_body_{detectBody}_reid_{useBodyReId}"
        if not (tracker := self.humanTrackers.get(name)):
            tracker = self.createHumanTracksBroker(
                name, detectFace=detectFace, detectBody=detectBody, useBodyReId=useBodyReId
            )
            self.humanTrackers[name] = tracker
        return tracker
