"""Module contains basic attribute estimator realization"""

from dataclasses import dataclass
from typing import List, Union

from lunavl.sdk.estimators.face_estimators.basic_attributes import BasicAttributes, BasicAttributesEstimator
from lunavl.sdk.estimators.face_estimators.facewarper import FaceWarp, FaceWarpedImage
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class BasicAttributesMonitoringPoint(EstimationPoint):
    """Basic attributes estimation monitoring"""

    name = "basic_attributes"


class AggregatedBasicAttributesMonitoringPoint(EstimationPoint):
    """Aggregated basic attributes estimation monitoring"""

    name = "aggregated_basic_attributes"


@dataclass
class BasicAttributesSettings(EstimatorSettingsBase):
    """Basic attributes estimator settings"""


class BasicAttributesQueueEstimator(
    QueueEstimatorBase[FaceWarpedImage, BasicAttributes, BasicAttributesEstimator, BasicAttributesSettings]
):
    """Basic attributes estimator"""

    async def _estimate(self, dataBatch: list[FaceWarpedImage]) -> list[BasicAttributes]:
        """Estimate basic attributes"""
        with EM.monitoring(BasicAttributesMonitoringPoint(batchSize=len(dataBatch))):
            basicAttributesBatch, _ = await self._estimator.estimateBasicAttributesBatch(
                dataBatch,
                asyncEstimate=True,
                aggregate=False,
                estimateAge=True,
                estimateGender=True,
                estimateEthnicity=True,
            )
            return basicAttributesBatch


class BasicAttributesEstimationBrokerBase(
    EstimationBrokerBase[FaceWarpedImage, BasicAttributes, BasicAttributesEstimator, BasicAttributesSettings]
):
    """Base class for basic attributes brokers"""

    def createEstimator(self, settings: BasicAttributesSettings) -> tuple[BasicAttributesEstimator, VLFaceEngine]:
        """Create basic attributes estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createBasicAttributesEstimator(launchOptions=lo))


class BasicAttributesEstimationBroker(BasicAttributesEstimationBrokerBase):
    """Basic attributes broker"""

    # broker name
    name = "basic_attributes"


class AggregatedBasicAttributesQueueEstimator(
    QueueEstimatorBase[
        list[Union[FaceWarp, FaceWarpedImage]],
        tuple[List[BasicAttributes], BasicAttributes],
        BasicAttributesEstimator,
        BasicAttributesSettings,
    ]
):
    """Aggregated basic attributes estimator"""

    async def _estimate(
        self, dataBatch: list[list[FaceWarpedImage]]
    ) -> list[tuple[list[BasicAttributes], BasicAttributes]]:
        """Estimate aggregated basic attributes"""
        res = []
        with EM.monitoring(AggregatedBasicAttributesMonitoringPoint(batchSize=len(dataBatch))):
            for warpedImageBatch in dataBatch:
                estimationResult = await self._estimator.estimateBasicAttributesBatch(
                    warpedImageBatch,
                    asyncEstimate=True,
                    aggregate=True,
                    estimateAge=True,
                    estimateGender=True,
                    estimateEthnicity=True,
                )
                res.append(estimationResult)
        return res


class AggregatedBasicAttributesEstimationBroker(BasicAttributesEstimationBrokerBase):
    """Aggregated basic attributes broker"""

    # broker name
    name = "aggregated_basic_attributes"
