"""Module contains Deepfake estimator realization"""

from dataclasses import dataclass
from statistics import mean
from typing import NamedTuple, Optional

from lunavl.sdk.detectors.facedetector import FaceDetection
from lunavl.sdk.estimators.face_estimators.deepfake import (
    Deepfake,
    DeepfakeEstimationMode as EDFM,
    DeepfakeEstimator,
    DeepfakePrediction,
)
from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.faceengine.setting_provider import DeepfakeEstimationMode

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class DeepfakePredicted:
    @classmethod
    def initialize(cls, realThreshold: float):
        """
        Initialize by estimator settings.
        Args:
            realThreshold: Deepfake real threshold from estimator settings
        """
        cls.defaultScoreThreshold = realThreshold

    def __init__(self, score: float, prediction: Optional[DeepfakePrediction] = None):
        self.score = score
        self.prediction = prediction

    def makePrediction(self, scoreThreshold: Optional[float]):
        """Make prediction by scoreThreshold."""
        if scoreThreshold is None:
            scoreThreshold = self.defaultScoreThreshold
        if self.score > scoreThreshold:
            self.prediction = DeepfakePrediction.Real
        else:
            self.prediction = DeepfakePrediction.Fake

    def asDict(self) -> dict:
        """Return Deepfake representation."""
        return {"prediction": self.prediction.value, "score": self.score}

    def __repr__(self) -> str:
        """
        Representation.

        Returns:
            str(self.asDict())
        """
        return str(self.asDict())


class DeepfakeMonitoringPoint(EstimationPoint):
    """
    Deepfake estimation monitoring point
    Attributes:
        mode: estimation mode (1-2)
    """

    name = "deepfake"
    __slots__ = ("mode",)

    def __init__(self, mode: int, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.mode = mode

    def getTags(self) -> dict:
        """Get monitoring tags. Add additional tag `descriptor_version`"""
        tags = super().getTags()
        tags["mode"] = self.mode
        return tags


def aggregateDeepfakeFromEstimates(DeepfakeEstimates: list[Deepfake], scoreThreshold: float) -> DeepfakePredicted:
    """
    Get aggregated Deepfake estimation from estimates.
    Args:
        DeepfakeEstimates: Deepfake input estimates
        scoreThreshold: Deepfake score threshold

    Returns:
        aggregated estimation (DeepfakePredicted)
    """
    deepfakeAggregated = DeepfakePredicted(
        score=mean(est.score for est in DeepfakeEstimates),
    )
    deepfakeAggregated.makePrediction(scoreThreshold=scoreThreshold)
    return deepfakeAggregated


@dataclass
class DeepfakeSettings(EstimatorSettingsBase):
    """Deepfake settings"""


class DeepfakeInputData(NamedTuple):
    """
    Input data for Deepfake estimation
    """

    # face detection
    detection: FaceDetection
    # score threshold for prediction
    scoreThreshold: Optional[float] = None


class _DeepfakeQueueEstimator(QueueEstimatorBase[FaceDetection, Deepfake, DeepfakeEstimator, DeepfakeSettings]):
    """Deepfake estimator"""

    async def _estimate(self, dataBatch: list[DeepfakeInputData]) -> list[Deepfake]:
        """Estimate Deepfake.

        After estimation for every estimated Deepfake  will be applied corresponding score threshold for
        correct prediction.
        """
        detections = [detection[0] for detection in dataBatch]
        with EM.monitoring(DeepfakeMonitoringPoint(batchSize=len(dataBatch), mode=self._estimator.mode.value)):
            estimations: list[Deepfake] = await self._estimator.estimateBatch(detections, asyncEstimate=True)
        for idx, deepfake in enumerate(estimations):
            if (scoreThreshold := dataBatch[idx][1]) is not None:
                if deepfake.score < scoreThreshold:
                    deepfake.prediction = DeepfakePrediction.Fake
                else:
                    deepfake.prediction = DeepfakePrediction.Real
        return estimations


class DeepfakeV1QueueEstimator(_DeepfakeQueueEstimator):
    pass


class DeepfakeV2QueueEstimator(_DeepfakeQueueEstimator):
    pass


class DeepfakeQueueEstimationBrokerBase(
    EstimationBrokerBase[FaceDetection, Deepfake, DeepfakeEstimator, DeepfakeSettings]
):
    """Deepfake broker"""

    mode: DeepfakeEstimationMode

    def createEstimator(self, settings: DeepfakeSettings) -> tuple[DeepfakeEstimator, VLFaceEngine]:
        """Create Deepfake estimator"""
        mode = EDFM.M1 if self.mode == DeepfakeEstimationMode.M1 else EDFM.M2
        estimator, faceEngine = createEstimator(
            settings, lambda fe, lo: fe.createDeepfakeEstimator(launchOptions=lo, mode=mode)
        )
        DeepfakePredicted.initialize(faceEngine.faceEngineProvider.deepfakeEstimator.realThreshold)
        return estimator, faceEngine


class DeepfakeV1QueueEstimationBroker(DeepfakeQueueEstimationBrokerBase):
    """Deepfake broker"""

    # broker name
    name = "deepfake_v1"
    mode: DeepfakeEstimationMode = DeepfakeEstimationMode.M1


class DeepfakeV2QueueEstimationBroker(DeepfakeQueueEstimationBrokerBase):
    """Deepfake broker"""

    # broker name
    name = "deepfake_v2"
    mode: DeepfakeEstimationMode = DeepfakeEstimationMode.M2
