from dataclasses import dataclass

from lunavl.sdk.detectors.facedetector import FaceDetection
from lunavl.sdk.estimators.face_estimators.dynamic_range import DynamicRangeEstimator
from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.launch_options import DeviceClass, LaunchOptions

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM
from ...utils.logger import getLogger


class DynamicRangeMonitoringPoint(EstimationPoint):
    """Dynamic range estimation monitoring point"""

    name = "dynamic_range"


@dataclass
class DynamicRangeSettings(EstimatorSettingsBase):
    """Dynamic range settings"""

    # launchOptions = LaunchOptions(deviceClass=DeviceClass.cpu)


class DynamicRangeQueueEstimator(QueueEstimatorBase[FaceDetection, float, DynamicRangeEstimator, DynamicRangeSettings]):
    """Dynamic range estimator"""

    async def _estimate(self, dataBatch: list[FaceDetection]) -> list[float]:
        """Estimate dynamic range"""
        with EM.monitoring(DynamicRangeMonitoringPoint(batchSize=len(dataBatch))):
            return await self._estimator.estimateBatch(dataBatch, asyncEstimate=True)


class DynamicRangeEstimationBroker(
    EstimationBrokerBase[FaceDetection, float, DynamicRangeEstimator, DynamicRangeSettings]
):
    """Dynamic range broker"""

    # broker name
    name = "dynamic_range"

    def createEstimator(self, settings: DynamicRangeSettings) -> tuple[DynamicRangeEstimator, VLFaceEngine]:
        """Create dynamic range estimator"""
        if settings.launchOptions.deviceClass.value != DeviceClass.cpu.value:
            settings = DynamicRangeSettings(launchOptions=LaunchOptions(deviceClass=DeviceClass.cpu))
            getLogger(f"DynamicRangeEstimator").warning(
                "DynamicRangeEstimator can not run using GPU device. DynamicRangeEstimator will use CPU device."
            )
        return createEstimator(settings, lambda fe, lo: fe.createDynamicRangeEstimator(launchOptions=lo))
