"""Module contains face descriptor estimator realization"""

from dataclasses import dataclass
from typing import Tuple, Union

from lunavl.sdk.descriptors.descriptors import FaceDescriptor, FaceDescriptorBatch
from lunavl.sdk.estimators.face_estimators.face_descriptor import FaceDescriptorEstimator
from lunavl.sdk.estimators.face_estimators.facewarper import FaceWarp, FaceWarpedImage
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class FDMonitoringPointBase(EstimationPoint):
    """
    Face descriptor monitoring pot base class.

    Attributes:
        descriptorVersion (int): version of descriptors which was estimated
    """

    __slots__ = "descriptorVersion"

    def __init__(self, descriptorVersion: int, executionTime: float = 0, batchSize: int = 0):
        super().__init__(batchSize=batchSize, executionTime=executionTime)
        self.descriptorVersion = descriptorVersion

    def getTags(self) -> dict:
        """Get monitoring tags. Add additional tag `descriptor_version`"""
        tags = super().getTags()
        tags["descriptor_version"] = self.descriptorVersion
        return tags


class FDMonitoringPoint(FDMonitoringPointBase):
    """Monitoring point for face descriptor estimation"""

    name = "face_descriptor"


class AFDMonitoringPoint(FDMonitoringPointBase):
    """Monitoring point for aggregated face descriptor estimation"""

    name = "aggregated_face_descriptor"


@dataclass
class FaceDescriptorSettings(EstimatorSettingsBase):
    """Face descriptor estimation settings"""

    # descriptor version
    descriptorVersion: int = 65


class FaceDescriptorQueueEstimator(
    QueueEstimatorBase[FaceWarp, FaceDescriptor, FaceDescriptorEstimator, FaceDescriptorSettings]
):
    """Face descriptor queue estimator realization"""

    async def _estimate(self, dataBatch: list[FaceWarp]) -> FaceDescriptorBatch:
        """
        Batch estimation face descriptor
        Args:
            dataBatch: face warp batch
        Returns:
            descriptors batch in the corresponding dataBatch order
        """
        with EM.monitoring(
            FDMonitoringPoint(batchSize=len(dataBatch), descriptorVersion=self.settings.descriptorVersion)
        ):
            descriptors, _ = await self._estimator.estimateDescriptorsBatch(dataBatch, asyncEstimate=True)
            return descriptors


class FaceDescriptorEstimationBroker(
    EstimationBrokerBase[FaceWarp, FaceDescriptor, FaceDescriptorEstimator, FaceDescriptorSettings]
):
    """
    Body descriptor estimation broker. Each broker corresponds to only one descriptor version.
    """

    @property
    def name(self):
        """Broker name"""
        return f"face_descriptor_{self._estimatorSettings.descriptorVersion}"

    def createEstimator(self, settings: FaceDescriptorSettings) -> tuple[FaceDescriptorEstimator, VLFaceEngine]:
        """Create body descriptor estimator"""
        return createEstimator(
            settings, lambda fe, lo: fe.createFaceDescriptorEstimator(settings.descriptorVersion, launchOptions=lo)
        )


class AggregatedFaceDescriptorQueueEstimator(
    QueueEstimatorBase[
        list[Union[FaceWarp, FaceWarpedImage]],
        Tuple[FaceDescriptorBatch, FaceDescriptor],
        FaceDescriptorEstimator,
        FaceDescriptorSettings,
    ]
):
    """Aggregated face descriptor queue estimator realization"""

    async def _estimate(self, dataBatch: list[list[FaceWarp]]) -> list[Tuple[FaceDescriptorBatch, FaceDescriptor]]:
        """
        Batch estimation aggregate face descriptor
        Args:
            dataBatch: batch of face warp batch
        Returns:
            list. Item of list is tuple <descriptors batch, aggregate descriptor> in the corresponding dataBatch order
        """
        with EM.monitoring(
            AFDMonitoringPoint(batchSize=len(dataBatch), descriptorVersion=self.settings.descriptorVersion)
        ):
            res = []
            for warpedImageBatch in dataBatch:
                estimationResult = await self._estimator.estimateDescriptorsBatch(
                    warpedImageBatch, asyncEstimate=True, aggregate=True
                )
                res.append(estimationResult)
            return res


class AggregatedFaceDescriptorEstimationBroker(
    EstimationBrokerBase[
        list[Union[FaceWarp, FaceWarpedImage]],
        Tuple[FaceDescriptorBatch, FaceDescriptor],
        FaceDescriptorEstimator,
        FaceDescriptorSettings,
    ]
):
    """
    Aggregated face descriptor estimation broker. Each broker corresponds to only one descriptor version.
    """

    @property
    def name(self):
        """Broker name"""
        return f"aggregated_face_descriptor_{self._estimatorSettings.descriptorVersion}"

    def createEstimator(self, settings: FaceDescriptorSettings):
        """Create aggregated body descriptor estimator"""
        return createEstimator(
            settings, lambda fe, lo: fe.createFaceDescriptorEstimator(settings.descriptorVersion, launchOptions=lo)
        )
