"""Module contains face detection background estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.detectors.facedetector import FaceDetection
from lunavl.sdk.estimators.face_estimators.background import FaceDetectionBackground, FaceDetectionBackgroundEstimator
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class FaceDetectionBackgroundMonitoringPoint(EstimationPoint):
    """Face detection background estimation monitoring point"""

    name = "face_detection_background"


@dataclass
class FaceDetectionBackgroundSettings(EstimatorSettingsBase):
    """Face detection background settings"""


class FaceDetectionBackgroundQueueEstimator(
    QueueEstimatorBase[
        FaceDetection, FaceDetectionBackground, FaceDetectionBackgroundEstimator, FaceDetectionBackgroundSettings
    ]
):
    """Face detection background estimator"""

    async def _estimate(self, dataBatch: list[FaceDetection]) -> list[FaceDetectionBackground]:
        """Estimate face detection background"""
        with EM.monitoring(FaceDetectionBackgroundMonitoringPoint(batchSize=len(dataBatch))):
            return self._estimator.estimateBatch(dataBatch, asyncEstimate=False)


class FaceDetectionBackgroundEstimationBroker(
    EstimationBrokerBase[
        FaceDetection, FaceDetectionBackground, FaceDetectionBackgroundEstimator, FaceDetectionBackgroundSettings
    ]
):
    """Face detection background broker"""

    # broker name
    name = "face_detection_background"

    def createEstimator(
        self, settings: FaceDetectionBackgroundSettings
    ) -> tuple[FaceDetectionBackgroundEstimator, VLFaceEngine]:
        """Create face detection background estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createFaceDetectionBackgroundEstimator(launchOptions=lo))
