"""
Module contains face detector estimator realization
"""

from dataclasses import dataclass, field

from lunavl.sdk.detectors.base import ImageForDetection, ImageForRedetection
from lunavl.sdk.detectors.facedetector import (
    FaceDetection,
    FaceDetector,
    FaceRedetectResult,
    FacesDetectResult,
    FacesRedetectResult,
)
from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.faceengine.setting_provider import DetectorType, FaceEngineSettingsProvider
from lunavl.sdk.image_utils.image import VLImage

from ..fe import DefaultSettingsProvider
from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM
from ...utils.validators import assertImageForReDetection


class FaceDetMonitoringPoint(EstimationPoint):
    """
    Face detector monitoring point. We get first (random) image for monitoring an image size and a body detection size.

    Attributes:
        detectionCount: detection count on all images from batch
        imageWidth: first image width from batch
        imageHeight: first image height from batch
        detectionWidth: first body detection width on first image from batch
        detectionHeight: first body detection height on first image from batch
    """

    __slots__ = (
        "detectionCount",
        "imageWidth",
        "imageHeight",
        "detectionWidth",
        "detectionHeight",
    )

    name = "face_detection"

    def __init__(self, imageWidth: int, imageHeight: int, executionTime: float = 0, batchSize: int = 0):
        super().__init__(batchSize=batchSize, executionTime=executionTime)
        self.detectionCount = 0
        self.imageWidth = imageWidth
        self.imageHeight = imageHeight
        self.detectionWidth = None
        self.detectionHeight = None

    def updateFromFirstImageResult(self, detectionResult: list[list[FaceDetection]]):
        """
        Update point of detection results
        Args:
            detectionResult: detection result
        """
        # firstImageDetections[0] may be None for re-detect
        if (firstImageDetections := detectionResult[0]) and firstImageDetections[0]:
            rect = firstImageDetections[0].boundingBox.rect
            self.detectionWidth = rect.width
            self.detectionHeight = rect.height
        self.detectionCount = sum((len(detections) for detections in detectionResult))

    def getFields(self) -> dict:
        """Get fields for monitoring"""
        fields = super().getFields()
        fields["image_width"] = self.imageWidth
        fields["image_height"] = self.imageHeight
        fields["detection_count"] = self.detectionCount
        fields["face_detection_width"] = self.detectionWidth
        fields["face_detection_height"] = self.detectionHeight

        return fields


class FaceReDetMonitoringPointBase(FaceDetMonitoringPoint):
    """
    Face detector monitoring point. We get first (random) image for monitoring an image size, bbox
    and a face detection size.

    Attributes:
        bboxWidth: first bbox width on first image from batch
        bboxHeight: first bbox height on first image from batch
    """

    __slots__ = ("bboxWidth", "bboxHeight")

    name = "face_redetection"

    def __init__(
        self,
        bboxWidth: int,
        bboxHeight: int,
        imageWidth: int,
        imageHeight: int,
        executionTime: float = 0,
        batchSize: int = 0,
    ):
        super().__init__(
            batchSize=batchSize,
            executionTime=executionTime,
            imageWidth=imageWidth,
            imageHeight=imageHeight,
        )

        self.bboxWidth = bboxWidth
        self.bboxHeight = bboxHeight

    def getFields(self) -> dict:
        """Get fields for monitoring"""
        fields = super().getFields()
        fields["bbox_width"] = self.bboxWidth
        fields["bbox_height"] = self.bboxHeight

        return fields


@dataclass
class FaceDetV3Settings:
    """
    Face detector version 3 settings
    """

    # (int): target face size for redetect
    redetectFaceTargetSize: int | None = None
    # (bool): whether minFaceSize parameter must be taken as a strict rule
    strictMinSize: bool = True
    # (int): min face size for down scaling by load side
    minFaceSize: int | None = None
    # (int): target face after preprocessing for redetect
    redetectTensorSize: int | None = None
    # (float): redetect score threshold
    redetectScoreThreshold: float | None = None
    # (float): score threshold for detections
    scoreThreshold: float | None = None

    def updateSDKFaceDetV3Settings(self, faceEngineConf: FaceEngineSettingsProvider) -> FaceEngineSettingsProvider:
        """
        Update FaceDetV3 section of the sdk config.
        Args:
            faceEngineConf:  sdk config

        Returns:
            updated config
        """
        faceEngineConf.faceDetV3Settings.strictMinSize = self.strictMinSize
        faceEngineConf.faceDetV3Settings.minFaceSize = self.minFaceSize
        faceEngineConf.faceDetV3Settings.redetectFaceTargetSize = self.redetectFaceTargetSize
        faceEngineConf.faceDetV3Settings.redetectScoreThreshold = self.redetectScoreThreshold
        if self.scoreThreshold is not None:
            faceEngineConf.faceDetV3Settings.scoreThreshold = self.scoreThreshold
        faceEngineConf.faceDetV3Settings.redetectTensorSize = self.redetectTensorSize
        return faceEngineConf

    def __post_init__(self):
        """
        Set up none values from default sdk config.
        """
        faceEngineConf = DefaultSettingsProvider.faceengine
        if self.strictMinSize is None:
            self.strictMinSize = faceEngineConf.faceDetV3Settings.strictMinSize
        if self.minFaceSize is None:
            self.minFaceSize = faceEngineConf.faceDetV3Settings.minFaceSize
        if self.redetectFaceTargetSize is None:
            self.redetectFaceTargetSize = faceEngineConf.faceDetV3Settings.redetectFaceTargetSize
        if self.redetectTensorSize is None:
            self.redetectTensorSize = faceEngineConf.faceDetV3Settings.redetectTensorSize
        if self.redetectScoreThreshold is None:
            self.redetectScoreThreshold = faceEngineConf.faceDetV3Settings.redetectScoreThreshold
        if self.scoreThreshold is None:
            self.scoreThreshold = faceEngineConf.faceDetV3Settings.scoreThreshold


@dataclass
class FaceDetectorSettings(EstimatorSettingsBase):
    """Face detector settings"""

    # face det config
    detectorSettings: FaceDetV3Settings = field(default_factory=FaceDetV3Settings)
    # max faces count per image
    maxFacesCount: int = 128

    def createFEConf(self) -> tuple[FaceEngineSettingsProvider, bool]:
        """
        Create faceEngine conf for detector creation.
        Returns:
             face engine config + runtime config
        """
        faceEngineConf, _ = super().createFEConf()
        faceEngineConf = self.detectorSettings.updateSDKFaceDetV3Settings(faceEngineConf)
        return faceEngineConf, False


class FaceDetectionEstimationBrokerMixin:
    """Face detector creation mixin"""

    def createEstimator(self, settings: FaceDetectorSettings) -> tuple[FaceDetector, VLFaceEngine]:
        """Create face detector"""
        return createEstimator(
            settings, lambda fe, lo: fe.createFaceDetector(DetectorType.FACE_DET_V3, launchOptions=lo)
        )


class FaceDetectionQueueEstimator(
    QueueEstimatorBase[VLImage | ImageForDetection, list[FaceDetection], FaceDetector, FaceDetectorSettings]
):
    """Face detection estimator without landmarks68 estimator"""

    async def _estimate(self, dataBatch: list[VLImage | ImageForDetection]) -> list[list[FaceDetection]]:
        """Detect faces on images without landmarks68"""
        firstImage = dataBatch[0]
        if isinstance(firstImage, VLImage):
            rect = firstImage.rect
        else:
            rect = firstImage.detectArea
        point = FaceDetMonitoringPoint(
            batchSize=len(dataBatch),
            imageWidth=rect.width,
            imageHeight=rect.height,
        )
        with EM.monitoring(point):
            res: FacesDetectResult = await self._estimator.detect(
                dataBatch,
                asyncEstimate=True,
                detect68Landmarks=False,
                limit=self.settings.maxFacesCount,
                detect5Landmarks=False,
            )
            point.updateFromFirstImageResult(res)

        return res


class FaceDetectionEstimationBroker(
    FaceDetectionEstimationBrokerMixin,
    EstimationBrokerBase[VLImage | ImageForDetection, list[FaceDetection], FaceDetector, FaceDetectorSettings],
):
    """Face detection without landmarks68 broker"""

    # broker name
    name = "face_detection"


class FaceReDetectionQueueEstimator(
    QueueEstimatorBase[ImageForRedetection, list[FaceRedetectResult], FaceDetector, FaceDetectorSettings]
):
    """Body ReDetection estimator without landmarks65 estimation"""

    def assertData(self, image: ImageForRedetection):
        """assert image for redetection"""
        assertImageForReDetection(self._estimator._detector, image)

    async def _estimate(self, dataBatch: list[ImageForRedetection]) -> FacesRedetectResult:
        """ReDetect bodies on images without landmarks65"""
        point = FaceReDetMonitoringPointBase(
            batchSize=len(dataBatch),
            imageWidth=dataBatch[0].image.rect.width,
            imageHeight=dataBatch[0].image.rect.height,
            bboxWidth=dataBatch[0].bBoxes[0].width,
            bboxHeight=dataBatch[0].bBoxes[0].height,
        )
        with EM.monitoring(point):
            res = await self._estimator.redetect(dataBatch, asyncEstimate=True, detect68Landmarks=False)
            point.updateFromFirstImageResult(res)
        return res


class FaceReDetectionEstimationBroker(
    FaceDetectionEstimationBrokerMixin,
    EstimationBrokerBase[ImageForRedetection, list[FaceRedetectResult], FaceDetector, FaceDetectorSettings],
):
    """Face ReDetection with landmarks68 broker"""

    # broker name
    name = "face_redetection"
