"""Module contains eyes estimator realization"""

from dataclasses import dataclass
from typing import NamedTuple

from lunavl.sdk.estimators.face_estimators.face_occlusion import (
    FaceOcclusion,
    FaceOcclusionEstimator,
    OcclusionEstimation,
    WarpWithLandmarks,
)
from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.faceengine.setting_provider import FaceEngineSettingsProvider

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class FaceOcclusionMonitoringPoint(EstimationPoint):
    """Face occlusion estimation monitoring"""

    name = "face_occlusion"


@dataclass
class FaceOcclusionSettings(EstimatorSettingsBase):
    """Face occlusion estimator settings"""

    normalHairCoeff: float = 0.15
    overallOcclusionThreshold: float = 0.07
    foreheadThreshold: float = 0.2
    eyeThreshold: float = 0.15
    noseThreshold: float = 0.2
    mouthThreshold: float = 0.15
    lowerFaceThreshold: float = 0.2

    def createFEConf(self) -> tuple[FaceEngineSettingsProvider, bool]:
        """
        Create faceEngine conf for estimator creation.
        Patch setting when creating the config

        Returns:
             face engine config + runtime config
        """
        faceEngineConf, _ = super().createFEConf()
        faceEngineConf.faceOcclusionEstimatorSettings.normalHairCoeff = self.normalHairCoeff
        faceEngineConf.faceOcclusionEstimatorSettings.overallOcclusionThreshold = self.overallOcclusionThreshold
        faceEngineConf.faceOcclusionEstimatorSettings.foreheadThreshold = self.foreheadThreshold
        faceEngineConf.faceOcclusionEstimatorSettings.eyeThreshold = self.eyeThreshold
        faceEngineConf.faceOcclusionEstimatorSettings.noseThreshold = self.noseThreshold
        faceEngineConf.faceOcclusionEstimatorSettings.mouthThreshold = self.mouthThreshold
        faceEngineConf.faceOcclusionEstimatorSettings.lowerFaceThreshold = self.lowerFaceThreshold
        return faceEngineConf, False


class FaceOcclusionInputData(NamedTuple):
    """
    Input data for face occlusion estimation
    """

    # face detection
    detection: WarpWithLandmarks
    # overall threshold for prediction
    overallThreshold: float | None = None
    # hair score threshold for prediction
    hairThreshold: float | None = None
    # forehead score threshold for prediction
    foreheadThreshold: float | None = None
    # left eye score threshold for prediction
    leftEyeThreshold: float | None = None
    # right eye score threshold for prediction
    rightEyeThreshold: float | None = None
    # nose score threshold for prediction
    noseThreshold: float | None = None
    # mouth score threshold for prediction
    mouthThreshold: float | None = None
    # lower face score threshold for prediction
    lowerFaceThreshold: float | None = None


class FaceOcclusionQueueEstimator(
    QueueEstimatorBase[WarpWithLandmarks, FaceOcclusion, FaceOcclusionEstimator, FaceOcclusionSettings]
):
    """
    Face occlusion estimator.
    Note. The overall occlusion score is recalculated taking into account the following logic:
          only that part of the hair occlusion that exceeds the established hair oclusion threshold contributes to the overall occlusion score.
    """

    async def _estimate(self, dataBatch: list[FaceOcclusionInputData]) -> list[FaceOcclusion]:
        """Estimate face occlusion"""
        detections = [detection[0] for detection in dataBatch]
        with EM.monitoring(FaceOcclusionMonitoringPoint(batchSize=len(dataBatch))):
            estimations = await self._estimator.estimateBatch(detections, asyncEstimate=True)
            for idx, faceOcclusion in enumerate(estimations):
                occlusionScore = faceOcclusion.overall.score
                if (hairThreshold := dataBatch[idx].hairThreshold) is not None:
                    defaultHairScore = max(faceOcclusion.hairScore - self.settings.normalHairCoeff, 0)
                    occlusionScore -= defaultHairScore
                    hairScore = max(faceOcclusion.hairScore - hairThreshold, 0)
                    occlusionScore += hairScore
                faceOcclusion.overall = OcclusionEstimation(
                    score=occlusionScore,
                    state=(
                        occlusionScore > threshold
                        if (threshold := dataBatch[idx].overallThreshold) is not None
                        else True
                    ),
                )
                for param in ("forehead", "leftEye", "rightEye", "nose", "mouth", "lowerFace"):
                    score = getattr(faceOcclusion, param).score
                    threshold = getattr(dataBatch[idx], f"{param}Threshold")
                    estimation = OcclusionEstimation(
                        score=score, state=int(score > threshold) if threshold is not None else 0
                    )
                    setattr(faceOcclusion, param, estimation)
            return estimations


class FaceOcclusionEstimationBroker(
    EstimationBrokerBase[WarpWithLandmarks, FaceOcclusion, FaceOcclusionEstimator, FaceOcclusionSettings]
):
    """Eyes broker"""

    # broker name
    name = "face_occlusion"

    def createEstimator(self, settings: FaceOcclusionSettings) -> tuple[FaceOcclusionEstimator, VLFaceEngine]:
        """Create eyes estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createFaceOcclusionEstimator(launchOptions=lo))
