"""Module contains face warp estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.detectors.facedetector import FaceDetection
from lunavl.sdk.estimators.face_estimators.eyes import WarpWithLandmarks5
from lunavl.sdk.estimators.face_estimators.facewarper import FaceWarp, FaceWarper
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class FaceWarpMonitoringPoint(EstimationPoint):
    """Face warp estimation monitoring"""

    name = "face_warp"


@dataclass
class FaceWarperSettings(EstimatorSettingsBase):
    """Face warp estimator settings"""


class FaceWarpFaceWarpQueueEstimator(QueueEstimatorBase[FaceDetection, FaceWarp, FaceWarper, FaceWarperSettings]):
    """Face warp estimator"""

    async def _estimate(self, dataBatch: list[FaceDetection]) -> list[WarpWithLandmarks5]:
        """Estimate face warps"""
        with EM.monitoring(FaceWarpMonitoringPoint(batchSize=len(dataBatch))):
            warps = []
            for detection in dataBatch:
                warp = self._estimator.warp(detection)
                transformedLandmarks5 = self._estimator.makeWarpTransformationWithLandmarks(detection, "L5")
                warps.append(WarpWithLandmarks5(warp, transformedLandmarks5))
            return warps


class FaceWarpEstimationBroker(EstimationBrokerBase[FaceDetection, FaceWarp, FaceWarper, FaceWarperSettings]):
    """Face warp broker"""

    # broker name
    name = "face_warp"

    def createEstimator(self, settings: FaceWarperSettings) -> tuple[FaceWarper, VLFaceEngine]:
        """Create face warper"""
        return createEstimator(settings, lambda fe, lo: fe.createFaceWarper())
