"""Module contains face warp quality estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.estimators.face_estimators.facewarper import FaceWarpedImage
from lunavl.sdk.estimators.face_estimators.warp_quality import Quality, WarpQualityEstimator
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class FaceWarpQualityMonitoringPoint(EstimationPoint):
    """Face warp quality estimation monitoring"""

    name = "face_warp_quality"


@dataclass
class FaceWarpQualitySettings(EstimatorSettingsBase):
    """Face warp quality estimator settings"""


class FaceWarpQualityQueueEstimator(
    QueueEstimatorBase[FaceWarpedImage, Quality, WarpQualityEstimator, FaceWarpQualitySettings]
):
    """Face warp quality estimator"""

    async def _estimate(self, dataBatch: list[FaceWarpedImage]) -> list[Quality]:
        """Estimate face warp quality"""
        with EM.monitoring(FaceWarpQualityMonitoringPoint(batchSize=len(dataBatch))):
            return await self._estimator.estimateBatch(dataBatch, asyncEstimate=True)


class FaceWarpQualityEstimationBroker(
    EstimationBrokerBase[FaceWarpedImage, Quality, WarpQualityEstimator, FaceWarpQualitySettings]
):
    """Face warp quality broker"""

    # broker name
    name = "face_warp_quality"

    def createEstimator(self, settings: FaceWarpQualitySettings) -> tuple[WarpQualityEstimator, VLFaceEngine]:
        """Create face warp quality estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createWarpQualityEstimator(launchOptions=lo))
