"""Module contains fisheye estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.estimators.face_estimators.facewarper import FaceWarpedImage
from lunavl.sdk.estimators.face_estimators.fisheye import Fisheye, FisheyeEstimator
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class FisheyeMonitoringPoint(EstimationPoint):
    """Fisheye estimation monitoring point"""

    name = "fisheye"


@dataclass
class FisheyeSettings(EstimatorSettingsBase):
    """Fisheye settings"""


class FisheyeQueueEstimator(QueueEstimatorBase[FaceWarpedImage, Fisheye, FisheyeEstimator, FisheyeSettings]):
    """Fisheye estimator"""

    async def _estimate(self, dataBatch: list[FaceWarpedImage]) -> list[Fisheye]:
        """Estimate fisheye"""
        with EM.monitoring(FisheyeMonitoringPoint(batchSize=len(dataBatch))):
            return await self._estimator.estimateBatch(dataBatch, asyncEstimate=True)


class FisheyeEstimationBroker(EstimationBrokerBase[FaceWarpedImage, Fisheye, FisheyeEstimator, FisheyeSettings]):
    """Fisheye broker"""

    # broker name
    name = "fisheye"

    def createEstimator(self, settings: FisheyeSettings) -> tuple[FisheyeEstimator, VLFaceEngine]:
        """Create fisheye estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createFisheyeEstimator(launchOptions=lo))
