"""Module contains gaze direction estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.estimators.face_estimators.eyes import GazeDirection, GazeEstimator, WarpWithLandmarks5
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM
from ...utils.async_tools import gather


class GazeDirectionMonitoringPoint(EstimationPoint):
    """Gaze direction estimation monitoring point"""

    name = "gaze_direction"


@dataclass
class GazeDirectionSettings(EstimatorSettingsBase):
    """Gaze direction estimator settings"""


class GazeDirectionQueueEstimator(
    QueueEstimatorBase[WarpWithLandmarks5, GazeDirection, GazeEstimator, GazeDirectionSettings]
):
    """Gaze direction estimator"""

    async def _estimate(self, dataBatch: list[WarpWithLandmarks5]) -> list[GazeDirection]:
        """Estimate gaze direction"""
        with EM.monitoring(GazeDirectionMonitoringPoint(batchSize=len(dataBatch))):
            return await gather((self._estimator.estimate(warp, asyncEstimate=True) for warp in dataBatch))


class GazeEstimationBroker(
    EstimationBrokerBase[WarpWithLandmarks5, GazeDirection, GazeDirectionQueueEstimator, GazeDirectionSettings]
):
    """Gaze direction broker"""

    # broker name
    name = "gaze_direction"

    def createEstimator(self, settings: GazeDirectionSettings) -> tuple[GazeDirectionQueueEstimator, VLFaceEngine]:
        """Create gaze direction estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createGazeEstimator(launchOptions=lo))
