"""Module contains head pose estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.detectors.facedetector import FaceDetection
from lunavl.sdk.estimators.face_estimators.head_pose import HeadPose, HeadPoseEstimator
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class HeadPoseMonitoringPoint(EstimationPoint):
    """Head pose estimation monitoring point"""

    name = "head_pose"


@dataclass
class HeadPoseSettings(EstimatorSettingsBase):
    """Glasses settings"""


class HeadPoseQueueEstimator(QueueEstimatorBase[FaceDetection, HeadPose, HeadPoseEstimator, HeadPoseSettings]):
    """Head pose estimator"""

    async def _estimate(self, dataBatch: list[FaceDetection]) -> list[HeadPose]:
        """Estimate head pose"""
        with EM.monitoring(HeadPoseMonitoringPoint(batchSize=len(dataBatch))):
            return await self._estimator.estimateBatch(dataBatch, asyncEstimate=True)


class HeadPoseEstimationBroker(EstimationBrokerBase[FaceDetection, HeadPose, HeadPoseEstimator, HeadPoseSettings]):
    """Head pose broker"""

    # broker name
    name = "head_pose"

    def createEstimator(self, settings: HeadPoseSettings) -> tuple[HeadPoseEstimator, VLFaceEngine]:
        """Create head pose estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createHeadPoseEstimator(launchOptions=lo))
