"""Module contains image color type estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.estimators.face_estimators.facewarper import FaceWarpedImage
from lunavl.sdk.estimators.face_estimators.image_type import ImageColorType, ImageColorTypeEstimator
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class ImageColorTypeMonitoringPoint(EstimationPoint):
    """Image color type estimation monitoring point"""

    name = "image_color_type"


@dataclass
class ImageColorTypeSettings(EstimatorSettingsBase):
    """Image color type settings"""


class ImageColorTypeQueueEstimator(
    QueueEstimatorBase[FaceWarpedImage, ImageColorType, ImageColorTypeEstimator, ImageColorTypeSettings]
):
    """Image color type estimator"""

    async def _estimate(self, dataBatch: list[FaceWarpedImage]) -> list[ImageColorType]:
        """Estimate Image color type"""
        with EM.monitoring(ImageColorTypeMonitoringPoint(batchSize=len(dataBatch))):
            return await self._estimator.estimateBatch(dataBatch, asyncEstimate=True)


class ImageColorTypeEstimationBroker(
    EstimationBrokerBase[FaceWarpedImage, ImageColorType, ImageColorTypeEstimator, ImageColorTypeSettings]
):
    """Image color type broker"""

    # broker name
    name = "image_color_type"

    def createEstimator(self, settings: ImageColorTypeSettings) -> tuple[ImageColorTypeEstimator, VLFaceEngine]:
        """Create image color type estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createImageColorTypeEstimator(launchOptions=lo))
