"""Module contains face face landmarks estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.detectors.facedetector import FaceDetection, FaceLandmarks, Landmarks5, Landmarks68
from lunavl.sdk.estimators.face_estimators.landmarks import FaceLandmarksEstimator
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class Landmarks5MonitoringPoint(EstimationPoint):
    """Face landmarks5 estimation monitoring"""

    name = "landmarks5"


@dataclass
class Landmarks5Settings(EstimatorSettingsBase):
    """Landmarks5 estimator settings"""


class Landmarks5QueueEstimator(
    QueueEstimatorBase[FaceDetection, Landmarks5, FaceLandmarksEstimator, Landmarks5Settings]
):
    """Landmarks5 estimator"""

    async def _estimate(self, dataBatch: list[FaceDetection]) -> list[Landmarks5]:
        """Estimate face Landmarks5"""
        with EM.monitoring(Landmarks5MonitoringPoint(batchSize=len(dataBatch))):
            landmarks = await self._estimator.estimateBatch(
                dataBatch, asyncEstimate=True, landmarksType=FaceLandmarks.Landmarks5
            )
            return landmarks


class Landmarks5EstimationBroker(
    EstimationBrokerBase[FaceDetection, Landmarks5, FaceLandmarksEstimator, Landmarks5Settings]
):
    """Face landmarks5 broker"""

    # broker name
    name = "landmarks5"

    def createEstimator(self, settings: Landmarks5Settings) -> tuple[FaceLandmarksEstimator, VLFaceEngine]:
        """Create landmarks5 estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createFaceLandmarksEstimator(launchOptions=lo))


class Landmarks68MonitoringPoint(EstimationPoint):
    """Face landmarks68 estimation monitoring"""

    name = "landmarks68"


@dataclass
class Landmarks68Settings(EstimatorSettingsBase):
    """Landmarks68 estimator settings"""


class Landmarks68QueueEstimator(
    QueueEstimatorBase[FaceDetection, Landmarks68, FaceLandmarksEstimator, Landmarks68Settings]
):
    """Landmarks68 estimator"""

    async def _estimate(self, dataBatch: list[FaceDetection]) -> list[Landmarks68]:
        """Estimate face Landmarks68"""
        with EM.monitoring(Landmarks68MonitoringPoint(batchSize=len(dataBatch))):
            landmarks = await self._estimator.estimateBatch(
                dataBatch, asyncEstimate=True, landmarksType=FaceLandmarks.Landmarks68
            )
            return landmarks


class Landmarks68EstimationBroker(
    EstimationBrokerBase[FaceDetection, Landmarks68, FaceLandmarksEstimator, Landmarks68Settings]
):
    """Face landmarks68 broker"""

    # broker name
    name = "landmarks68"

    def createEstimator(self, settings: Landmarks68Settings) -> tuple[FaceLandmarksEstimator, VLFaceEngine]:
        """Create landmarks68 estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createFaceLandmarksEstimator(launchOptions=lo))
