"""Module contains livenessv1 estimator realization"""

from dataclasses import dataclass
from statistics import mean
from typing import NamedTuple, Optional

from lunavl.sdk.detectors.facedetector import FaceDetection
from lunavl.sdk.estimators.face_estimators.livenessv1 import LivenessPrediction, LivenessV1, LivenessV1Estimator
from lunavl.sdk.faceengine.engine import VLFaceEngine

from .liveness_predictor import LivenessPredicted
from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class LivenessV1MonitoringPoint(EstimationPoint):
    """LivenessV1 estimation monitoring point"""

    name = "livenessV1"


class LivenessInputData(NamedTuple):
    """
    Input data for livenessV1 estimation
    """

    # face detection
    detection: FaceDetection
    # score threshold for prediction
    scoreThreshold: Optional[float] = None


def aggregateLivenessFromEstimates(
    livenessEstimates: list[LivenessV1], scoreThreshold: Optional[float] = None
) -> LivenessPredicted:
    """
    Get aggregated liveness from estimates.
    Args:
        livenessEstimates: liveness input estimates
        scoreThreshold: liveness score threshold

    Returns:
        aggregated estimation (LivenessPredicted)
    """
    livenessAggregated = LivenessPredicted(
        score=mean(est.score for est in livenessEstimates),
    )
    livenessAggregated.makePrediction(scoreThreshold=scoreThreshold)
    return livenessAggregated


@dataclass
class LivenessV1Settings(EstimatorSettingsBase):
    """LivenessV1 settings"""


class LivenessV1QueueEstimator(QueueEstimatorBase[FaceDetection, LivenessV1, LivenessV1Estimator, LivenessV1Settings]):
    """LivenessV1 estimator"""

    async def _estimate(self, dataBatch: list[LivenessInputData]) -> list[LivenessV1]:
        """Estimate livenessV1.

        After estimation for every estimated liveness  will be applied corresponding score threshold for correct prediction.
        """
        detections = [detection[0] for detection in dataBatch]
        with EM.monitoring(LivenessV1MonitoringPoint(batchSize=len(dataBatch))):
            estimations: list[LivenessV1] = await self._estimator.estimateBatch(detections, asyncEstimate=True)
        for idx, liveness in enumerate(estimations):
            if (scoreThreshold := dataBatch[idx][1]) is not None:
                if liveness.score < scoreThreshold:
                    liveness.prediction = LivenessPrediction.Spoof
                else:
                    liveness.prediction = LivenessPrediction.Real
        return estimations


class Livenessv1QueueEstimationBroker(
    EstimationBrokerBase[FaceDetection, LivenessV1, LivenessV1Estimator, LivenessV1Settings]
):
    """Livenessv1 broker"""

    # broker name
    name = "livenessv1"

    def createEstimator(self, settings: LivenessV1Settings) -> tuple[LivenessV1Estimator, VLFaceEngine]:
        """Create livenessv1 estimator"""
        estimator, faceEngine = createEstimator(settings, lambda fe, lo: fe.createLivenessV1Estimator(launchOptions=lo))
        LivenessPredicted.initialize(faceEngine.faceEngineProvider.livenessV1Estimator.realThreshold)
        return estimator, faceEngine
