"""Module contains livenessv1 estimator realization"""

from dataclasses import dataclass

from lunavl.sdk.detectors.facedetector import FaceDetection
from lunavl.sdk.estimators.face_estimators.portrait_style import PortraitStyle, PortraitStyleEstimator
from lunavl.sdk.faceengine.engine import VLFaceEngine

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class PortraitStyleMonitoringPoint(EstimationPoint):
    """Portrait style estimation monitoring point"""

    name = "portrait_style"


@dataclass
class PortraitStyleEstimatorSettings(EstimatorSettingsBase):
    """Portrait style settings"""


class PortraitStyleQueueEstimator(
    QueueEstimatorBase[FaceDetection, PortraitStyle, PortraitStyleEstimator, PortraitStyleEstimatorSettings]
):
    """Portrait style estimator"""

    async def _estimate(self, dataBatch: list[FaceDetection]) -> list[PortraitStyle]:
        """Estimate portrait_style"""
        with EM.monitoring(PortraitStyleMonitoringPoint(batchSize=len(dataBatch))):
            return self._estimator.estimateBatch(dataBatch, asyncEstimate=False)


class PortraitStyleEstimationBroker(
    EstimationBrokerBase[FaceDetection, PortraitStyle, PortraitStyleEstimator, PortraitStyleEstimatorSettings]
):
    """Portrait style broker"""

    # broker name
    name = "portrait_style"

    def createEstimator(self, settings: PortraitStyleEstimatorSettings) -> tuple[PortraitStyleEstimator, VLFaceEngine]:
        """Create portrait_style estimator"""
        return createEstimator(settings, lambda fe, lo: fe.createPortraitStyleEstimator(launchOptions=lo))
