from typing import Optional

from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.faceengine.setting_provider import (
    FaceEngineSettingsProvider,
    RuntimeSettings,
    RuntimeSettingsProvider,
    SystemSettings,
)

# keep runtime global configuration. Runtime is initialized with first FaceEngine instance and live while exist any FE.
# default FaceEngine
_FACE_ENGINE: Optional[VLFaceEngine] = None
# global runtime conf
_RUNTIME: Optional[RuntimeSettingsProvider] = None


def initFE(runtimeConf: RuntimeSettingsProvider, faceEngineConf: Optional[FaceEngineSettingsProvider] = None):
    """Init global runtime and default FaceEngine"""
    global _FACE_ENGINE
    global _RUNTIME
    if _RUNTIME:
        raise RuntimeError("Double FaceEngine initialization")
    _RUNTIME = runtimeConf
    _DefaultSettingsProvider._runtime = runtimeConf.runtimeSettings
    if faceEngineConf is None:
        _FACE_ENGINE = VLFaceEngine(runtimeConf=runtimeConf)
    else:
        _FACE_ENGINE = VLFaceEngine(runtimeConf=runtimeConf, faceEngineConf=faceEngineConf)
        _DefaultSettingsProvider._system = faceEngineConf.systemSettings


def resetFE():
    """Reset FaceEngine"""
    global _FACE_ENGINE
    global _RUNTIME
    if not _FACE_ENGINE:
        raise RuntimeError("FaceEngine not initialized yet")
    _FACE_ENGINE.coreFaceEngine.reset()
    _FACE_ENGINE = None
    _RUNTIME = None


def getFE(faceEngineConf: Optional[FaceEngineSettingsProvider] = None) -> VLFaceEngine:
    """
    Get FaceEngine instance with corresponding configuration
    Returns:
        global FE if faceEngineConf is None otherwise will be generated new FE.
    """
    global _FACE_ENGINE
    global _RUNTIME
    if _RUNTIME is None:
        raise RuntimeError("Runtime was not initialized")
    if faceEngineConf:
        return VLFaceEngine(runtimeConf=_RUNTIME, faceEngineConf=faceEngineConf)
    return _FACE_ENGINE


class _DefaultSettingsProvider(type):
    """Default SDK settings provider."""

    _runtime: RuntimeSettings | None = None
    _system: SystemSettings | None = None

    @property
    def runtime(self):
        conf = RuntimeSettingsProvider()
        if self._runtime:
            conf.runtimeSettings.verboseLogging = self._runtime.verboseLogging
        return conf

    @property
    def faceengine(self):
        conf = FaceEngineSettingsProvider()
        if self._system:
            conf.systemSettings.verboseLogging = self._system.verboseLogging
        return conf


class DefaultSettingsProvider(metaclass=_DefaultSettingsProvider): ...
