from dataclasses import dataclass
from typing import Literal, Tuple

import FaceEngine
from lunavl.sdk.estimators.estimators_utils.extractor_utils import validateInputByBatchEstimator
from lunavl.sdk.estimators.image_estimators.people_count import (
    ImageForPeopleEstimation,
    PeopleCount,
    PeopleCountEstimator,
)
from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.faceengine.setting_provider import PeopleCountEstimatorType
from lunavl.sdk.image_utils.geometry import Rect, Union
from lunavl.sdk.image_utils.image import VLImage

from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM

PEOPLE_COUNT_ESTIMATOR_TYPE = Literal["DEFAULT", "PEOPLE_COUNT_V1", "PEOPLE_COUNT_V2"]


class PeopleCountMonitoringPoint(EstimationPoint):
    """People count estimation monitoring point"""

    name = "people_count"


@dataclass
class PeopleCountSettings(EstimatorSettingsBase):
    """People count settings"""

    estimatorType: PEOPLE_COUNT_ESTIMATOR_TYPE = "PEOPLE_COUNT_V2"


class PeopleCountQueueEstimator(QueueEstimatorBase[VLImage, PeopleCount, PeopleCountEstimator, PeopleCountSettings]):
    def assertData(self, data: Union[VLImage, Tuple[VLImage, Rect], ImageForPeopleEstimation]) -> None:
        """assert input data"""
        if isinstance(data, VLImage):
            return
        detectArea = data[1].coreRectI
        image = data[0]
        validateInputByBatchEstimator(
            self._estimator._coreEstimator,
            [image.coreImage],
            [detectArea],
            FaceEngine.CrowdRequest.estimateHeadCountAndCoords,
        )

    async def _estimate(self, dataBatch: list[VLImage]) -> list[PeopleCount]:
        """Estimate people count"""
        with EM.monitoring(PeopleCountMonitoringPoint(batchSize=len(dataBatch))):
            peopleCount: list[int] = await self._estimator.estimateBatch(dataBatch, asyncEstimate=True)
            return peopleCount


class PeopleCountEstimationBroker(
    EstimationBrokerBase[Union[VLImage, Tuple[VLImage, Rect]], PeopleCount, PeopleCountEstimator, PeopleCountSettings]
):
    """People count broker"""

    name = "people_count"

    def createEstimator(self, settings: PeopleCountSettings) -> tuple[PeopleCountEstimator, VLFaceEngine]:
        """Create people count estimator"""

        return createEstimator(
            settings,
            lambda fe, lo: fe.createPeopleCountEstimatorV2(
                PeopleCountEstimatorType[settings.estimatorType], launchOptions=lo
            ),
        )
