"""
Module contains human detector estimator realization.
"""

from dataclasses import dataclass

from lunavl.sdk.detectors.base import ImageForDetection
from lunavl.sdk.detectors.humandetector import HumanDetection, HumanDetector
from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.faceengine.setting_provider import FaceEngineSettingsProvider
from lunavl.sdk.image_utils.image import VLImage

from ..fe import DefaultSettingsProvider
from ..worker import EstimationBrokerBase, EstimatorSettingsBase, QueueEstimatorBase, createEstimator
from ...monitoring_utils.estimation_monitoring import EstimationPoint
from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class HumanDetMonitoringPoint(EstimationPoint):
    """
    Human detector monitoring point. We get first (random) image and detection for monitoring an image size and
    detections size.

    Attributes:
        detectionCount: human detection count on all images from batch
        faceDetectionCount: face detection count on all images from batch
        bodyDetectionCount: face detection count on all images from batch
        imageWidth: first image width from batch
        imageHeight: first image height from batch
        bodyDetectionWidth: first body detection width on first image from batch
        bodyDetectionHeight: first body detection height on first image from batch
        faceDetectionWidth: first face detection width on first image from batch
        faceDetectionHeight: first face detection height on first image from batch
    """

    __slots__ = (
        "faceDetectionCount",
        "bodyDetectionCount",
        "detectionCount",
        "imageWidth",
        "imageHeight",
        "detectionWidth",
        "detectionHeight",
        "bodyDetectionWidth",
        "bodyDetectionHeight",
        "faceDetectionWidth",
        "faceDetectionHeight",
    )
    name = "human_detection"

    def __init__(self, imageWidth: int, imageHeight: int, executionTime: float = 0, batchSize: int = 0):
        super().__init__(batchSize=batchSize, executionTime=executionTime)
        self.faceDetectionCount = 0
        self.bodyDetectionCount = 0
        self.detectionCount = 0
        self.imageWidth = imageWidth
        self.imageHeight = imageHeight
        self.bodyDetectionWidth = None
        self.bodyDetectionHeight = None
        self.faceDetectionWidth = None
        self.faceDetectionHeight = None

    def updateFromFirstImageResult(self, detectionResult: list[list[HumanDetection]]):
        """
        Update point of detection results
        Args:
            detectionResult: detection result
        """
        for imageDetections in detectionResult:
            for detection in imageDetections:
                if face := detection.face:
                    self.faceDetectionCount += 1
                    if not self.faceDetectionHeight:
                        self.faceDetectionHeight = face.boundingBox.rect.height
                        self.faceDetectionWidth = face.boundingBox.rect.width
                if body := detection.body:
                    if not self.bodyDetectionHeight:
                        self.bodyDetectionHeight = body.boundingBox.rect.height
                        self.bodyDetectionWidth = body.boundingBox.rect.width
                    self.bodyDetectionCount += 1
                self.detectionCount += 1

    def getFields(self) -> dict:
        """Get fields for monitoring"""
        fields = super().getFields()
        fields["image_width"] = self.imageWidth
        fields["image_height"] = self.imageHeight
        fields["face_detection_count"] = self.faceDetectionCount
        fields["body_detection_count"] = self.bodyDetectionCount
        fields["detection_count"] = self.detectionCount
        fields["body_detection_width"] = self.bodyDetectionWidth
        fields["body_detection_height"] = self.bodyDetectionHeight
        fields["face_detection_width"] = self.faceDetectionWidth
        fields["face_detection_height"] = self.faceDetectionHeight

        return fields


@dataclass
class HumanDetectorSettings(EstimatorSettingsBase):
    """Human detector settings"""

    # (bool): whether minFaceSize parameter must be taken as a strict rule.
    strictMinSize: bool = True
    # (int): min face size for down scaling by load side
    minFaceSize: int | None = None
    # (float): face detection score threshold
    faceThreshold: float | None = None
    # (float): face detection score threshold
    bodyThreshold: float | None = None
    # (float): face and body association detection score threshold
    associationThreshold: float | None = None

    def createFEConf(self) -> tuple[FaceEngineSettingsProvider, bool]:
        """
        Create faceEngine conf for detector creation.
        Returns:
             face engine config
        """
        faceEngineConf, _ = super().createFEConf()
        faceEngineConf = self.updateSettings(faceEngineConf)
        return faceEngineConf, False

    def updateSettings(self, faceEngineConf: FaceEngineSettingsProvider) -> FaceEngineSettingsProvider:
        """
        Update HumanDetector section of the sdk config.
        Args:
            faceEngineConf:  sdk config

        Returns:
            updated config
        """
        faceEngineConf.humanDetectorSettings.strictMinSize = self.minFaceSize
        faceEngineConf.humanDetectorSettings.minFaceSize = self.minFaceSize
        if self.faceThreshold is not None:
            faceEngineConf.humanDetectorSettings.faceThreshold = self.faceThreshold
        if self.bodyThreshold is not None:
            faceEngineConf.humanDetectorSettings.bodyThreshold = self.bodyThreshold
        if self.associationThreshold is not None:
            faceEngineConf.humanDetectorSettings.associationThreshold = self.associationThreshold
        return faceEngineConf

    def __post_init__(self):
        """
        Set up none values from default sdk config.
        """
        faceEngineConf = DefaultSettingsProvider.faceengine
        if self.strictMinSize is None:
            self.strictMinSize = faceEngineConf.humanDetectorSettings.strictMinSize
        if self.minFaceSize is None:
            self.minFaceSize = faceEngineConf.humanDetectorSettings.minFaceSize
        if self.bodyThreshold is None:
            self.bodyThreshold = faceEngineConf.humanDetectorSettings.bodyThreshold
        if self.faceThreshold is None:
            self.faceThreshold = faceEngineConf.humanDetectorSettings.faceThreshold
        if self.associationThreshold is None:
            self.associationThreshold = faceEngineConf.humanDetectorSettings.associationThreshold


class HumanDetectionQueueEstimator(
    QueueEstimatorBase[VLImage | ImageForDetection, list[HumanDetection], HumanDetector, HumanDetectorSettings]
):
    """Body detection estimator"""

    async def _estimate(self, dataBatch: list[VLImage | ImageForDetection]) -> list[list[HumanDetection]]:
        """Detect bodies on images"""
        firstImage = dataBatch[0]
        if isinstance(firstImage, VLImage):
            rect = firstImage.rect
        else:
            rect = firstImage.detectArea

        point = HumanDetMonitoringPoint(
            batchSize=len(dataBatch),
            imageWidth=rect.width,
            imageHeight=rect.height,
        )
        with EM.monitoring(point):
            res = await self._estimator.detect(
                dataBatch,
                asyncEstimate=True,
            )
            point.updateFromFirstImageResult(res)
            return res


class HumanDetectionEstimationBroker(
    EstimationBrokerBase[VLImage | ImageForDetection, list[HumanDetection], HumanDetector, HumanDetectorSettings]
):
    """Human detection base broker"""

    name = "human_detection"

    def createEstimator(self, settings: HumanDetectorSettings) -> tuple[HumanDetector, VLFaceEngine]:
        """Create human detector"""
        return createEstimator(settings, lambda fe, lo: fe.createHumanDetector(launchOptions=lo))
