"""
Module contains settings for a loop work
"""

from copy import deepcopy
from dataclasses import dataclass, field
from typing import Protocol

from .body_estimators.body_attributes_estimator import BodyAttributesSettings
from .body_estimators.body_descriptor_estimator import BodyDescriptorSettings
from .body_estimators.body_detector import BodyDetectorSettings
from .body_estimators.body_warp_estimator import BodyWarperSettings
from .face_estimators.ags_estimator import AGSSettings
from .face_estimators.basic_attributes_estimator import BasicAttributesSettings
from .face_estimators.deepfake_estimator import DeepfakeSettings
from .face_estimators.dynamic_range_estimator import DynamicRangeSettings
from .face_estimators.emotions_estimator import EmotionsSettings
from .face_estimators.eyebrow_expression_estimator import EyebrowExpressionSettings
from .face_estimators.eyes_estimator import EyesSettings
from .face_estimators.face_descriptor_estimator import FaceDescriptorSettings
from .face_estimators.face_detection_background_estimator import FaceDetectionBackgroundSettings
from .face_estimators.face_detector import FaceDetectorSettings
from .face_estimators.face_natural_light_estimator import FaceNaturalLightSettings
from .face_estimators.face_occlusion_estimator import FaceOcclusionSettings
from .face_estimators.face_warp_estimator import FaceWarperSettings
from .face_estimators.face_warp_quality_estimator import FaceWarpQualitySettings
from .face_estimators.facial_hair_estimator import FacialHairSettings
from .face_estimators.fisheye_estimator import FisheyeSettings
from .face_estimators.gaze_direction_estimator import GazeDirectionSettings
from .face_estimators.glasses_estimator import GlassesSettings
from .face_estimators.head_pose_estimator import HeadPoseSettings
from .face_estimators.headwear import HeadwearSettings
from .face_estimators.image_color_type import ImageColorTypeSettings
from .face_estimators.landmarks_estimator import Landmarks5Settings, Landmarks68Settings
from .face_estimators.liveness1_estimator import LivenessV1Settings
from .face_estimators.mask_estimator import MaskSettings
from .face_estimators.mouth_estimator import MouthStatesSettings
from .face_estimators.portrait_style_estimator import PortraitStyleEstimatorSettings
from .face_estimators.red_eyes_estimator import RedEyesSettings
from .image_estimators.image_modification import ImageModificationSettings
from .image_estimators.orientation_mode_estimator import ImageOrientationSettings
from .image_estimators.people_count_estimator import PeopleCountSettings
from .others_estimators.human_detector import HumanDetectorSettings
from .te.human_tracker import HumanTrackSettings
from .worker import EstimatorSettingsBase


@dataclass
class BrokerSettingsProtocol(Protocol):
    """Estimation broker settings protocol"""

    # optimal batch size for estimation
    optimalBatchSize: int
    # estimator settings
    estimatorSettings: EstimatorSettingsBase
    # count worker  which a broker will start
    workerCount: int


@dataclass
class ImageOrientationBrokerSettings:
    """Image orientation broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: ImageOrientationSettings = field(default_factory=ImageOrientationSettings)
    workerCount: int = 1


@dataclass
class ImageModificationBrokerSettings:
    """Image modification broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: ImageModificationSettings = field(default_factory=ImageModificationSettings)
    workerCount: int = 1


@dataclass
class PeopleCountBrokerSettings:
    """Image orientation broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: PeopleCountSettings = field(default_factory=PeopleCountSettings)
    workerCount: int = 1


@dataclass
class DynamicRangeBrokerSettings:
    """Dynamic range broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: DynamicRangeSettings = field(default_factory=DynamicRangeSettings)
    workerCount: int = 1


@dataclass
class AGSBrokerSettings:
    """AGS broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: AGSSettings = field(default_factory=AGSSettings)
    workerCount: int = 1


@dataclass
class PortraitStyleBrokerSettings:
    """Portrait style broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: PortraitStyleEstimatorSettings = field(default_factory=PortraitStyleEstimatorSettings)
    workerCount: int = 1


@dataclass
class FaceDetectionBrokerSettings:
    """Face detection broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FaceDetectorSettings = field(default_factory=FaceDetectorSettings)
    workerCount: int = 2


@dataclass
class BodyDetectionBrokerSettings:
    """Body detection broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: BodyDetectorSettings = field(default_factory=BodyDetectorSettings)
    workerCount: int = 2


@dataclass
class FaceDescriptorBrokerSettings:
    """Face descriptor broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FaceDescriptorSettings = field(default_factory=FaceDescriptorSettings)
    workerCount: int = 2


@dataclass
class BodyDescriptorBrokerSettings:
    """Body descriptor broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: BodyDescriptorSettings = field(default_factory=BodyDescriptorSettings)
    workerCount: int = 2


@dataclass
class FaceWarpBrokerSettings:
    """Face warp broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FaceWarperSettings = field(default_factory=FaceWarperSettings)
    workerCount: int = 1


@dataclass
class BodyWarpBrokerSettings:
    """Body warp broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: BodyWarperSettings = field(default_factory=BodyWarperSettings)
    workerCount: int = 1


@dataclass
class BasicAttributesBrokerSettings:
    """Basic attributes broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: BasicAttributesSettings = field(default_factory=BasicAttributesSettings)
    workerCount: int = 1


@dataclass
class EmotionsBrokerSettings:
    """Emotions broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: EmotionsSettings = field(default_factory=EmotionsSettings)
    workerCount: int = 1


@dataclass
class MaskBrokerSettings:
    """Mask broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: MaskSettings = field(default_factory=MaskSettings)
    workerCount: int = 1


@dataclass
class FaceWarpQualityBrokerSettings:
    """Face warp quality broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FaceWarpQualitySettings = field(default_factory=FaceWarpQualitySettings)
    workerCount: int = 1


@dataclass
class MouthStateBrokerSettings:
    """Mouth state broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: MouthStatesSettings = field(default_factory=MouthStatesSettings)
    workerCount: int = 1


@dataclass
class LivenessV1BrokerSettings:
    """LivenessV1 broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: LivenessV1Settings = field(default_factory=LivenessV1Settings)
    workerCount: int = 1


@dataclass
class DeepfakeBrokerSettings:
    """Deepfake broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: DeepfakeSettings = field(default_factory=DeepfakeSettings)
    workerCount: int = 1


@dataclass
class HeadPoseBrokerSettings:
    """HeadPose broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: HeadPoseSettings = field(default_factory=HeadPoseSettings)
    workerCount: int = 1


@dataclass
class GlassesBrokerSettings:
    """Glasses broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: GlassesSettings = field(default_factory=GlassesSettings)
    workerCount: int = 1


@dataclass
class FaceOcclusionBrokerSettings:
    """Face occlusion broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FaceOcclusionSettings = field(default_factory=FaceOcclusionSettings)
    workerCount: int = 1


@dataclass
class FacialHairBrokerSettings:
    """Facial hair broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FacialHairSettings = field(default_factory=FacialHairSettings)
    workerCount: int = 1


@dataclass
class EyesBrokerSettings:
    """Eyes broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: EyesSettings = field(default_factory=EyesSettings)
    workerCount: int = 1


@dataclass
class GazeDirectionBrokerSettings:
    """Gaze direction broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: GazeDirectionSettings = field(default_factory=GazeDirectionSettings)
    workerCount: int = 1


@dataclass
class EyebrowExpressionBrokerSettings:
    """Eyebrow expression broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: EyebrowExpressionSettings = field(default_factory=EyebrowExpressionSettings)
    workerCount: int = 1


@dataclass
class HeadwearBrokerSettings:
    """Headwear expression broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: HeadwearSettings = field(default_factory=HeadwearSettings)
    workerCount: int = 1


@dataclass
class ImageColorTypeBrokerSettings:
    """Image color type broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: ImageColorTypeSettings = field(default_factory=ImageColorTypeSettings)
    workerCount: int = 1


@dataclass
class RedEyesBrokerSettings:
    """Red eyes broker settings,realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: RedEyesSettings = field(default_factory=RedEyesSettings)
    workerCount: int = 1


@dataclass
class FisheyeBrokerSettings:
    """Fisheye broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FisheyeSettings = field(default_factory=FisheyeSettings)
    workerCount: int = 1


@dataclass
class FaceNaturalLightBrokerSettings:
    """Face natural light broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FaceNaturalLightSettings = field(default_factory=FaceNaturalLightSettings)
    workerCount: int = 1


@dataclass
class FaceDetectionBackgroundBrokerSettings:
    """Face detection background broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: FaceDetectionBackgroundSettings = field(default_factory=FaceDetectionBackgroundSettings)
    workerCount: int = 1


@dataclass
class Landmarks5BrokerSettings:
    """Landmark5 broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: Landmarks5Settings = field(default_factory=Landmarks5Settings)
    workerCount: int = 1


@dataclass
class Landmarks68BrokerSettings:
    """Landmark68 broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: Landmarks68Settings = field(default_factory=Landmarks68Settings)
    workerCount: int = 1


@dataclass
class BodyAttributesBrokerSettings:
    """Body attributes broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: BodyAttributesSettings = field(default_factory=BodyAttributesSettings)
    workerCount: int = 1


@dataclass
class HumanDetectionsBrokerSettings:
    """Human detections broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 10
    estimatorSettings: HumanDetectorSettings = field(default_factory=HumanDetectorSettings)
    workerCount: int = 1


@dataclass
class HumanTracksBrokerSettings:
    """Human detections broker settings, realize BrokerSettingsProtocol"""

    optimalBatchSize: int = 1000
    estimatorSettings: HumanTrackSettings = field(default_factory=HumanTrackSettings)

    def __deepcopy__(self, memodict={}):
        return self.__class__(self.optimalBatchSize, deepcopy(self.estimatorSettings))


@dataclass
class SDKLoopSettings:
    """
    SDKLoop settings (engine).
    """

    # image orientation broker settings
    imageOrientationBrokerSettings: ImageOrientationBrokerSettings = field(
        default_factory=ImageOrientationBrokerSettings
    )
    # image modification broker settings
    imageModificationBrokerSettings: ImageModificationBrokerSettings = field(
        default_factory=ImageModificationBrokerSettings
    )
    # people count broker settings
    peopleCountBrokerSettings: PeopleCountBrokerSettings = field(default_factory=PeopleCountBrokerSettings)
    # dynamic range broker settings
    dynamicRangeBrokerSettings: DynamicRangeBrokerSettings = field(default_factory=DynamicRangeBrokerSettings)
    # ags broker settings
    agsBrokerSettings: AGSBrokerSettings = field(default_factory=AGSBrokerSettings)
    # human  detector settings
    humanDetectionBrokerSettings: HumanDetectionsBrokerSettings = field(default_factory=HumanDetectionsBrokerSettings)
    portraitStyleBrokerSettings: PortraitStyleBrokerSettings = field(default_factory=PortraitStyleBrokerSettings)

    # face detection broker settings (common for detector, detector+landmarks68, redetector,
    # redetector+landmarks68 brokers)
    faceDetectionBrokerSettings: FaceDetectionBrokerSettings = field(default_factory=FaceDetectionBrokerSettings)
    # face warp broker settings
    faceWarpBrokerSettings: FaceWarpBrokerSettings = field(default_factory=FaceWarpBrokerSettings)
    # face descriptor broker settings (common for all brokers, setup default face descriptor version)
    faceDescriptorBrokerSettings: FaceDescriptorBrokerSettings = field(default_factory=FaceDescriptorBrokerSettings)
    # basic attributes broker settings
    basicAttributesBrokerSettings: BasicAttributesBrokerSettings = field(default_factory=BasicAttributesBrokerSettings)
    # emotions broker settings
    emotionsBrokerSettings: EmotionsBrokerSettings = field(default_factory=EmotionsBrokerSettings)
    # mask broker settings
    maskBrokerSettings: MaskBrokerSettings = field(default_factory=MaskBrokerSettings)
    # face warp quality broker settings
    faceWarpQualityBrokerSettings: FaceWarpQualityBrokerSettings = field(default_factory=FaceWarpQualityBrokerSettings)
    # mouth state broker settings
    mouthStateBrokerSettings: MouthStateBrokerSettings = field(default_factory=MouthStateBrokerSettings)
    # livenessV1 broker settings
    livenessV1BrokerSettings: LivenessV1BrokerSettings = field(default_factory=LivenessV1BrokerSettings)
    # deepfake broker settings
    deepfakeBrokerSettings: DeepfakeBrokerSettings = field(default_factory=DeepfakeBrokerSettings)
    # head pose broker settings
    headPoseBrokerSettings: HeadPoseBrokerSettings = field(default_factory=HeadPoseBrokerSettings)
    # gaze direction broker settings
    gazeDirectionBrokerSettings: GazeDirectionBrokerSettings = field(default_factory=GazeDirectionBrokerSettings)
    # eyes broker settings
    eyesBrokerSettings: EyesBrokerSettings = field(default_factory=EyesBrokerSettings)
    # face occlusion broker settings
    faceOcclusionBrokerSettings: FaceOcclusionBrokerSettings = field(default_factory=FaceOcclusionBrokerSettings)
    # facial hair broker settings
    facialHairBrokerSettings: FacialHairBrokerSettings = field(default_factory=FacialHairBrokerSettings)
    # glasses broker settings
    glassesBrokerSettings: GlassesBrokerSettings = field(default_factory=GlassesBrokerSettings)
    # eyebrow expression broker settings
    eyebrowExpressionBrokerSettings: EyebrowExpressionBrokerSettings = field(
        default_factory=EyebrowExpressionBrokerSettings
    )
    # fisheye broker settings
    fisheyeBrokerSettings: FisheyeBrokerSettings = field(default_factory=FisheyeBrokerSettings)
    # red-eyes broker settings
    redEyesBrokerSettings: RedEyesBrokerSettings = field(default_factory=RedEyesBrokerSettings)
    # face detection background broker settings
    faceDetectionBackgroundBrokerSettings: FaceDetectionBackgroundBrokerSettings = field(
        default_factory=FaceDetectionBackgroundBrokerSettings
    )
    # headwear broker settings
    headwearBrokerSettings: HeadwearBrokerSettings = field(default_factory=HeadwearBrokerSettings)
    # image color type broker settings
    imageColorTypeBrokerSettings: ImageColorTypeBrokerSettings = field(default_factory=ImageColorTypeBrokerSettings)
    # Face natural light broker settings
    faceNaturalLightBrokerSettings: FaceNaturalLightBrokerSettings = field(
        default_factory=FaceNaturalLightBrokerSettings
    )
    # face landmakrs5 broker settings
    landmarks5BrokerSettings: Landmarks5BrokerSettings = field(default_factory=Landmarks5BrokerSettings)
    # face landmakrs68 broker settings
    landmarks68BrokerSettings: Landmarks68BrokerSettings = field(default_factory=Landmarks68BrokerSettings)

    # body detection broker settings (common for detector, redetector, brokers)
    bodyDetectionBrokerSettings: BodyDetectionBrokerSettings = field(default_factory=BodyDetectionBrokerSettings)
    # body warp broker settings
    bodyWarpBrokerSettings: BodyWarpBrokerSettings = field(default_factory=BodyWarpBrokerSettings)
    # body descriptor broker settings (common for all brokers, setup default body descriptor version)
    bodyDescriptorBrokerSettings: BodyDescriptorBrokerSettings = field(default_factory=BodyDescriptorBrokerSettings)
    # body attributes broker settings
    bodyAttributesBrokerSettings: BodyAttributesBrokerSettings = field(default_factory=BodyAttributesBrokerSettings)

    humanTracksBrokerSettings: HumanTracksBrokerSettings = field(default_factory=HumanTracksBrokerSettings)
