import os
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Literal, NamedTuple

from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.image_utils.geometry import Rect
from lunavl.sdk.image_utils.image import MemoryResidence, VLImage
from lunavl.sdk.launch_options import DeviceClass, LaunchOptions
from lunavl.sdk.trackengine.engine import VLTrackEngine
from lunavl.sdk.trackengine.setting_provider import LoggingMode, TrackEngineSettingsProvider, TrackerType
from lunavl.sdk.trackengine.structures import Frame as CoreFrame, StreamParams, TrackingResult

from sdk_loop.estimators.fe import getFE
from sdk_loop.estimators.worker import EstimationBrokerBase, QueueEstimatorBase, _setResult
from sdk_loop.monitoring_utils.estimation_monitoring import EstimationPoint
from sdk_loop.queues.sdkqueue import QTask

from ...monitoring_utils.storages import ESTIMATION_MONITORING as EM


class HumanTrackMonitoringPoint(EstimationPoint):
    """
    Human tracking monitoring point.

    Attributes:
        trackCount: detection count on all images from batch
    """

    __slots__ = ("trackCount",)

    name = "human_track"

    def __init__(
        self,
        executionTime: float = 0,
        batchSize: int = 0,
    ):
        super().__init__(batchSize=batchSize, executionTime=executionTime)
        self.trackCount = 0

    def updateFromResult(self, trackingResult: list[TrackingResult]):
        """
        Update point of tracking results
        Args:
            trackingResult: detection result
        """
        self.trackCount = sum((len(result.humanTracks) for result in trackingResult))

    def getFields(self) -> dict:
        """Get fields for monitoring"""
        fields = super().getFields()
        fields["track_count"] = self.trackCount
        return fields


@dataclass(slots=True)
class TESettings:
    """
    The `tracking engine` settings that are allowed to be modified.

    Other fields are taken from the trackengine.conf file.
    """

    # detect human face or not
    detectFace: Literal[0, 1] = 1
    # detect human body or not
    detectBody: Literal[0, 1] = 0
    # use body reid algorithm for more accuracy tracking or not
    useBodyReId: Literal[0, 1] = 0
    # full detect rate (each several frame run full object detect)
    detectorStep: int | None = None
    # frame will be scaled to this size in pixels for detection step
    scaleResultSize: float | None = None
    # If track wasn't updated by detect/redetect for this number of frames, then track is finished
    skipFrames: int | None = None

    def updateSDKTeSettings(self, teConf: TrackEngineSettingsProvider) -> TrackEngineSettingsProvider:
        """Update TrackEngineSettingsProvider from itself"""
        teConf.detectors.useFaceDetector = self.detectFace
        teConf.detectors.useBodyDetector = self.detectBody
        teConf.bodyTracking.useBodyReid = self.useBodyReId
        if self.detectorStep is not None:
            teConf.other.detectorStep = self.detectorStep
        if self.scaleResultSize is not None:
            teConf.other.scaleResultSize = self.scaleResultSize
        if self.skipFrames is not None:
            teConf.other.skipFrames = self.skipFrames
        return teConf

    def __post_init__(self):
        trackEngineProvider = TrackEngineSettingsProvider()
        if self.detectorStep is None:
            self.detectorStep = trackEngineProvider.other.detectorStep
        if self.scaleResultSize is None:
            self.scaleResultSize = trackEngineProvider.other.scaleResultSize
        if self.skipFrames is None:
            self.skipFrames = trackEngineProvider.other.skipFrames

    def __deepcopy__(self, memodict):
        """Deepcopy realization"""
        return self.__class__(
            detectFace=self.detectFace,
            detectBody=self.detectBody,
            useBodyReId=self.useBodyReId,
            detectorStep=self.detectorStep,
            scaleResultSize=self.scaleResultSize,
            skipFrames=self.skipFrames,
        )


@dataclass(slots=True)
class HumanTrackSettings:
    """Track engine settings."""

    # trackengine launch options
    launchOptions: LaunchOptions = field(default_factory=LaunchOptions)
    # te settings
    te: TESettings = field(default_factory=TESettings)

    def __deepcopy__(self, memodict={}):
        return self.__class__(self.launchOptions, deepcopy(self.te))


LOGGING_MODE = LoggingMode(os.getenv("TE_LOGGING_MODE", "l2c"))


def createTE(teConfig: HumanTrackSettings) -> tuple[VLTrackEngine, VLFaceEngine]:
    """
    Create TrackEngine instance with corresponding configuration.
    """

    trackEngineProvider = TrackEngineSettingsProvider()
    trackEngineProvider = teConfig.te.updateSDKTeSettings(trackEngineProvider)
    trackEngineProvider.faceTracking.faceLandmarksDetection = 0
    trackEngineProvider.bodyTracking.humanLandmarksDetection = 0
    trackEngineProvider.other.trackerType = TrackerType.VL_TRACKER
    trackEngineProvider.other.frgSubtractor = 0
    trackEngineProvider.logging.mode = LOGGING_MODE
    trackEngineProvider.logging.severity = 3  # error
    fe = getFE()
    te = VLTrackEngine(fe, trackEngineConf=trackEngineProvider, launchOptions=teConfig.launchOptions)
    return te, fe


class Frame(NamedTuple):
    """Container for input frames"""

    image: VLImage
    streamId: int
    frameNumber: int
    lastFrame: bool = False

    def toCoreFrame(self) -> CoreFrame:
        """Convert Frame  from"""
        return CoreFrame(self.image, self.frameNumber, self.streamId)


class TEWorker(QueueEstimatorBase[Frame, TrackingResult, VLTrackEngine, HumanTrackSettings]):
    """
    Human tracking estimation worker.

    Attributes:
        pendingTasks: container for frame result waiters. Track engine not always return result by current frame.
        Engine can return result by some previous frames. Worker set these results to corresponding futures.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pendingTasks: dict[str, QTask] = {}

    async def do(self, tasks: list[QTask[Frame, TrackingResult]]):
        """
        Run a single batch estimation for several tasks, and mark tasks as pending.
        Distribute the results obtained from the current batch among pending tasks
        (assuming some results could correspond to tasks from previous batches).
        """
        newTasks = {f"{task.data.streamId}.{task.data.frameNumber}": task for task in tasks}
        self.pendingTasks |= newTasks
        tId = "-".join(newTasks)

        self._logger.debug("start human tracking %s, batch size %i", tId, len(tasks))
        try:
            frames = []
            for task in tasks:
                frame = task.data.toCoreFrame()
                device = self.settings.launchOptions.deviceClass
                frameMemory = frame.image.getMemoryResidence()
                if device == DeviceClass.cpu:
                    if frameMemory == MemoryResidence.GPU:
                        frame = CoreFrame(
                            VLImage(frame.image.coreImage, memoryResidence=MemoryResidence.CPU),
                            frame.frameNumber,
                            frame.streamId,
                        )
                else:
                    if frameMemory == MemoryResidence.CPU:
                        frame.image = CoreFrame(
                            VLImage(frame.image.coreImage, memoryResidence=MemoryResidence.GPU),
                            frame.frameNumber,
                            frame.streamId,
                        )
                frames.append(frame)
            results = await self._estimate(frames)
        except Exception as e:
            self._logger.error("human tracking %s failed", tId)
            for task in tasks:
                frame = task.data
                _setResult(self.pendingTasks.pop(f"{task.data.streamId}.{frame.frameNumber}"), e)
            return
        for result in results:
            _setResult(self.pendingTasks.pop(f"{result.streamId}.{result.frameId}"), result)
        self._logger.debug("human tracking %s succeeded", tId)
        # finalize all ended streams
        for streamId in (task.data.streamId for task in tasks if task.data.lastFrame):
            results = self._estimator.closeStream(streamId)
            for result in results:
                _setResult(self.pendingTasks.pop(f"{result.streamId}.{result.frameId}"), result)

    async def _estimate(self, dataBatch: list[CoreFrame]) -> list[TrackingResult]:
        """Estimate human tracks."""
        point = HumanTrackMonitoringPoint(
            batchSize=len(dataBatch),
        )
        with EM.monitoring(point):
            res = await self._estimator.track(dataBatch, asyncEstimate=True)
            point.updateFromResult(res)
        return res


class TEBroker(EstimationBrokerBase[Frame, TrackingResult, VLTrackEngine, HumanTrackSettings]):
    """Human tracking broker."""

    # broker name
    name = "human_track"

    @property
    def settings(self) -> HumanTrackSettings:
        return self._estimatorSettings

    def createEstimator(self, settings: HumanTrackSettings) -> tuple[VLTrackEngine, VLFaceEngine]:
        """Create track engine."""

        return createTE(settings)

    def registerStream(self) -> int:
        """Register stream for human tracking."""

        streamId = self._estimator.registerStream()
        return streamId

    def configureStream(self, streamId: int, roi: Rect) -> StreamParams:
        """
        Configure stream.

        Args:
            streamId: stream id
            roi: rectangle for perform full detection on frame
        """

        params = self._estimator.getStreamParams(streamId)
        params.roi = roi
        return self._estimator.reconfigure(streamId, params)
