"""Module contains base estimator and broker classes"""

import asyncio
from abc import ABC, abstractmethod
from asyncio import CancelledError
from dataclasses import dataclass, field
from typing import Callable, Generic, Iterable, Optional, Type, TypeVar, Union

from lunavl.sdk.faceengine.engine import VLFaceEngine
from lunavl.sdk.faceengine.setting_provider import FaceEngineSettingsProvider
from lunavl.sdk.launch_options import LaunchOptions

from .fe import DefaultSettingsProvider, getFE
from ..queues.sdkqueue import NamedQueue, QTask
from ..utils.logger import getLogger

# SDK estimator (detector, extractor, ...)
SDKEstimator = TypeVar("SDKEstimator")
# input data for sdk estimator
EstimationData = TypeVar("EstimationData")
# estimation result
Estimation = TypeVar("Estimation")


@dataclass
class EstimatorSettingsBase:
    """Base estimator settings class"""

    launchOptions: LaunchOptions = field(default_factory=LaunchOptions)

    def createFEConf(self) -> tuple[FaceEngineSettingsProvider, bool]:
        """
        Create  FaceEngine config for following estimator creation.
        Returns:
            <faceEngine config, default settings or not>
        """
        faceEngineConf = DefaultSettingsProvider.faceengine
        return faceEngineConf, True


def createEstimator(
    settings: EstimatorSettingsBase, estimatorGetter: Callable[[VLFaceEngine, LaunchOptions], SDKEstimator]
) -> tuple[SDKEstimator, VLFaceEngine]:
    """
    Factory for estimators
    Args:
        settings: estimation settings
        estimatorGetter: estimator getter from VLFaceEngine

    Returns:
        <estimator, FaceEngine which generated estimator>
    """
    configuration, default = settings.createFEConf()
    if default:
        faceEngine = getFE()
    else:
        faceEngine = getFE(configuration)
    estimator = estimatorGetter(faceEngine, settings.launchOptions)
    return estimator, faceEngine


# estimator settings generic
EstimatorSettings = TypeVar("EstimatorSettings", bound=EstimatorSettingsBase)


def _setResult(task: QTask[EstimationData, Estimation], result: Union[Estimation, Exception]):
    """Set estimation result"""
    try:
        if isinstance(result, Exception):
            task.callback.set_exception(result)
        else:
            task.callback.set_result(result)
    except Exception:
        pass


class QueueEstimatorBase(ABC, Generic[EstimationData, Estimation, SDKEstimator, EstimatorSettings]):
    """
    A feature estimator. this estimator collects input data from a queue and batch it.

    Pipeline (see `QueueEstimatorBase._loop`):
        0. Start to listen the queue.
        1. Gets tasks with data from queue while data count less than optimal batch size and queue is not empty.
        2. Estimates features.
        4. Calls callbacks for each collected task an set result to callback.

    Attributes:
        _isRunning (bool): running or not estimator
        _queue (NamedQueue[EstimationData, Estimation]): input task queue
        _executorLoopTask (asyncio.Task): asyncio task with running loop
        _maxBatchSize (int): optimal estimator batch size
        _logger (Logger): estimator logger
        _estimationName (str): estimation name (simple helper)
        settings (EstimatorSettings): estimator settings

    """

    # estimation counter
    _counter = 0
    # worker count
    _workerCount = 0

    def __init__(
        self,
        taskQueue: NamedQueue[EstimationData, Estimation],
        estimator: SDKEstimator,
        maxBatchSize: int,
        settings: EstimatorSettings,
        estimationName: str = "",
    ):
        self.__class__._workerCount += 1
        self._workerCount = self.__class__._workerCount
        self._isRunning = False
        self._queue = taskQueue
        self._executorLoopTask = asyncio.create_task(self._loop())
        self._estimator = estimator
        self._maxBatchSize = maxBatchSize
        self._estimationName = estimationName
        self._logger = getLogger(f"{self._estimationName}_{self._workerCount}_estimator")

        self.settings = settings

    @abstractmethod
    async def _estimate(self, dataBatch: list[EstimationData]) -> Iterable[Estimation]:
        """
        Estimate features
        Args:
            dataBatch:  batch of data for estimation.

        Returns:
            estimated data
        """

    def isRunning(self) -> bool:
        """Get is running or not estimator"""
        return self._isRunning

    def assertData(self, data: EstimationData) -> None:
        """Validate input data"""

    async def _loop(self):
        """
        Infinite estimation loop

        Pipeline (see `QueueEstimatorBase._loop`):
            0. Start to listen the queue.
            1. Gets tasks with data from queue while data count less than optimal batch size and queue is not empty.
            2. Estimates features.
            4. Calls callbacks for each collected task an set result to callback.

        Loop will stop if estimator is not running (or asyncio task is was cancelled).

        """
        self._isRunning = True
        self._logger.debug("estimator starts to wait tasks")
        while self.isRunning:
            # get first task
            task = await self._queue.get()
            tasks = [task]
            # collect optimal batch
            while len(tasks) < self._maxBatchSize and not self._queue.empty():
                tasks.append(self._queue.get_nowait())
            # estimate features
            await self.do(tasks)
            for _ in range(len(tasks)):
                self._queue.task_done()
        self._logger.debug("estimator stop wait task")

    async def do(self, tasks: list[QTask[EstimationData, Estimation]]):
        """
        Do task.
            1. Estimates data from batch.
            2. Calls result (including exceptions) setters for each task from batch

        If estimation is failed, exception will be set to each task.
        Args:
            tasks: input tasks
        """
        self._counter += 1

        tId = self._counter
        dataBatch = []
        correctTasks = []
        for task in tasks:
            try:
                self.assertData(task.data)
                dataBatch.append(task.data)
                correctTasks.append(task)
            except Exception as e:
                _setResult(task, e)

        batchSize = len(dataBatch)
        self._logger.info(f"start estimation {tId}, batch size {batchSize}")
        if correctTasks:
            try:
                results = await self._estimate(dataBatch)
            except Exception as e:
                self._logger.error(f"estimation {tId} are failed", exc_info=e)
                for task in correctTasks:
                    _setResult(task, e)
                return
            for idx, result in enumerate(results):
                _setResult(correctTasks[idx], result)
        self._logger.info(f"estimation {tId} are succeeded")

    async def stop(self, timeout: Optional[float] = None):
        """
        Stop estimator loop. Proceed all current task from queue and stop work
        Args:
            timeout: timeout (None - without timeout)
        """
        self._isRunning = False
        if not self._queue.unfinishedTasks:
            # no task in queue, stop wait
            self._executorLoopTask.cancel()
            try:
                # wait estimation
                await self._executorLoopTask
            except CancelledError:
                pass
            except Exception:
                pass

        else:
            # estimate something, wait
            try:
                async with asyncio.timeout(timeout):
                    await self._executorLoopTask
            except asyncio.TimeoutError:
                self._logger.warning(f"{self._queue.unfinishedTasks} task was not finished")

            self._logger.debug("estimator are stopped")


class EstimationBrokerBase(ABC, Generic[EstimationData, Estimation, SDKEstimator, EstimatorSettings]):
    """
    Estimator broker.

    Broker create a future, put task with the future to a queue and wait the future with results.
    Attributes:
        _estimatorSettings: estimator settings
        _estimator (SDKEstimator): sdk estimator, common for all workers
        _queue (NamedQueue): task queue, transport between broker and workers
        _futureFactory (Callable[[], asyncio.Future]): asyncio future factory
        _workers (list[QueueEstimatorBase]):  list of workers
        _lock (bool): active broker or not for new task


    """

    # estimation name
    name: str

    def __init__(
        self,
        optimalBatchSize: Optional[int],
        estimatorSettings: EstimatorSettings,
        workerClass: Type[QueueEstimatorBase[EstimationData, Estimation, SDKEstimator, EstimatorSettings]],
        workerCount: int = 1,
        queueName="",
        **kwargs,
    ):
        self._estimatorSettings = estimatorSettings
        self._queue: NamedQueue[EstimationData, Estimation] = NamedQueue(
            name=queueName if queueName else f"{self.name}_queue"
        )
        self._futureFactory: Callable[[], asyncio.Future] = asyncio.get_running_loop().create_future
        self._estimator, self._faceEngine = self.createEstimator(estimatorSettings)
        self._workers = [
            workerClass(
                taskQueue=self._queue,
                maxBatchSize=optimalBatchSize,
                estimator=self._estimator,
                settings=estimatorSettings,
                estimationName=self.name,
            )
            for _ in range(workerCount)
        ]
        self._lock = False

    @abstractmethod
    def createEstimator(self, estimatorSettings: EstimatorSettings) -> tuple[SDKEstimator, VLFaceEngine]:
        """
        Create estimator abstract method
        Args:
            estimatorSettings:  estimator settings

        Returns:
            estimator and FaceEngine
        """

    async def stop(self, timeout: Optional[float] = None):
        """
        Stop broker and estimators
        Args:
            timeout: timeout (none - without timeot)
        """
        self._lock = True
        await asyncio.gather(*[worker.stop(timeout) for worker in self._workers])
        self._workers = []

    def isRunning(self) -> bool:
        """Broker is stop if all estimators is stopped"""
        return any(*(worker.isRunning() for worker in self._workers))

    async def estimate(self, data: EstimationData) -> Estimation:
        """
        Estimate features.

        Args:
            data: input data for estimator
        Raises:
            StopIteration: if broker is locked

        Returns:
            estimation result
        """
        if self._lock:
            raise StopIteration
        callBackFuture = self._futureFactory()
        self._queue.put_nowait(QTask(data, callBackFuture))
        return await callBackFuture
