"""Module contains utils for parsing image exif"""

from collections import namedtuple
from enum import Enum
from typing import Any, Iterable

from PIL.Image import Image as PilImage
from PIL.TiffImagePlugin import IFDRational


class Orientation(Enum):
    """Exif orientation enum"""

    NOTHING = 1
    FLIPPED_LEFT_RIGHT = 2
    ROTATED_180 = 3
    FLIPPED_TOP_BOTTOM = 4
    TRANSPOSED = 5
    ROTATED_90 = 6
    TRANSVERSED = 7
    ROTATED_270 = 8


def convertCoordinatesToStr(coords: Iterable[IFDRational]) -> str:
    """Convert gps coordinates to string. Accepts nan and returns zeros in this case."""
    result = []
    for coord in coords:
        try:
            result.append(coord.numerator / coord.denominator)
        except ZeroDivisionError:
            result.append(0.0)
    return ", ".join(map(str, result))


def getGPS(gpsTag: dict) -> dict:
    """
    Get gps tag in api format

    Args:
        gpsTag: pillow gps tag

    Returns:
        dict with 'latitude' and 'longitude'
    """
    gpsInfo = {
        "latitude": convertCoordinatesToStr(gpsTag.get(0x0002, (IFDRational(0),) * 3)),
        "longitude": convertCoordinatesToStr(gpsTag.get(0x0004, (IFDRational(0),) * 3)),
    }
    return gpsInfo


ExifTagHandler = namedtuple("ExifTagHandler", ("apiTagName", "tagValueGetter", "apiTagValueMaker"))
#  getters from exif for supported tags
tagHandlers = [
    ExifTagHandler("flash", lambda exif: exif.get_ifd(0x8769).get(0x9209), str),
    ExifTagHandler("gps", lambda exif: exif.get_ifd(0x8825), getGPS),
    ExifTagHandler("artist", lambda exif: exif.get(0x013B), str),
    ExifTagHandler("model", lambda exif: exif.get(0x0110), str),
    ExifTagHandler("make", lambda exif: exif.get(0x010F), str),
    ExifTagHandler("software", lambda exif: exif.get(0x0131), str),
    ExifTagHandler("orientation", lambda exif: exif.get(0x0112), int),
]


def getExif(image: PilImage) -> dict[str, Any]:
    """
    Get exif data from image.
    See PIL.ExifTags.TAGS
    For details see also:
    https://pillow.readthedocs.io/en/stable/releasenotes/8.2.0.html#image-getexif-exif-and-gps-ifd

    Args:
        image: input image

    Returns:
        dict with exif data
    """

    exifInfo = image.getexif()
    exifData = {
        handler.apiTagName: handler.apiTagValueMaker(tagValue)
        for handler in tagHandlers
        if (tagValue := handler.tagValueGetter(exifInfo))
    }
    return exifData
