from abc import ABC, abstractmethod
from enum import Enum
from typing import Generic, Iterator, TypeVar

FilterValue = TypeVar("FilterValue")
ObjectValue = TypeVar("ObjectValue")


def _serialize(value):
    """
    Serialize value to json serializable object.

    Raises:
        ValueError: if structure not supported
    """
    if isinstance(value, (int, float, str, dict)):
        return value
    elif isinstance(value, Enum):
        return _serialize(value.value)
    elif isinstance(value, (set, list)):
        return [_serialize(item) for item in value]
    elif hasattr(value, "asDict"):
        return value.asDict()
    raise ValueError("Unknown type for serealization")


class FiltrationResult(Generic[FilterValue, ObjectValue]):
    """
    Filtration result container.

    Attributes:
        value: value
        threshold: filter value
        result: filtration result
        filterName: filter name
    """

    __slots__ = ("value", "threshold", "result", "filterName")

    def __init__(self, threshold: FilterValue, value: ObjectValue, result: bool, filterName: str):
        self.value = value
        self.threshold = threshold
        self.result = result
        self.filterName = filterName

    def asDict(self) -> dict:
        """convert filtration result to serializable dict"""
        value = _serialize(self.value)
        threshold = _serialize(self.threshold)
        return {"filter_name": self.filterName, "object_value": value, "threshold_value": threshold}


class BaseFilter(Generic[FilterValue, ObjectValue], ABC):
    __slots__ = ("threshold",)

    def __init__(self, threshold: FilterValue):
        self.threshold = threshold

    @abstractmethod
    def filter(self, value: ObjectValue) -> FiltrationResult[FilterValue, ObjectValue]:
        """"""

    def __repr__(self):
        return f"threshold: {self.threshold}"


class FiltrationResults:
    """
    Container for filtration results
    Attributes:
        _filters (list[FiltrationResult]): filtration results
    """

    def __init__(self):
        self._filters = []

    def append(self, filtrationResult: FiltrationResult):
        """Add new result"""
        self._filters.append(filtrationResult)

    def __iter__(self) -> Iterator[FiltrationResult]:
        """Iterator by results"""
        return iter(self._filters)

    def isFiltered(self) -> bool:
        """If any filtration result is True return True otherwise False"""
        return any((filtrationResult.result for filtrationResult in self._filters))

    def asDict(self) -> list[dict]:
        """convert to serializable list"""
        return [filtrationResult.asDict() for filtrationResult in self._filters]

    @property
    def filters(self) -> list[FiltrationResult]:
        """Get filters"""
        return self._filters
