from typing import Literal

from attr import dataclass
from lunavl.sdk.estimators.face_estimators.deepfake import DeepfakePrediction
from lunavl.sdk.estimators.face_estimators.face_occlusion import FaceOcclusion
from lunavl.sdk.estimators.face_estimators.livenessv1 import LivenessPrediction
from lunavl.sdk.estimators.face_estimators.mask import MaskState

from .base import BaseFilter, FiltrationResult


class MaskFilter(BaseFilter[list[MaskState], MaskState]):
    def filter(self, value: MaskState) -> FiltrationResult[list[MaskState], MaskState]:
        result = value not in self.threshold
        return FiltrationResult(self.threshold, value=value, result=result, filterName="mask_states")


class LivenessFilter(BaseFilter[list[LivenessPrediction], LivenessPrediction]):
    def filter(self, value: LivenessPrediction) -> FiltrationResult[list[LivenessPrediction], LivenessPrediction]:
        result = value not in self.threshold
        return FiltrationResult(self.threshold, value=value, result=result, filterName="liveness_states")


class DeepfakeFilter(BaseFilter[list[DeepfakePrediction], DeepfakePrediction]):
    def filter(self, value: DeepfakePrediction) -> FiltrationResult[list[DeepfakePrediction], DeepfakePrediction]:
        result = value not in self.threshold
        return FiltrationResult(self.threshold, value=value, result=result, filterName="deepfake_states")


class HeadPoseYawFilter(BaseFilter[float, float]):
    def filter(self, value: float) -> FiltrationResult[float, float]:
        result = abs(value) > self.threshold
        return FiltrationResult(self.threshold, value=value, result=result, filterName="yaw_threshold")


class HeadPosePitchFilter(BaseFilter[float, float]):
    def filter(self, value: float) -> FiltrationResult[float, float]:
        result = abs(value) > self.threshold
        return FiltrationResult(self.threshold, value=value, result=result, filterName="pitch_threshold")


class HeadPoseRollFilter(BaseFilter[float, float]):
    def filter(self, value: float) -> FiltrationResult[float, float]:
        result = abs(value) > self.threshold
        return FiltrationResult(self.threshold, value=value, result=result, filterName="roll_threshold")


class FaceGCFilter(BaseFilter[float, float]):
    def filter(self, value: float) -> FiltrationResult[float, float]:
        result = value < self.threshold
        return FiltrationResult(self.threshold, value=value, result=result, filterName="score_threshold")


@dataclass(slots=True)
class FaceOcclusionThreshold:
    """
    Face occlusion threshold params
    """

    # occlusion threshold
    face: Literal[0, 1] | None = None
    # forehead threshold
    forehead: Literal[0, 1] | None = None
    # eye threshold
    eye: Literal[0, 1] | None = None
    # nose threshold
    nose: Literal[0, 1] | None = None
    # mouth threshold
    mouth: Literal[0, 1] | None = None
    # lower face threshold
    lowerFace: Literal[0, 1] | None = None

    def asDict(self):
        return {
            "face": self.face,
            "forehead": self.forehead,
            "eye": self.eye,
            "nose": self.nose,
            "mouth": self.mouth,
            "lower_face": self.lowerFace,
        }


class FaceOcclusionFilter(BaseFilter[FaceOcclusionThreshold, FaceOcclusion]):
    def filter(self, value: FaceOcclusion) -> FiltrationResult[FaceOcclusionThreshold, FaceOcclusion]:
        threshold = self.threshold
        result = 0
        if threshold.face is not None and value.overall.state != threshold.face:
            result = 1
        if threshold.forehead is not None and value.forehead.state != threshold.forehead:
            result = 1
        if threshold.eye is not None and (
            value.leftEye.state != threshold.eye or value.rightEye.state != threshold.eye
        ):
            result = 1
        if threshold.nose is not None and value.nose.state != threshold.nose:
            result = 1
        if threshold.mouth is not None and value.mouth.state != threshold.mouth:
            result = 1
        if threshold.lowerFace is not None and value.lowerFace.state != threshold.lowerFace:
            result = 1
        _value = value.asDict()["face_occlusion"]
        _value["face_occluded"] = value.overall.state
        return FiltrationResult(threshold=threshold, value=_value, result=result, filterName="face_occlusion_states")
