"""Module contains global monitoring pont storages"""

from contextlib import contextmanager
from time import perf_counter
from typing import ContextManager

from .monitoring import MonitoringPoint


class Storage:
    """
    Container for monitoring points

    Attributes:
        name: storage name
        enable: enable or not monitoring point storing
        _values: map - keys - series name, values - list monitoring points
    """

    def __init__(self, name: str, enable: bool = True):
        self._values = {}
        self.name = name
        self.enable = enable

    @contextmanager
    def popValues(self) -> ContextManager[dict[str, list[MonitoringPoint]]]:
        """pop points from  the storage"""
        yield self._values
        # if correct processing points, remove them from the store
        self._values = {}

    def append(self, point: MonitoringPoint):
        """Add point to the storage"""
        self._values.setdefault(point.name, []).append(point)

    @contextmanager
    def monitoring(self, point: MonitoringPoint):
        """
        Collect monitoring point ti storage (a event duration measurement).
        """
        if not self.enable:
            yield point
            return
        start = perf_counter()
        try:
            yield point
        finally:
            point.executionTime = perf_counter() - start
            self._values.setdefault(point.name, []).append(point)


# The storage contains an estimation monitoring info from all estimators
ESTIMATION_MONITORING = Storage("sdk_estimation", enable=False)
# The storage contains a load image info
IMAGE_LOAD_MONITORING = Storage("image_load", enable=False)
# The storage contains a task processing monitoring info
TASK_EXECUTION_MONITORING = Storage("task_execution", enable=False)
