"""Module contains task filters containers"""

from typing import Optional

import attr
from attr import dataclass
from lunavl.sdk.estimators.face_estimators.deepfake import DeepfakePrediction
from lunavl.sdk.estimators.face_estimators.livenessv1 import LivenessPrediction
from lunavl.sdk.estimators.face_estimators.mask import MaskState

from ..models.filtration.base import BaseFilter
from ..models.filtration.face_filters import (
    DeepfakeFilter,
    FaceGCFilter,
    FaceOcclusionFilter,
    FaceOcclusionThreshold,
    HeadPosePitchFilter,
    HeadPoseRollFilter,
    HeadPoseYawFilter,
    LivenessFilter,
    MaskFilter,
)


class FaceDetectionFilters:
    """
    Face detection filters
    Attributes:
        yawFilter (Optional[HeadPoseYawFilter]): filter by yaw angle
        pitchFilter Optional[HeadPosePitchFilter]): filter by pitch angle
        rollFilter (Optional[HeadPoseRollFilter]): filter by roll angle
        livenessFilter (Optional[LivenessFilter]): filter by liveness
        deepfakeFilter (Optional[DeepfakeFilter]): filter by liveness
        maskFilter (Optional[MaskFilter]): filter by masks
        gcFilter (Optional[FaceGCFilter]): filter by descriptor garbage score
        faceOcclusion(FaceOcclusionThreshold | None): filter by face occlusion parameters
    """

    def __init__(
        self,
        yawThreshold: Optional[float] = None,
        pitchThreshold: Optional[float] = None,
        rollThreshold: Optional[float] = None,
        livenessStates: Optional[list[LivenessPrediction]] = None,
        maskStates: Optional[list[MaskState]] = None,
        gcThreshold: Optional[float] = None,
        deepfakeStates: Optional[list[DeepfakePrediction]] = None,
        faceOcclusion: FaceOcclusionThreshold | None = None,
    ):
        self.yawFilter = HeadPoseYawFilter(yawThreshold) if yawThreshold is not None else None
        self.pitchFilter = HeadPosePitchFilter(pitchThreshold) if pitchThreshold is not None else None
        self.rollFilter = HeadPoseRollFilter(rollThreshold) if rollThreshold is not None else None
        self.livenessFilter = LivenessFilter(livenessStates) if livenessStates is not None else None
        self.maskFilter = MaskFilter(maskStates) if maskStates is not None else None
        self.gcFilter = FaceGCFilter(gcThreshold) if gcThreshold is not None else None
        self.deepfakeFilter = DeepfakeFilter(deepfakeStates) if deepfakeStates is not None else None
        self.faceOcclusionFilter = FaceOcclusionFilter(faceOcclusion) if faceOcclusion is not None else None

    def __str__(self) -> str:
        """Convert to string"""
        return str(self.asDict())

    def asDict(self) -> dict[str, Optional[BaseFilter]]:
        """Convert filters to dict"""
        return {
            "yawFilter": self.yawFilter,
            "pitchFilter": self.pitchFilter,
            "rollFilter": self.rollFilter,
            "livenessFilter": self.livenessFilter,
            "maskFilter": self.maskFilter,
            "gcFilter": self.gcFilter,
            "deepfakeFilter": self.deepfakeFilter,
            "faceOcclusionFilter": self.faceOcclusionFilter,
        }


@dataclass(slots=True)
class Filters:
    """Task filters"""

    faceDetection: FaceDetectionFilters = attr.field(factory=FaceDetectionFilters)
