"""Module contains strategies for getting best detection on image"""

from functools import cmp_to_key
from typing import TypeVar

from lunavl.sdk.detectors.base import BaseDetection

#: generic type for detector subtasks
from ..models.sample import Sample

DetectorType = TypeVar("DetectorType", bound=BaseDetection)
CONF_THR = 0.65  # confidence detection comparison threshold
AREA_THR = 1.5  # area detection comparison threshold


def _compareCentreManhattan(detection1: DetectorType, detection2: DetectorType) -> int:
    """
    Compare two detections by manhattan distance to the center of the image
    Args:
        detection1: first detection to compare
        detection2: second detection to compare

    Returns:
        1 if detection1 is better than detection2 by manhattan distance criterion else -1

    """
    imageCenter = detection1.image.rect.center
    center1 = detection1.boundingBox.rect.center
    center2 = detection2.boundingBox.rect.center
    distance1 = abs(center1.x - imageCenter.x) + abs(center1.y - imageCenter.y)
    distance2 = abs(center2.x - imageCenter.x) + abs(center2.y - imageCenter.y)
    if distance1 < distance2:
        return 1
    return -1


def _compareConfidence(detection1: DetectorType, detection2: DetectorType) -> int:
    """
    Compare two detections by confidence
    Args:
        detection1: first detection to compare
        detection2: second detection to compare

    Returns:
        1 if detection1 is better than detection2 by confidence criterion else -1
    """
    if detection1.boundingBox.score > detection2.boundingBox.score:
        return 1
    return -1


def _comparerV3(sample1: Sample, sample2: Sample) -> int:
    """
    Compare two samples by combined criteria of detection confidence, detection area and
        manhattan distance from detection center to the center of the image
    Args:
        sample1: first sample to compare with face detection
        sample2: second sample to compare with face detection

    Returns:
        1 if sample1 face detection is better than sample2 face detection by combined criterion else -1

    """
    if sample2.face is None:
        return 1
    if sample1.face is None:
        return -1
    detection1 = sample1.face.detection
    detection2 = sample2.face.detection
    if detection1.boundingBox.score > CONF_THR and detection2.boundingBox.score > CONF_THR:
        areaRatio = detection1.boundingBox.rect.getArea() / detection2.boundingBox.rect.getArea()
        if areaRatio > AREA_THR:
            return 1
        if areaRatio < 1 / AREA_THR:
            return -1
        return _compareCentreManhattan(detection1, detection2)
    return _compareConfidence(detection1, detection2)


def getBestDetection(detections: list[Sample]) -> Sample:
    """
    Get best sample from the image.

    Returns:
        best sample from the image
    """
    detections.sort(key=cmp_to_key(_comparerV3), reverse=True)
    return detections[0]
