"""
Module contains functions for image and coordinates rotation
"""

from typing import Iterable, List, Union

from lunavl.sdk.detectors.base import ImageForRedetection
from lunavl.sdk.estimators.image_estimators.orientation_mode import OrientationType
from lunavl.sdk.image_utils.geometry import COORDINATE_TYPE, Rect
from lunavl.sdk.image_utils.image import RotationAngle, VLImage


def getImageRotationAngle(imageOrientationMode: OrientationType) -> RotationAngle:
    """
    Get rotation angle for image/bounding boxes according due to original image orientation mode.
    Image/bounding box rotated at the received angle must have normal orientation mode.
    Args:
        imageOrientationMode: original image orientation mode
    Returns:
        image/bounding box rotation angle
    """
    if imageOrientationMode == OrientationType.LEFT:
        return RotationAngle.ANGLE_270
    if imageOrientationMode == OrientationType.RIGHT:
        return RotationAngle.ANGLE_90
    if imageOrientationMode == OrientationType.UPSIDE_DOWN:
        return RotationAngle.ANGLE_180
    return RotationAngle.ANGLE_0


def rotateRect(rect: Rect, imageSize: Iterable[COORDINATE_TYPE], angle: RotationAngle, flip: bool = False) -> Rect:
    """
    Rotate rectangle according due to image size and rotation angle
    Args:
        rect: dict with x, y, width and height
        imageSize: iterable with image width and height
        angle: rotation angle
        flip: flip image or not
    Returns:
        dict with rotated x, y, width and height
    >>> rotateRect(rect=Rect(x=494, y=72, width=405, height=538), imageSize=[1024, 768], angle=RotationAngle.ANGLE_0)
    x = 494, y = 72, width = 405, height = 538
    >>> rotateRect(rect=Rect(x=158, y=494, width=538, height=405), imageSize=[768, 1024], angle=RotationAngle.ANGLE_90)
    x = 494, y = 72, width = 405, height = 538
    >>> rotateRect(rect=Rect(x=125, y=158, width=405, height=538), imageSize=[1024, 768], angle=RotationAngle.ANGLE_180)
    x = 494, y = 72, width = 405, height = 538
    >>> rotateRect(rect=Rect(x=72, y=125, width=538, height=405), imageSize=[768, 1024], angle=RotationAngle.ANGLE_270)
    x = 494, y = 72, width = 405, height = 538
    """
    xCoordinate, yCoordinate, width, height = rect.x, rect.y, rect.width, rect.height
    imageWidth, imageHeight = imageSize
    if angle == RotationAngle.ANGLE_90:
        xCoordinate, yCoordinate, width, height = yCoordinate, imageWidth - width - xCoordinate, height, width
        imageWidth, imageHeight = imageHeight, imageWidth
    if angle == RotationAngle.ANGLE_180:
        xCoordinate, yCoordinate = imageWidth - width - xCoordinate, imageHeight - height - yCoordinate
    if angle == RotationAngle.ANGLE_270:
        xCoordinate, yCoordinate, width, height = imageHeight - height - yCoordinate, xCoordinate, height, width
        imageWidth, imageHeight = imageHeight, imageWidth
    if flip:
        return Rect(x=imageWidth - xCoordinate - width, y=yCoordinate, width=width, height=height)
    return Rect(x=xCoordinate, y=yCoordinate, width=width, height=height)


def rotateBBoxes(
    bboxes: List[Rect], imageWidth: int, imageHeight: int, angle: RotationAngle, flip: bool = False
) -> List[Rect]:
    """Rotate bounding boxes to normal orientation"""
    rotatedBBoxes = []
    for bbox in bboxes:
        rotatedRect = rotateRect(rect=bbox, imageSize=[imageWidth, imageHeight], angle=angle, flip=flip)
        rotatedBBoxes.append(rotatedRect)
    return rotatedBBoxes


def rotateImage(
    originImage: Union[VLImage, ImageForRedetection], angle: RotationAngle
) -> Union[VLImage, ImageForRedetection]:
    """
    Rotate image to normal state to normal orientation mode
    Args:
        originImage: incoming image
        angle: original image orientation mode
    Returns:
        image with normal orientation mode
    """
    if isinstance(originImage, VLImage):
        return VLImage.rotate(originImage, angle)
    if isinstance(originImage, ImageForRedetection):
        return ImageForRedetection(
            image=VLImage.rotate(originImage.image, angle),
            bBoxes=rotateBBoxes(
                bboxes=originImage.bBoxes,
                imageWidth=originImage.image.rect.width,
                imageHeight=originImage.image.rect.height,
                angle=angle,
            ),
        )
    raise RuntimeError("Unsupported image type")
